/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.core;

//import java.util.Set;
import java.util.*;
//import java.io.Serializable;

/**
 * representing semantic contents in a document and/or its imported documents
 * @author Z. Pan
 * @version 1.0
 */

/**
 * change history:
 * 2004/02/03 Pan: add descendants.
 */


public interface Document extends java.io.Serializable{

  /**
   * Add an assertion to this document.
   */
  public void addAssertion(ObjectId subject, ObjectId predicate, ObjectId object);

  /**
   * Delete an assertion in this document.
   */
  public void deleteAssertion(Assertion a);

  /**
   * Get all assertions.
   */
  public Assertion[] getAllAssertions();

  /**
   * Get the Iterator of all the assertions.
   */
  public Iterator getAssertionIterator();

  /**
   * Get the imports statements.
   */
  public DocId[] getImports();

  /**
   * Add an imports statement.
   */
  public void addImports(DocId id);

  /**
   * Delete an imports statement.
   */
  public void deleteImports(DocId id);

  /**
   * get the ID of this document
   */
  public DocId getID();

  /**
   * set the ID of this document
   */
  public void setID(DocId id);

  /**
   * Add a set of distinct individuals' IDs
   * $change: collection allow duplicate members, which we wanna
   * accommodate here.
   */
  public void addDistinctIndividualsSet(Collection dis);

  /**
   * delete a set of distinct individuals' IDs
   */
  public void deleteDistinctIndividualsSet(Collection dis);

  /**
   * get sets of distinct individuals' IDs
   */
  public Collection[] getDistinctIndividualsSets();

  /**
   * Add a set of equivalent individuals' IDs
   */
  public void addSameIndividualsSet(Collection dis);

  /**
   * delete a set of equivalent individuals' IDs
   */
  public void deleteSameIndividualsSet(Collection dis);

  /**
   * get sets of equivalent individuals' IDs
   */
  public Collection[] getSameIndividualsSets();


  /**
   * Create a Individual object.
   */
  public Individual createIndividual();


    /**
   * Add an individual.
   */
  public void addIndividual(Individual indi);

  /**
   * Delete an individual.
   */
  public void deleteIndividual(Individual indi);

  /**
   * Get an individual.
   */
  public Individual getIndividual(ObjectId id);

  /**
   * Get all the individuals.
   */
  public Individual[] getAllIndividuals();

  /**
   * Get the Iterator of all the individuals.
   */
  public Iterator getIndividualIterator();

  /**
   * Set the nameSpaces of this document.
   */
  public void setNameSpaces(Map ns);

  /**
   * Get the nameSpaces of this document.
   */
  public Map getNameSpaces();

  /**
   * Set the xml:Base of this document.
   */
  public void setXMLBase(String base);

  /**
   * Get the xml:Base of this document.
   */
  public String getXMLBase();

  /**
   * Add a description.
   */
  public void addDescription(String s);

  /**
   * Delete a description.
   */
  public void deleteDescriprion(String s);

  /**
   * Get the first description.
   */
  public String getDescription();

  /**
   * Get all the descriptions.
   */
  public Iterator getAllDescriptions();

  /**
   * Set the label.
   */
  public void setLabel(String s);

  /**
   * Get the label.
   */
  public String getLabel();

  /**
   * Set descendants.
   */
  public void setDescendants(Stack des);

  /**
   * Get descendants.
   */
  public Stack getDescendants();

  /**
   * Get the number of triples in this document.
   */
  public int getSize();

  /**
   * Set the number of triples in this document.
   */
  public void setSize(int s);

  /**
   * Flush this document.
   * May be needed for database storage models.
   */
  public void flush();
  }
