/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.core;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

/**
 * encapsulating the web ontology </p>
 * @author Z. Pan
 * @version 1.0
 */

public interface Ontology extends Document{


  /**
   * add a class to this ontology
   */
  public void addOntClass(OntClass c);

  /**
   * get the specified class
   */
  public OntClass getOntClass(ObjectId id);

  /**
   * delete the specified class
   */
  public void deleteOntClass(ObjectId id);

  /**
   * get all the classes in this ontology
   */
  public OntClass[] getAllClasses();

  /**
   * Get the Iterator of all the classes.
   */
  public Iterator getClassIterator();

  /**
   * add a property to this ontology
   */
  public void addProperty(OntProperty p);

  /**
   * get the specified property
   */
  public OntProperty getProperty(ObjectId id);

  /**
   * get the object (class/property) given the ID
   */
  public OntObject getOntObject(ObjectId id);

  /**
   * delete a specified property
   */
  public void deleteProperty(ObjectId id);

  /**
   * get all the properties in this ontology
   */
  public OntProperty[] getAllProperties();

  /**
   * Get the Iterator of all the properties.
   */
  public Iterator getPropertyIterator();

  /**
   * add a property to this ontology
   */
  public void addRestriction(PropertyRestriction p);

  /**
   * get the specified property
   */
  public PropertyRestriction getRestriction(ObjectId id);

  /**
   * delete a specified property
   */
  public void deleteRestriction(ObjectId id);

  /**
   * get all the properties in this ontology
   */
  public PropertyRestriction[] getAllRestrictions();

  /**
   * Get the priorVersion ontologies which are backwardCompatibleWith
   */
  public DocId[] getCompatiblePriorOnts();

  /**
   * Add a priorVersion ontology which is backwardCompatibleWith.
   */
  public void addCompatiblePriorOnts(DocId id);

  /**
   * Delete a priorVersion ontology which is backwardCompatibleWith.
   */
  public void deleteCompatiblePriorOnts(DocId id);

  /**
   * Get the priorVersion ontologies may be incompatibleWith
   */
  public DocId[] getPriorOnts();

  /**
   * Add a priorVersion ontology may be incompatibleWith.
   */
  public void addPriorOnts(DocId id);

  /**
   * Delete a priorVersion ontology may be incompatibleWith.
   */
  public void deletePriorOnts(DocId id);

  /**
   * get the version of this ontology
   */
  public String getVersion();

  /**
   * Set the version of this ontology.
   */
  public void setVersion(String ver);

  /**
   * Create a Class object.
   */
  public OntClass createClass();

  /**
   * Create a Property object.
   */
  public OntProperty createProperty();

  /**
   * Create a PropertyRestriction object.
   */
  public PropertyRestriction createRestriction();

  /**
   * To see if ontology has the class.
   */
  public boolean hasClass(String name);

  /**
   * To see if ontology has the property.
   */
  public boolean hasProperty(String name);

  /**
   * Get All the classes plus the imported Classes.
   */
  public OntClass[] getAllPlusImportedClasses();

  /**
   * Get All the classes plus the imported properties.
   */
  public OntProperty[] getAllPlusImportedProperties();

  /**
   * To tell if the ontology imorted specified class.
   */
  public boolean importedClass(ObjectId cid);

  /**
   * To tell if the ontology imported specified property.
   */
  public boolean importedProperty(ObjectId pid);

  /**
   * Add an imported class.
   */
  public void addImportedClass(ObjectId cid, edu.lehigh.swat.hawk.core.OntClass c);

  /**
   * Add an imported property.
   */
  public void addImportedProperty(ObjectId pid, edu.lehigh.swat.hawk.core.OntProperty p);

  /**
   * To tell if the ontology asserted specified class.
   */
  public boolean assertedClass(ObjectId cid);

  /**
   * To tell if the ontology asserted specified property.
   */
  public boolean assertedProperty(ObjectId pid);

  /**
   * Add an asserted class.
   */
  public void addAssertedClass(ObjectId cid);

  /**
   * Add an asserted property.
   */
  public void addAssertedProperty(ObjectId pid);

  /**
   * Add annotations.
   */
  public boolean addAnnotations(ObjectId sId, ObjectId oId, String value);

  /**
   * Get annotations by subject ID.
   */
  public Vector<Annotation> getAnnotationsBySubject(ObjectId pid);

  /**
   * Get annotations by object ID.
   */
  public Vector<Annotation> getAnnotationsByObject(ObjectId pid);

  /**
   * Get all annotations.
   */
  public Vector<Annotation> getAllAnnotations();


  /******************** Axiom Design Methods ***************************/

  public void addAxiom(Axiom a);
  public void addAxiomByObject(ObjectId obj, Axiom a);
  public void addAxiomBySubject(ObjectId sub, Axiom a);
  public void addAxiomByType(Integer type, Axiom a);
  public void addAxiomNodeRestriction(AxiomNode res);
  public void addAllAxiomNodeRestrictions(Vector<AxiomNode> res);
  public List<Axiom> getAllAxioms();
  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsByObject();
  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsBySubject();
  public Hashtable<Integer, Vector<Axiom>> getAllAxiomsByType();
  public List<Axiom> getAxiomByObject(ObjectId object);
  public List<Axiom> getAxiomBySubject(ObjectId subject);
  public List<Axiom> getAxiomByType(Integer type);
  public Vector<AxiomNode> getAxiomNodeRestrictions();

  // Ameet Dec 11, 2008
  public List<Axiom> getAllPlusImportedAxioms();
  public Hashtable<ObjectId, Vector<Axiom>> getAllPlusImportedAxiomsByObject();
  public Hashtable<ObjectId, Vector<Axiom>> getAllPlusImportedAxiomsBySubject();
  public Hashtable<Integer, Vector<Axiom>> getAllPlusImportedAxiomsByType();
  public void addImportedAxiom(Axiom a);

  public final static int SUBCLASSOF = 1;
  public final static int SUBPROPERTYOF = 2;
  public final static int EQUIVALENTCLASS = 3;
  public final static int EQUIVALENTPROPERTY = 4;
  public final static int INVERSEOF = 5;
  public final static int DISJOINTWITH = 6;
  public final static int COMPLEMENTOF = 7;
  public final static int INTERSECTIONOF = 8;
  public final static int UNIONOF = 9;
  public final static int ONEOF = 10;
  public final static int ATOMICNODE = 0;
  public final static int CARDINALITYRESTRICTION = -1;
  public final static int HASVALUERESTRICTION = -2;
  public final static int SOMEVALUERESTRICTION = -3;
  public final static int ALLVALUERESTRICTION = -4;
  public final static int DOMAIN = -5;
  public final static int RANGE = -6;

  /*********************************************************************/
}
