/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.owl;
import java.io.*;
import edu.lehigh.swat.hawk.core.*;
import edu.lehigh.swat.hawk.storage.StorageModel;
//import edu.lehigh.swat.hawk.storage.simplememory.Ontology;

import com.hp.hpl.jena.rdf.arp.JenaReader;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.rdf.model.impl.SelectorImpl;
import java.util.*;


/**
 * OWL reader, reads OWL documents.
 * @author Z. Pan
 * @version 1.0
 */

public class OwlRdfReader extends OntReader{

  /**
   * Constructor
   */
  public OwlRdfReader(StorageModel s) throws Exception{
    super();
    smodel = s;
    try{
      resource = Class.forName("com.hp.hpl.jena.rdf.model.Resource");
    }catch(Exception e){
      e.printStackTrace();
    }
  }

  /**
   * Constructor
   */
  public OwlRdfReader(StorageModel s, String blfile) throws Exception{
    super();
    smodel = s;
    try{
      resource = Class.forName("com.hp.hpl.jena.rdf.model.Resource");
    }catch(Exception e){
      e.printStackTrace();
    }
    BufferedReader br = new BufferedReader(new FileReader(blfile));
    String line = null;
    blacklist = new Vector();
    while ( (line = br.readLine()) != null) {
        blacklist.add(line);
    }
    br.close();
  }

  /**
   * Use empty string as XMl base.
   */
  public Document read(InputStream in)throws Exception{
    return this.read(in,"");
  }

  /**
   * Read document from Reader.
   */
/*  public Document read(Reader reader, String xmlBase){
    Document d = smodel.getDocument(xmlBase);
    if(d!=null)
      return d;

    PreParser pp = new PreParser(reader);
    HashMap map = pp.getNamespaces();
    if(xmlBase=="")
      xmlBase=pp.getXMLBase();
// HashMap map = new HashMap();
    JenaReader jenaReader = new JenaReader();
    Model model = ModelFactory.createDefaultModel();
    jenaReader.read(model, reader, xmlBase);
    return readJenaModel(model, xmlBase, map);
  }
*/

  /**
   * Read document from given URL
   */
  public Document read(String url) throws Exception{
   // url = url.toLowerCase();
    if(blacklist!=null && blacklist.contains(url)){
        System.out.println("HAWK: url is in blacklist!");
        return null;
    }

    if(proxy!=null){
    	String phy = proxy.getURI(url);		// Retrieve the physical URI of the ontology
    	if(!phy.startsWith("http")&&!phy.startsWith("file")){
    		InputStream in = new BufferedInputStream(new FileInputStream(phy));
    		return this.read(in);
    	}else
        url = phy;
    }

    if(url.equalsIgnoreCase("//www.w3.org/2002/07/owl")||
        url.equalsIgnoreCase("http://www.w3.org/2000/01/rdf-schema")||
        url.equalsIgnoreCase("http://www.w3.org/1999/02/22-rdf-syntax-ns")||
        url.equalsIgnoreCase("http://xmlns.com/foaf/0.1/")){
        return null;
    }
//    HashMap map = null;

//    System.out.println("URL is "+url);
    Document d = smodel.getDocument(url);
    if(d!=null)
      return d;

    String xmlBase = "";
    InputStream ins = null;
    try{
      PreParser pp = new PreParser(url);
//      map = pp.getNamespaces();
      xmlBase = pp.getXMLBase();
      ins = pp.getInputStream();
      if(ins == null)
          return null;
    }catch(Exception e){
      e.printStackTrace();
    }

    if(xmlBase==null||xmlBase.equals(""))
      xmlBase = url;
    d = smodel.getDocument(xmlBase);
    if(d!=null)
      return d;
    JenaReader jenaReader = new JenaReader();
    Model model = ModelFactory.createDefaultModel();
    if(ins==null)
        return null;
    jenaReader.read(model,ins,xmlBase);
    ins.close();
    return readJenaModel(model, xmlBase);
  }


  /**
   * Read an input in OWL format to the model and return the document object
   */
  public Document read(InputStream in, String xmlBase) throws Exception{
    //long t0= System.currentTimeMillis(); //temp
    ByteArrayInputStream bin=null, bin2=null;
    ByteArrayOutputStream bout = new ByteArrayOutputStream();
 //   try{
      byte[] buf = new byte[10240];
      int len = 0;
      while((len = in.read(buf))>0){
        bout.write(buf,0,len);
      }
      //System.out.println("inputstream begin: \n"+bout.toString()+"\n!!!the end!!!");
      //len = bout.size();
      //buf = new byte[len];
      buf = bout.toByteArray();
      bin = new ByteArrayInputStream(buf);
      bin2 = new ByteArrayInputStream(buf);

//    }catch (Exception e) {
//        e.printStackTrace();
//    }
//try{
    PreParser pp = new PreParser(bin);
    //HashMap map = pp.getNamespaces();
    if(xmlBase=="")
      xmlBase=pp.getXMLBase();
  //}catch(Exception e){};
    Document d = smodel.getDocument(xmlBase);
    if(d!=null)
      return d;

    JenaReader jenaReader = new JenaReader();
   //  long t1= System.currentTimeMillis(); //temp
    Model model = ModelFactory.createDefaultModel();
        //long t2 = System.currentTimeMillis(); //temp
    jenaReader.read(model, bin2, xmlBase);
    //long t3 = System.currentTimeMillis() - t2; //temp
    //System.out.println("jena time: "+t3);
//System.out.println(t1-t0);
  //  System.out.println(t2-t1);
 // System.out.println(xmlBase);
    return readJenaModel(model, xmlBase);
  }

  /**
   * Process the document from the triple-based model in Jena.
   * @param jenaReader
   * @param model
   * @param xmlBase
   * @return
   */
  private Document readJenaModel(Model model,
                                 String xmlBase) throws Exception{
    Resource itself = model.getResource(xmlBase);
    Document doc = null;
    edu.lehigh.swat.hawk.core.Ontology ont = null;

    ResIterator ti = model.listSubjectsWithProperty(RDF.type,OWL.Ontology);
    while(ti.hasNext())
    {
      Resource r = ti.nextResource();
      if(r!=null&&r.getURI()!=null&&!r.getURI().equals("")){
          xmlBase = r.getURI().replace('#',' ').trim();
          Document d = smodel.getDocument(xmlBase);
          if(d!=null)
            return d;
      }
    } // z. pan on 6/29/05: some ontologies don't have base but give info in header.



/*    if(itself!=null&&model.contains(itself, RDF.type, OWL.Ontology)){
      ont = smodel.createOntology();
      doc = ont;
    }
    else{
      doc = smodel.createDocument();
    }
*/  //better way to identify ontology/document
    boolean isOnto = false;
    ResIterator rirc = model.listSubjectsWithProperty(RDF.type, RDFS.Class);
    ResIterator ric = model.listSubjectsWithProperty(RDF.type, OWL.Class);
    ResIterator riop = model.listSubjectsWithProperty(RDF.type,OWL.ObjectProperty);
    ResIterator ridp = model.listSubjectsWithProperty(RDF.type, OWL.DatatypeProperty);
    ResIterator rip = model.listSubjectsWithProperty(RDF.type, RDF.Property);
    ResIterator rion = model.listSubjectsWithProperty(RDF.type, OWL.Ontology);
//    if(rirc==null||ric==null||riop==null||ridp==null||rip==null||rion==null)
//      isOnto = false;
    if(ric.hasNext()||riop.hasNext()||ridp.hasNext()||rip.hasNext()
        ||rirc.hasNext()) //zhp2: loose up requirement on OWL.Ontology
      isOnto = true;

    int tcount = model.listStatements().toList().size();
    int icount=0;
    if(model.listObjectsOfProperty(OWL.imports)!=null)
        icount=model.listObjectsOfProperty(OWL.imports).toList().size();
    if(rion!=null&&(tcount-icount)<3)
        isOnto=true;

//temproray line:
    if(xmlBase.contains("Source"))
        isOnto = false;
//end temproray.

    if(isOnto){
      //System.out.println("HAWK: this is an ontology!");
      ont =  smodel.createOntology();
      doc = ont;
    }
    else{
      doc = smodel.createDocument();
    }

    DocId docid = new DocId(xmlBase);
    doc.setID(docid);

    //int tcount = model.listStatements().toList().size();
    doc.setSize(tcount);
//System.out.println("add document: "+docid.getId());
    smodel.addDocument(doc);
//    HashMap ns = jenaReader.getNameSpaces();
    Map map = model.getNsPrefixMap();
    Map nsMap = new HashMap();
    Iterator it = map.keySet().iterator();
    if(it!=null){
        while(it.hasNext()){
            String prefix = (String) it.next();
            String ns = (String)map.get(prefix);
            nsMap.put(ns,prefix);
        }
    }
    doc.setNameSpaces(nsMap);


    //
    // z.pan: RDFs use NameSpaces, not imports
    /*
    Iterator it = map.keySet().iterator();
    if(it!=null){
        while(it.hasNext()){
    System.out.println((String)it.next());
                    String ns = (String) it.next();
            ns.replace('#',' ').trim();
            if(smodel.getDocument(ns)==null)
                this.read(ns);

       }
    }
    */ //end additions for RDF NameSpace


    doc.setXMLBase(xmlBase);
//  System.out.println(xmlBase);
    NodeIterator ni = model.listObjectsOfProperty(itself,RDFS.comment);
    while(ni.hasNext()){
        Literal comment = (Literal)ni.nextNode();
        doc.addDescription(comment.getLexicalForm());
    }

    ni = model.listObjectsOfProperty(itself,RDFS.label);
    if(ni.hasNext()){
        Literal label = (Literal)ni.nextNode();
        doc.setLabel(label.getLexicalForm());
    }

    ni = model.listObjectsOfProperty(OWL.imports);
    while(ni.hasNext()){
      Resource l = (Resource)ni.nextNode();
      DocId dId = new DocId(l.getURI());
      if(this.getLoadImports()){
    	  Document idoc = this.read(dId.getId());
          if(idoc!=null)
            dId = idoc.getID();
      }
      doc.addImports(dId);
      if(ont!=null)
          smodel.pushImports(ont);
    }

    if(ont!=null){

      ni = model.listObjectsOfProperty(itself,OWL.versionInfo);
      if(ni.hasNext()){
        Literal ver = (Literal)ni.nextNode();
        ont.setVersion(ver.getLexicalForm());
      }

      ni = model.listObjectsOfProperty(OWL.priorVersion);
      while(ni.hasNext()){
        Resource l = (Resource)ni.nextNode();
        DocId dId = new DocId(l.getURI());
        ont.addPriorOnts(dId);
      }

      ni = model.listObjectsOfProperty(OWL.backwardCompatibleWith);
      while(ni.hasNext()){
        Resource l = (Resource)ni.nextNode();
        DocId dId = new DocId(l.getURI());
        this.read(l.getURI());
        ont.addCompatiblePriorOnts(dId);
      }

      //OWL.incompatibleWith??

      readOntology(ont, model);
      ont.flush();
      return ont;

   }// end if isOntology

   else{
     readIndividuals(doc,model);
     doc.flush();
     return doc;
   }
  }

  /**
   * Parse and construct an ontology.
   */
  protected void readOntology(Ontology ont, Model model){


    long t0 = System.currentTimeMillis();//test
    readClasses(ont, model);
      long t1 = System.currentTimeMillis()-t0; //test
//      System.out.println("class: "+t1); //test

      t0 = System.currentTimeMillis();//test
    readProperties(ont, model);
       t1 = System.currentTimeMillis()-t0; //test
//      System.out.println("property: "+t1); //test

       //Read annotation Properties, added by yang
       readAnnotations(ont, model);

       t0 = System.currentTimeMillis();//test
       readIndividuals(ont, model);
       t1 = System.currentTimeMillis()-t0; //test
//     System.out.println("individual: "+t1); //test


      t0 = System.currentTimeMillis();//test
    readRestrictions(ont,model);
       t1 = System.currentTimeMillis()-t0; //test
 //   System.out.println("restriction: "+t1); //test

      t0 = System.currentTimeMillis();//test
    readAxioms(ont, model);
     t1 = System.currentTimeMillis()-t0; //test
 //   System.out.println("axiom: "+t1); //test

//    descendants.pop();
//zhp2 commented out on 7/3/08
     mergeEquivalents(ont);
     resolveAnonymousRestrictions(ont);


  }

  /**
   * Construct classes and add to the ontology.
   */
  private void readClasses(Ontology ont, Model model){

    ResIterator ri = model.listSubjectsWithProperty(RDF.type,OWL.Class);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      addClass(ont,r,"",getComment(r,model),getLabel(r,model));
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.DeprecatedClass);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      addClass(ont,r,"deprecated",getComment(r,model),getLabel(r,model));
    }

    ri = model.listSubjectsWithProperty(RDF.type,RDFS.Class);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      addClass(ont,r,"",getComment(r,model),getLabel(r,model));
    }

  }

  /**
   * Construct properties and add to the ontology.
   */
  private void readProperties(Ontology ont, Model model){   //properties
//    long t0 = System.currentTimeMillis();//test

	ResIterator ri = model.listSubjectsWithProperty(RDF.type,OWL.DatatypeProperty);

    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,new Boolean(true));
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.ObjectProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,new Boolean(false));
    }

    ri = model.listSubjectsWithProperty(RDF.type,RDF.Property);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,null);
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.AnnotationProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,null);
      if(op!= null){
        op.setAnnotation(true);
      }
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.DeprecatedProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,null);
      if(op != null){
        op.setDeprecated(true);
      }
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.TransitiveProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,new Boolean(false));
      if(op != null){
        op.setTransitive(true);
      }
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.SymmetricProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,new Boolean(false));
      if(op != null){
        op.setSymmetric(true);
      }
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.FunctionalProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,new Boolean(false));
      if(op != null){
        op.setFunctional(true);
      }
    }

    ri = model.listSubjectsWithProperty(RDF.type,OWL.InverseFunctionalProperty);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      OntProperty op = addProperty(ont,r,model,new Boolean(false));
      if(op != null){
        op.setInverseFuncional(true);
      }
    }
  }

  private void readAnnotations(Ontology ont, Model model)
  {
      //read localnames of all AnnotationProperties including all imported into array
      ArrayList<String> annPropArray = new ArrayList<String>();
//	  ArrayList<Ontology> arrayOnt = new ArrayList<Ontology>();
//	  arrayOnt.add(ont);
//	  for(int i=0; i<arrayOnt.size(); i++) {
      OntProperty[] ontProps = ont.getAllPlusImportedProperties();
      for (int i=0; i<ontProps.length; i++) {
          if (ontProps[i].isAnnotation()) {
//			  System.out.println(ontProps[i].getID().getLocalName());
              annPropArray.add(ontProps[i].getID().getLocalName());
          }
      }
      //traverse all classes' statement to check if there are usages of AnnotationProperties
      ResIterator classRI = model.listSubjectsWithProperty(RDF.type,OWL.Class);
      while(classRI.hasNext()){
          Resource r = classRI.nextResource();
          StmtIterator itstmt = r.listProperties();
          while(itstmt.hasNext())
          {
              Statement ss = itstmt.nextStatement();
              String strPropValue = "";
              for (int i=0; i<annPropArray.size(); i++)
                  if (ss.getPredicate().getLocalName().equals(annPropArray.get(i))) {
                      if (ss.getObject().isLiteral()) {
                          strPropValue = ss.getObject().toString();
                          if(strPropValue.indexOf("^^")>0) //zhp2 on 11/24/08
                              strPropValue = strPropValue.substring(0,
                                      strPropValue.indexOf("^^"));
                          if(strPropValue.length()>0)
                             ont.addAnnotations(resourceToId(r),resourceToId(ss.getPredicate()),strPropValue);
                      }
                  }
          }
      }
      ResIterator objPropRI = model.listSubjectsWithProperty(RDF.type,OWL.ObjectProperty);
      while(objPropRI.hasNext()){
          Resource r = objPropRI.nextResource();
          StmtIterator itstmt = r.listProperties();
          while(itstmt.hasNext())
          {
              Statement ss = itstmt.nextStatement();
              String strPropValue = "";
              for (int i=0; i<annPropArray.size(); i++)
                  if (ss.getPredicate().getLocalName().equals(annPropArray.get(i))) {
                      if (ss.getObject().isLiteral()) {
                          strPropValue = ss.getObject().toString();
                          if(strPropValue.indexOf("^^")>0) //zhp2 on 11/24/08
                              strPropValue = strPropValue.substring(0,
                                      strPropValue.indexOf("^^"));
                          if(strPropValue.length()>0)
                             ont.addAnnotations(resourceToId(r), resourceToId(ss.getPredicate()),strPropValue);
                      }
                  }
          }
      }
      ResIterator dataPropRI = model.listSubjectsWithProperty(RDF.type,OWL.DatatypeProperty);
      while(dataPropRI.hasNext()){
          Resource r = dataPropRI.nextResource();
          StmtIterator itstmt = r.listProperties();
          while(itstmt.hasNext())
          {
              Statement ss = itstmt.nextStatement();
              String strPropValue = "";
              for (int i=0; i<annPropArray.size(); i++)
                  if (ss.getPredicate().getLocalName().equals(annPropArray.get(i))) {
                      if (ss.getObject().isLiteral()) {
                          strPropValue = ss.getObject().toString();
                          if(strPropValue.indexOf("^^")>0)  //zhp2 on 11/24/08
                              strPropValue = strPropValue.substring(0,
                                      strPropValue.indexOf("^^"));
                          if(strPropValue.length()>0)
                              ont.addAnnotations(resourceToId(r),resourceToId(ss.getPredicate()),strPropValue);
                      }
                  }
          }
      }
  }


  /**
   * Construct individuals and add to the ontology.
   */
  private void readIndividuals(Document doc, Model model){ //individuals
    StmtIterator si = model.listStatements(new SelectorImpl(null,RDF.type,(RDFNode)null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();

      Resource obj = null;
      if(stmt.getObject().isResource())
    	  obj = (Resource)stmt.getObject();
      ObjectId oid = resourceToId(obj);
      if(oid==null)
        continue;
      if(!isVocabularyResource(obj)){
        Resource sub = (Resource)stmt.getSubject();
        ObjectId iid = resourceToId(sub);
        Individual indiv = doc.getIndividual(iid);
   //     System.out.println(iid.getId());//test
        if(indiv==null){
          indiv = doc.createIndividual();
          indiv.setID(iid);
          indiv.addDescription(getComment(sub,model));
          indiv.setLabel(getLabel(sub,model));
          indiv.addSourceClass(oid);
          doc.addIndividual(indiv);
        }else{
          indiv.addSourceClass(oid);
          continue;
        }
        StmtIterator isi = model.listStatements(new SelectorImpl(sub,null,
            (RDFNode) null));
        while(isi.hasNext()){
          Statement st = isi.nextStatement();
          Property p = st.getPredicate();
          ObjectId pid = resourceToId(p);
          if(pid==null||p.equals(RDF.type)||p.equals(RDFS.comment)) //||p.equals(RDFS.label))
            continue;

          RDFNode vn = st.getObject();
          if(vn instanceof Resource){
            Resource vr = (Resource)vn;
            ObjectId vid = resourceToId(vr);
            indiv.addProperty(pid,vid);
          }
          else if(vn instanceof Literal)
          {
            Literal vl = (Literal)vn;
            String s = vl.getLexicalForm();
            String dtype;
            if(vl.getDatatype()!=null)
              dtype = vl.getDatatype().getURI();
            else
              dtype = "";
            DatatypeValue dv = new DatatypeValue(s,dtype);
            indiv.addProperty(pid,dv);
          }
        }
      }
    }

// handle untyped individuals and their properties
    si = model.listStatements();
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Property p = stmt.getPredicate();
      ObjectId pid = resourceToId(p);
      if(pid==null||p.equals(RDF.type)||p.equals(RDFS.comment)||p.equals(RDFS.label))
            continue;
       if(!isVocabularyProperty(p)){
           Resource sub = (Resource) stmt.getSubject();
           StmtIterator tsi = model.listStatements(sub, RDF.type, (RDFNode)null);  //zhp2 on 2/26/09, added 3 line from here.
           if(tsi.hasNext())
               continue;
           ObjectId sid = resourceToId(sub);
           if (sid == null)
               continue;
           Individual indiv = doc.getIndividual(sid);

           ObjectId OWLThing = new ObjectId("http://www.w3.org/2002/07/owl#Thing");
           if(doc.getIndividual(sid)!=null&&!doc.getIndividual(sid).getSourceClasses()[0].equals(OWLThing))
               continue;

           else if(indiv==null){
               indiv = doc.createIndividual();
               indiv.setID(sid);
               indiv.addDescription(getComment(sub,model));
               indiv.setLabel(getLabel(sub,model));
               indiv.addSourceClass(OWLThing);
               doc.addIndividual(indiv);
           }

           RDFNode vn = stmt.getObject();
           if(vn instanceof Resource){
             Resource vr = (Resource)vn;
             ObjectId vid = resourceToId(vr);
             indiv.addProperty(pid,vid);
           }
           else if(vn instanceof Literal)
           {
             Literal vl = (Literal)vn;
             String s = vl.getLexicalForm();
             String dtype;
             if(vl.getDatatype()!=null)
               dtype = vl.getDatatype().getURI();
             else
               dtype = "";
             DatatypeValue dv = new DatatypeValue(s,dtype);
             indiv.addProperty(pid,dv);
          }
       }
  }
/*
    ResIterator ti = model.listSubjects();
    while(ti.hasNext()){
      Resource r = ti.nextResource();
      NodeIterator obji = model.listObjectsOfProperty(r,RDF.type);
      if(!obji.hasNext()){
        ObjectId oid = resourceToId(r);
      }
    }
*/

    //sameAs
    si = model.listStatements(new SelectorImpl(null,OWL.sameAs,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
/*      HashSet set = new HashSet();
      set.add(sid);
      set.add(oid);
      ont.addSameIndividualsSet(set);
*/    Individual i1 = doc.getIndividual(sid);
      Individual i2 = doc.getIndividual(oid);
      if(i1!=null&&i2!=null){
        i1.addSameAs(oid);
        i2.addSameAs(sid);
        Collection c1 = i1.getSameSet();
        Collection c2 = i2.getSameSet();
        if(!c1.equals(c2)){
          Collection c = combineSets(c1,c2);
          doc.deleteSameIndividualsSet(c1);
          doc.deleteSameIndividualsSet(c2);
          doc.addSameIndividualsSet(c);
          i1.setSameSet(c);
          i2.setSameSet(c);
        }
      }else{
          Vector set = new Vector();
          set.add(oid);
          set.add(sid);
          doc.addSameIndividualsSet(set);
      }
    }
//differentFrom
    si = model.listStatements(new SelectorImpl(null,OWL.differentFrom,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      HashSet set = new HashSet();
      set.add(sid);
      set.add(oid);
      doc.addDistinctIndividualsSet(set);
      Individual i1 = doc.getIndividual(sid);
      Individual i2 = doc.getIndividual(oid);
      if(i1!=null&&i2!=null){
        i1.addDifferentFrom(oid);
        i2.addDifferentFrom(sid);
      }
    }

    //AllDifferent
    ResIterator ri = model.listSubjectsWithProperty(RDF.type,OWL.AllDifferent);
    while(ri.hasNext()){
      Resource ad = ri.nextResource();
      NodeIterator ni = model.listObjectsOfProperty(ad, OWL.distinctMembers);
      Resource members = (Resource) firstNode(ni);
      if(!model.contains(members,RDF.type,RDF.List))
        continue;
      Collection c = unpackList(model,members);
      Iterator it = c.iterator();
      if(!c.isEmpty()){
        Collection cc = new ArrayList();
        while(it.hasNext()){
          Resource re = (Resource)it.next();
          ObjectId oid = resourceToId(re);
          if(oid==null)
            continue;
          cc.add(oid);
          if(doc.getIndividual(oid)==null){
            Individual indiv = doc.createIndividual();
            indiv.setID(oid);
            NodeIterator ndi= model.listObjectsOfProperty(re,RDF.type);
            Resource rsc = (Resource)firstNode(ndi);
            ObjectId scid = resourceToId(rsc);
            indiv.addSourceClass(scid);
            doc.addIndividual(indiv);
          }
        }
        doc.addDistinctIndividualsSet(cc);
      }
    }

  }

  /**
   * Construct restrictions and add to the ontology.
   */
  private void readRestrictions(Ontology ont, Model model){//Restriction
    long t0 = System.currentTimeMillis();//test
    ResIterator ri = model.listSubjectsWithProperty(RDF.type,OWL.Restriction);
    while(ri.hasNext()){
      Resource r = ri.nextResource();
      ObjectId rid = resourceToId(r);
      if(rid==null)
        continue;
      PropertyRestriction res = ont.getRestriction(rid);
      if(res==null){
        res = ont.createRestriction();
        res.setID(rid);
        ont.addRestriction(res);
      }
      NodeIterator ni = model.listObjectsOfProperty(r,OWL.onProperty);
      if(!ni.hasNext())
        continue;
      Resource pr = (Resource)ni.nextNode();
      ObjectId opid = resourceToId(pr);
      res.setOnProperty(opid);

      ni = model.listObjectsOfProperty(r,OWL.allValuesFrom);
      if(ni.hasNext()){
        pr = (Resource)ni.nextNode();
        opid = resourceToId(pr);
        res.setAllValuesFrom(opid);
      }

      ni = model.listObjectsOfProperty(r,OWL.someValuesFrom);
      if(ni.hasNext()){
        pr = (Resource)ni.nextNode();
        opid = resourceToId(pr);
        res.setSomeValuesFrom(opid);
      }

      ni = model.listObjectsOfProperty(r,OWL.hasValue);
      if(ni.hasNext()){
        RDFNode pd = ni.nextNode();
        if(pd.canAs(resource)){
          pr = (Resource)pd;
          opid = resourceToId(pr);
          //System.out.println("Has Value "+opid.getId());
          //if(ont.getIndividual(opid)!=null)			// Problem traced!!!
          if(ont.getOntObject(opid) == null)
        	  res.setHasValue(opid);
          else{
        	  System.out.println("The resource - "+opid.getId()+" is a class/property and " +
        	  		"hence cannot be the value of hasValue property!!");
        	  ont.deleteRestriction(res.getID());
          }
        }
        else{
          Literal pl = (Literal)pd;
          if(pl.isLiteral()){
            String s = pl.getLexicalForm();
            DatatypeValue dv = new DatatypeValue(s,pl.getDatatypeURI());
            res.setHasValue(dv);
          }
        }
      }

      ni = model.listObjectsOfProperty(r,OWL.cardinality);
      Literal cl;
      if(ni.hasNext()){
        cl = (Literal)ni.nextNode();
        String ls = cl.getLexicalForm();
        int card = Integer.parseInt(ls);
        res.setMinCardinality(card);
        res.setMaxCardinality(card);
      }

      ni = model.listObjectsOfProperty(r,OWL.maxCardinality);
      if(ni.hasNext()){
        cl = (Literal)ni.nextNode();
        String ls = cl.getLexicalForm();
        int card = Integer.parseInt(ls);
        res.setMaxCardinality(card);
      }

      ni = model.listObjectsOfProperty(r,OWL.minCardinality);
      if(ni.hasNext()){
        cl = (Literal)ni.nextNode();
        String ls = cl.getLexicalForm();
        int card = Integer.parseInt(ls);
        res.setMinCardinality(card);
      }
    }
  }

  /**
   * Construct axioms and add to the ontology.
   */
  private void readAxioms(Ontology ont, Model model){  //axioms
    long t0 = System.currentTimeMillis();//test
    StmtIterator si = model.listStatements(new SelectorImpl(null,RDFS.subClassOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      ObjectId sid = resourceToId(sub);

      Resource obj = (Resource)stmt.getObject();
      ObjectId oid = resourceToId(obj);

      if(sid==null || oid==null)
        continue;
      /*ont.addAssertedClass(sid);
      ont.addAssertedClass(oid);

      OntClass subc = ont.getOntClass(sid);
      if(subc==null)                        //zhp2: tentative changes
          subc = ont.getRestriction(sid);   //
      OntClass superc = ont.getOntClass(oid);
      if(superc==null)
          superc = ont.getRestriction(oid);
      if(subc!=null){
        subc.addSuperClass(oid);*/
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);

        a.setLeftAxiomNode(subNode);
        a.setRightAxiomNode(objNode);
        ont.addAxiom(a);
        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF), a);
      /*}
      if(superc!=null)
        superc.addSubClass(sid);*/
    }

    si = model.listStatements(new SelectorImpl(null, RDFS.domain, (RDFNode) null));		// Read domain and range axioms as well
    while(si.hasNext()){																// Ameet, Oct 2 2008
          Statement stmt = si.nextStatement();
          Resource sub = (Resource)stmt.getSubject();
          Resource obj = (Resource)stmt.getObject();
          ObjectId sid = resourceToId(sub);
          ObjectId oid = resourceToId(obj);
          if(sid==null || oid==null)
            continue;

          Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.DOMAIN);
          AxiomNode subNode = new AxiomNode();
          subNode.setAtomicNode(sid);

          AxiomNode objNode = new AxiomNode();
          objNode.setAtomicNode(oid);

          a.setLeftAxiomNode(subNode);
          a.setRightAxiomNode(objNode);
          ont.addAxiom(a);

          //zhp2 on 8/8/09: need to be cosolidated with axioms.
          OntProperty lp = ont.getProperty(sid);
          if(lp==null)
              continue;
          lp.setDomain(oid);
    }

    si = model.listStatements(new SelectorImpl(null, RDFS.range, (RDFNode) null));
    while(si.hasNext()){
          Statement stmt = si.nextStatement();
          Resource sub = (Resource)stmt.getSubject();
          Resource obj = (Resource)stmt.getObject();
          ObjectId sid = resourceToId(sub);
          ObjectId oid = resourceToId(obj);
          if(sid==null || oid==null)
            continue;

          Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.RANGE);
          AxiomNode subNode = new AxiomNode();
          subNode.setAtomicNode(sid);

          AxiomNode objNode = new AxiomNode();
          objNode.setAtomicNode(oid);

          a.setLeftAxiomNode(subNode);
          a.setRightAxiomNode(objNode);
          ont.addAxiom(a);

          //zhp2 on 8/8/09: need to be cosolidated with axioms.
          OntProperty lp = ont.getProperty(sid);
          if(lp==null)
              continue;
          lp.setRange(oid);
    }

    si = model.listStatements(new SelectorImpl(null,RDFS.subPropertyOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      ont.addAssertedProperty(sid);
      ont.addAssertedProperty(oid);
     /*
      OntProperty subc = ont.getProperty(sid);
      OntProperty superc = ont.getProperty(oid);
      if(subc!=null){
        subc.addSuperProperty(oid);*/
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.SUBPROPERTYOF);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);

        a.setLeftAxiomNode(subNode);
        a.setRightAxiomNode(objNode);
        ont.addAxiom(a);
        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.SUBPROPERTYOF), a);
      /*}
      if(superc!=null)
        superc.addSubProperty(sid);*/
    }

    si = model.listStatements(new SelectorImpl(null,OWL.equivalentClass,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      /*ont.addAssertedClass(sid);
      ont.addAssertedClass(oid);

      OntClass subc = ont.getOntClass(sid);
      OntClass superc = ont.getOntClass(oid);

      if(subc!=null){
        subc.addEquivalentClass(oid);*/
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);

        a.setLeftAxiomNode(subNode);
        a.setRightAxiomNode(objNode);
        ont.addAxiom(a);
        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS), a);
      /*}
      if(superc!=null)
        superc.addEquivalentClass(sid);*/
    }

    si = model.listStatements(new SelectorImpl(null,OWL.equivalentProperty,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      ont.addAssertedProperty(sid);
      ont.addAssertedProperty(oid);
      /*
      OntProperty subc = ont.getProperty(sid);
      OntProperty superc = ont.getProperty(oid);
      if(subc!=null){
        subc.addEquivalentProperty(oid);*/
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTPROPERTY);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);

        a.setLeftAxiomNode(subNode);
        a.setRightAxiomNode(objNode);
        ont.addAxiom(a);
        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTPROPERTY), a);
      /*}
      if(superc!=null)
        superc.addEquivalentProperty(sid);*/
    }

    si = model.listStatements(new SelectorImpl(null,OWL.inverseOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      ont.addAssertedProperty(sid);
      ont.addAssertedProperty(oid);
     /*
      OntProperty subc = ont.getProperty(sid);
      OntProperty superc = ont.getProperty(oid);
      if(subc!=null){
        subc.addInverseOf(oid);*/
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTPROPERTY);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);
        a.setLeftAxiomNode(subNode);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);
        AxiomNode compNode = new AxiomNode();
        compNode.setJoinType(edu.lehigh.swat.hawk.core.Ontology.INVERSEOF);
        compNode.addChild(objNode);
        a.setRightAxiomNode(compNode);
        ont.addAxiom(a);

        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.INVERSEOF), a);

        //zhp2 on 8/8/09: need to be cosolidated with axioms.
        OntProperty lp = ont.getProperty(sid);
        OntProperty rp = ont.getProperty(oid);
        if(lp==null||rp==null)
            continue;
        if(lp.getDomain()==null&&rp.getRange()!=null)
            lp.setDomain(rp.getRange());
        if(rp.getDomain()==null&&lp.getRange()!=null)
            rp.setDomain(lp.getRange());
        if(lp.getRange()==null&&rp.getDomain()!=null)
            lp.setRange(rp.getDomain());
        if(rp.getRange()==null&&lp.getDomain()!=null)
                rp.setRange(lp.getDomain());

      /*}
      else{
          OntProperty prop = ont.createProperty();
          prop.setID(sid);
          prop.setDatatype(false);
          ont.addProperty(prop);
      }
      if(superc!=null)
        superc.addInverseOf(sid);
      else{
          OntProperty prop = ont.createProperty();
          prop.setID(oid);
          prop.setDatatype(false);
          ont.addProperty(prop);
      }*/
    }

    si = model.listStatements(new SelectorImpl(null,OWL.disjointWith,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      /*ont.addAssertedClass(sid);
      ont.addAssertedClass(oid);

      OntClass subc = ont.getOntClass(sid);
      OntClass superc = ont.getOntClass(oid);
      if(subc!=null){
        subc.addDisjointWith(oid);*/
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.DISJOINTWITH);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);

        a.setLeftAxiomNode(subNode);
        a.setRightAxiomNode(objNode);
        ont.addAxiom(a);
        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.DISJOINTWITH), a);
      /*}
      if(superc!=null)
        superc.addDisjointWith(sid);*/
    }

    si = model.listStatements(new SelectorImpl(null,OWL.complementOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      ObjectId oid = resourceToId(obj);
      if(sid==null || oid==null)
        continue;
      /*ont.addAssertedClass(sid);
      ont.addAssertedClass(oid);

      OntClass subc = ont.getOntClass(sid);
      OntClass superc = ont.getOntClass(oid);
      if(subc!=null){
        subc.addComplementOf(oid);*/
//zhp2 on7/7/08
      Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS);

      AxiomNode subNode = new AxiomNode();
      subNode.setAtomicNode(sid);
      a.setLeftAxiomNode(subNode);

      AxiomNode objNode = new AxiomNode();
      objNode.setAtomicNode(oid);
      AxiomNode compNode = new AxiomNode();
      compNode.setJoinType(edu.lehigh.swat.hawk.core.Ontology.COMPLEMENTOF);
      compNode.addChild(objNode);
      a.setRightAxiomNode(compNode);
//end zhp2, following was commented out by zhp2 on 7/7/08
/*
        Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.COMPLEMENTOF);

        AxiomNode subNode = new AxiomNode();
        subNode.setAtomicNode(sid);

        AxiomNode objNode = new AxiomNode();
        objNode.setAtomicNode(oid);

        a.setLeftAxiomNode(subNode);
        a.setRightAxiomNode(objNode);
*/
        ont.addAxiom(a);
        ont.addAxiomBySubject(sid, a);
        ont.addAxiomByObject(oid, a);
        ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.COMPLEMENTOF), a);
      //}
      //is restriction already be anon class in jena?
    }

    si = model.listStatements(new SelectorImpl(null,OWL.intersectionOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      if(sid==null)
        continue;
      //ont.addAssertedClass(sid);

//      if(!model.contains(obj,RDF.type,RDF.List))
//        continue;
     /*OntClass subc = ont.getOntClass(sid);
     if(subc==null){
         subc = ont.createClass();
         subc.setID(sid);
         ont.addOntClass(subc);
     }*/

     // Ameet Oct 22, 2007
     /*if(subc.getIntersectionOfs()!=null && (subc.getIntersectionOfs().length)>0){ // zpan: more than one intersects treated as equivalent class
         ObjectId oid = resourceToId(obj);
         ont.addAssertedClass(oid);   //zpan 04/08/07
         subc.addEquivalentClass(oid);
         continue;
     }*/

      Collection c = unpackList(model,obj);
      Iterator it = c.iterator();

      if(!c.isEmpty()){
    	  Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS);

          AxiomNode subNode = new AxiomNode();
          subNode.setAtomicNode(sid);
          a.setLeftAxiomNode(subNode);

          AxiomNode objNode = new AxiomNode();
          objNode.setJoinType(edu.lehigh.swat.hawk.core.Ontology.INTERSECTIONOF);

          while(it.hasNext()){
            Resource re = (Resource)it.next();
            ObjectId oid = resourceToId(re);
            if(oid==null)
              continue;
            /*ont.addAssertedClass(oid);  //zpan 04/08/07
            subc.addIntersectionOf(oid);*/
            AxiomNode child = new AxiomNode();
            child.setAtomicNode(oid);
            objNode.addChild(child);
          }
          a.setRightAxiomNode(objNode);
          ont.addAxiomBySubject(sid, a);
          ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS), a);
          ont.addAxiom(a);
      }

    }

    si = model.listStatements(new SelectorImpl(null,OWL.unionOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      if(sid==null)
        continue;
      //ont.addAssertedClass(sid);

//      if(!model.contains(obj,RDF.type,RDF.List))
//        continue;
      /*OntClass subc = ont.getOntClass(sid);
      if(subc==null){
        subc = ont.createClass();
        subc.setID(sid);
        ont.addOntClass(subc);
      }*/

      // Ameet Oct 22, 2007
    /*if(subc.getUnionOfs()!=null && (subc.getUnionOfs().length)>0){ // zpan: more than one union treated as equivalent class
          ObjectId oid = resourceToId(obj);
          ont.addAssertedClass(oid);  //zpan 04/08/07
          subc.addEquivalentClass(oid);
          continue;
     }*/

      Collection c = unpackList(model,obj);
      Iterator it = c.iterator();
      if(!c.isEmpty()){
    	  Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS);
/*
if(ont.getRestriction(sid)!=null){
     PropertyRestriction res = ont.getRestriction(sid);
     sid = res.getSomeValuesFrom();
}
*/
          AxiomNode subNode = new AxiomNode();
          subNode.setAtomicNode(sid);
          a.setLeftAxiomNode(subNode);

          AxiomNode objNode = new AxiomNode();
          objNode.setJoinType(edu.lehigh.swat.hawk.core.Ontology.UNIONOF);

          while(it.hasNext()){
            Resource re = (Resource)it.next();
            ObjectId oid = resourceToId(re);
            if(oid==null)
              continue;
            /*ont.addAssertedClass(oid);  //zpan 04/08/07
            subc.addUnionOf(oid);*/
            AxiomNode child = new AxiomNode();
            child.setAtomicNode(oid);
            objNode.addChild(child);
          }
          a.setRightAxiomNode(objNode);
          ont.addAxiomBySubject(sid, a);
          ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS), a);
          ont.addAxiom(a);
     }
  }

    si = model.listStatements(new SelectorImpl(null,OWL.oneOf,
        (RDFNode) null));
    while(si.hasNext()){
      Statement stmt = si.nextStatement();
      Resource sub = (Resource)stmt.getSubject();
      Resource obj = (Resource)stmt.getObject();
      ObjectId sid = resourceToId(sub);
      if(sid==null)
        continue;
      //ont.addAssertedClass(sid);

//      if(!model.contains(obj,RDF.type,RDF.List))
//        continue;
      Collection c = unpackList(model,obj);
      if(c==null)
        continue;
      Iterator it = c.iterator();
      /*OntClass subc = ont.getOntClass(sid);
      if(subc==null){
        PropertyRestriction res = ont.createRestriction();
        res.setID(sid);
        ont.addRestriction(res);
        subc = res;
      }*/

      if(!c.isEmpty()){
    	  Axiom a = new Axiom(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS);

          AxiomNode subNode = new AxiomNode();
          subNode.setAtomicNode(sid);
          a.setLeftAxiomNode(subNode);

          AxiomNode objNode = new AxiomNode();
          objNode.setJoinType(edu.lehigh.swat.hawk.core.Ontology.ONEOF);

          while(it.hasNext()){
            Resource re = (Resource)it.next();
            ObjectId oid = resourceToId(re);
            if(oid==null)
              continue;
            //subc.addOneOf(oid);
            AxiomNode child = new AxiomNode();
            child.setAtomicNode(oid);
            objNode.addChild(child);
          }
          a.setRightAxiomNode(objNode);
          ont.addAxiomBySubject(sid, a);
          ont.addAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS), a);
          ont.addAxiom(a);
      }
    }
  }
  /** This method wherever possible eliminates non top level axioms containing anonymous nodes either
   * in the subject or the object position.
   * @param ont
   */
  protected void mergeEquivalents(Ontology ont){

      Vector<Axiom> classAxioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS));
      Vector<Axiom> propAxioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTPROPERTY));
      Vector<Axiom> equivalentAxioms = new Vector<Axiom>();

      if(classAxioms != null && !equivalentAxioms.containsAll(classAxioms))
          equivalentAxioms.addAll(classAxioms);

      if(propAxioms != null && !equivalentAxioms.containsAll(propAxioms))
          equivalentAxioms.addAll(propAxioms);

      Vector<Axiom> tempAxioms = new Vector<Axiom>();
      Vector<Axiom> allAxioms = (Vector<Axiom>)ont.getAllAxioms();

      boolean el1, el2, eliminate = false;
      if(equivalentAxioms == null)
          return;
      for(Iterator<Axiom> it1=equivalentAxioms.iterator();it1.hasNext();){
          Axiom a = it1.next();
          AxiomNode subject = a.getLeftAxiomNode();
          AxiomNode object = a.getRightAxiomNode();

          // If subject is anonymous and object is completely defined
          if(subject.getJoinType() == Ontology.ATOMICNODE && isAnonymous(subject.getAtomicNode()) &&
                  ((object.getJoinType() > Ontology.ATOMICNODE && !containsAnonymousChildren(object)) ||
                          (object.getJoinType() == Ontology.ATOMICNODE && !isAnonymous(object.getAtomicNode())) ) ){

              for(Iterator<Axiom> it2=allAxioms.iterator();it2.hasNext();){
                  Axiom ax = it2.next();
                  if(!ax.equals(a)){
                      AxiomNode sub = ax.getLeftAxiomNode();
                      AxiomNode obj = ax.getRightAxiomNode();
                      el1 = eliminateAnonymousNode(ont, ax, sub, subject, object, 0);
                      el2 = eliminateAnonymousNode(ont, ax, obj, subject, object, 1);
                      if(!eliminate && (el1 || el2))
                          eliminate = true;
                  }
              }
//			  if(eliminate){
//				  ont.getAxiomBySubject(subject.getAtomicNode()).remove(a);
//				  if(object.getJoinType() == Ontology.ATOMICNODE)
//					  ont.getAxiomByObject(object.getAtomicNode()).remove(a);	// Ameet, Dec 12 2008, It is important to eliminate
//			  }																	// the axioms from all sets
          }
          // If object is anonymous and subject is completely defined
          else if(object.getJoinType() == Ontology.ATOMICNODE && isAnonymous(object.getAtomicNode()) &&
                  ((subject.getJoinType() > Ontology.ATOMICNODE && !containsAnonymousChildren(subject)) ||
                          (subject.getJoinType() == Ontology.ATOMICNODE && !isAnonymous(subject.getAtomicNode())) ) ){

              for(Iterator<Axiom> it2=allAxioms.iterator();it2.hasNext();){
                  Axiom ax = it2.next();
                  if(!ax.equals(a)){
                      AxiomNode sub = ax.getLeftAxiomNode();
                      AxiomNode obj = ax.getRightAxiomNode();
                      el1 = eliminateAnonymousNode(ont, ax, sub, object, subject, 0);
                      el2 = eliminateAnonymousNode(ont, ax, obj, object, subject, 1);
                      if(!eliminate && (el1 || el2))
                          eliminate = true;
                  }
              }
//              if(eliminate){
//				  ont.getAxiomByObject(object.getAtomicNode()).remove(a);
//				  if(subject.getJoinType() == Ontology.ATOMICNODE)
//					  ont.getAxiomBySubject(subject.getAtomicNode()).remove(a);		// Ameet, Dec 12 2008, It is important to eliminate
//              }																		// the axioms from all sets
          }
          if(eliminate){
              ont.getAllAxioms().remove(a);
              eliminate = false;
              tempAxioms.add(a);
          }
      }

      equivalentAxioms.removeAll(tempAxioms);
      if(classAxioms != null)
          classAxioms.removeAll(tempAxioms);
      if(propAxioms != null)
          propAxioms.removeAll(tempAxioms);
  }

  /** Returns true of the given ObjectId is an anonymous node
   *
   * @param id
   * @return
   */
  protected boolean isAnonymous(ObjectId id){
	  if(id!=null && id.getId().startsWith("[") && id.getId().endsWith("]"))
		  return true;
	  else
		  return false;
  }

  /** This method determines if a given AxiomNode contains any anonymous children
   *
   * @param node
   * @return
   */

  protected boolean containsAnonymousChildren(AxiomNode node){
	  Vector<AxiomNode> children = node.getChildren();
	  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
		  AxiomNode child = it.next();
		  if(child.getJoinType() == Ontology.ATOMICNODE){
			  if(isAnonymous(child.getAtomicNode()))
				  return true;
		  }else if(child.getJoinType() > Ontology.ATOMICNODE && containsAnonymousChildren(child))
			  return true;
	  }
	  return false;
  }

  /** If the given AxiomNode (node) or any of its descendants are equivalent to the anonymous node (aNode) then they
   *  are replaced by the replacement node (rNode)
   */
  protected boolean eliminateAnonymousNode(Ontology ont, Axiom a, AxiomNode node,
		  AxiomNode aNode, AxiomNode rNode, int d){
	  boolean eliminate = false;
	  if(node.getJoinType() == Ontology.ATOMICNODE && node.getAtomicNode().equals(aNode.getAtomicNode())){
		  if(!topLevel(ont, node)){
			  if(d == 0){
			  	  a.setLeftAxiomNode(rNode);
				  if(!eliminate)
					  eliminate = true;
			  }else if(d == 1){
				  a.setRightAxiomNode(rNode);
				  if(!eliminate)
					  eliminate = true;
			  }
		  }else if(isAnonymous(node.getAtomicNode())){  //zhp2 on 7/2/08
                     if(d>0)
                         a.setRightAxiomNode(rNode);
                     else
                         a.setLeftAxiomNode(rNode);

                     eliminate = true;
                  }
        }else if(node.getJoinType() > Ontology.ATOMICNODE){
		  Vector<AxiomNode> children = node.getChildren();
		  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
			  boolean temp = eliminateAnonymousNode(ont, node, it.next(), aNode, rNode);
			  if(!eliminate)
				  eliminate = temp;
		  }
	  }
	  return eliminate;
  }

  protected boolean eliminateAnonymousNode(Ontology ont, AxiomNode parent,
		  AxiomNode node, AxiomNode aNode, AxiomNode rNode){
	  boolean eliminate = false;
	  if(node.getJoinType() == Ontology.ATOMICNODE && node.getAtomicNode().equals(aNode.getAtomicNode())){
		  int index = parent.getChildren().indexOf(node);
		  parent.getChildren().set(index, rNode);
		  if(!eliminate)
			  eliminate = true;
	  }else if(node.getJoinType() > Ontology.ATOMICNODE){
		  Vector<AxiomNode> children = node.getChildren();
		  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
			  boolean temp = eliminateAnonymousNode(ont, node, it.next(), aNode, rNode);
			  if(!eliminate)
				  eliminate = temp;
		  }
	  }
	  return eliminate;
  }

  /** Returns true if the given AxiomNode is a top level node i.e. does not occur in the object position of any
   *  axiom
   * @param ont
   * @param node
   * @return
   */
  protected boolean topLevel(Ontology ont, AxiomNode node){
	  Vector<Axiom> allAxioms = (Vector<Axiom>)ont.getAllAxioms();

	  for(Iterator<Axiom> it2=allAxioms.iterator();it2.hasNext();){
		  Axiom ax = it2.next();
		  AxiomNode obj = ax.getRightAxiomNode();
		  if(obj.getJoinType() == Ontology.ATOMICNODE && obj.getAtomicNode().equals(node.getAtomicNode()))
			  return false;
		  else if(obj.getJoinType() > Ontology.ATOMICNODE){
			  Vector<AxiomNode> children = obj.getChildren();
			  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
				  if(!topLevel(it.next(), node))
					  return false;
			  }
		  }
	  }
	  return true;
  }

  protected boolean topLevel(AxiomNode node, AxiomNode aNode){

	  if(node.getJoinType() == Ontology.ATOMICNODE && node.getAtomicNode().equals(aNode.getAtomicNode()))
		  return false;
	  else if(node.getJoinType() > Ontology.ATOMICNODE){
		  Vector<AxiomNode> children = node.getChildren();
		  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
			  if(!topLevel(it.next(), aNode))
				  return false;
		  }
	  }
	  return true;
  }

  /** This method resolves all the restrictions nested within other axioms/restrictions and also stores the top level restrictions
   *  into the ontology
   *
   * @param ont
   */

  public void resolveAnonymousRestrictions(Ontology ont){
	  PropertyRestriction[] restrictions = ont.getAllRestrictions();
	  Vector<AxiomNode> atomicRestrictions = new Vector<AxiomNode>();
	  Vector<AxiomNode> nonAtomicRestrictions = new Vector<AxiomNode>();
	  Vector<Axiom> allAxioms = (Vector<Axiom>)ont.getAllAxioms();

		for(int i=0;i<restrictions.length;i++){		// Identify all the atomic restrictions
			PropertyRestriction pr = restrictions[i];
			if(!isAnonymous(pr.getOnProperty())){
				if(isAnonymous(pr.getSomeValuesFrom())){
					AxiomNode propRes = new AxiomNode();
					propRes.setRestriction(pr);
					nonAtomicRestrictions.add(propRes);
					continue;
				}else if(isAnonymous(pr.getAllValuesFrom())){
					AxiomNode propRes = new AxiomNode();
					propRes.setRestriction(pr);
					nonAtomicRestrictions.add(propRes);
					continue;
				}else if(isAnonymous(pr.getHasValue())){
					AxiomNode propRes = new AxiomNode();
					propRes.setRestriction(pr);
					nonAtomicRestrictions.add(propRes);
					continue;
				}
				AxiomNode propRes = new AxiomNode();
				propRes.setRestriction(pr);
				atomicRestrictions.add(propRes);
			}else{
				AxiomNode propRes = new AxiomNode();
				propRes.setRestriction(pr);
				nonAtomicRestrictions.add(propRes);
			}
		}
		Vector<AxiomNode> temp = new Vector<AxiomNode>();

		for(Iterator<AxiomNode> it=nonAtomicRestrictions.iterator();it.hasNext();){
			AxiomNode nonAtomicRes = it.next();
			if(nonAtomicRes.getOnProperty().getJoinType() == Ontology.ATOMICNODE &&
					isAnonymous(nonAtomicRes.getOnProperty().getAtomicNode()))
				resolveAnonResProperties(nonAtomicRes, temp, ont);
		}
		nonAtomicRestrictions.removeAll(temp);
		atomicRestrictions.addAll(temp);
		while(!nonAtomicRestrictions.isEmpty() || !atomicRestrictions.isEmpty()){
//			System.out.println(atomicRestrictions.size());
			temp.removeAllElements();
			temp.addAll(atomicRestrictions);
			int count=0;
			Vector<AxiomNode> tempA = new Vector<AxiomNode>();
			for(Iterator<AxiomNode> it=temp.iterator();it.hasNext();){
				count++;
				AxiomNode atomicRes = it.next();
				resolveNonAtomicRestrictions(nonAtomicRestrictions, atomicRes, atomicRestrictions, tempA);
			}
//				temp = new Vector<AxiomNode>();  //zhp2 on 06/13/2009 comment out 5 lines down from here, were in for loop
//				temp.addAll(atomicRestrictions);
//				it = temp.iterator();
//				for(int j=0;j<count;j++)
//					it.next();
//			}
			atomicRestrictions.removeAll(tempA);

			Vector<Axiom> tempAxioms = new Vector<Axiom>();
			allAxioms = (Vector<Axiom>)ont.getAllAxioms();
			for(Iterator<Axiom> it2=allAxioms.iterator();it2.hasNext();){
				Axiom a = it2.next();
				AxiomNode left = a.getLeftAxiomNode();
				AxiomNode right = a.getRightAxiomNode();

				if(left.getJoinType() == Ontology.ATOMICNODE && isAnonymous(left.getAtomicNode()) &&
					  ((right.getJoinType() > Ontology.ATOMICNODE && !containsAnonymousChildren(right)) ||
							  (right.getJoinType() == Ontology.ATOMICNODE && !isAnonymous(right.getAtomicNode())) ) ){
					resolveNonAtomicRestrictions(nonAtomicRestrictions, a, atomicRestrictions, tempAxioms, ont);
				}
			}
			allAxioms.removeAll(tempAxioms);

			allAxioms = (Vector<Axiom>)ont.getAllAxioms();
			temp.removeAllElements();

			for(Iterator<AxiomNode> it = atomicRestrictions.iterator();it.hasNext();){
				AxiomNode atomicRes = it.next();
				for(Iterator<Axiom> it2=allAxioms.iterator();it2.hasNext();){
					Axiom a = it2.next();
					resolveRestrictionsInAxioms(a, 0, atomicRes, temp);
					resolveRestrictionsInAxioms(a, 1, atomicRes, temp);
				}
			}
			atomicRestrictions.removeAll(temp);
//            atomicRestrictions.removeAllElements();  //added by zhp2 on 06/13/2009, some restrictions do not participate in any axioms.
 //           nonAtomicRestrictions.removeAllElements();  //added by zhp2 on 06/13/2009, some restrictions do not participate in any axioms.
			//zhp2 commented out on 7/3/08
            mergeEquivalents(ont);
		}
//        mergeEquivalents(ont); //zhp2 on 7/3/08
		ont.addAllAxiomNodeRestrictions(atomicRestrictions);
  }

  /** This method return the object of the axioms where the given sid plays the role of the subject
   *
   * @param sid
   * @param ont
   * @return
   */

  public Axiom getCompletelyDefinedDefinitions(ObjectId sid, Ontology ont){

	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(sid);

	  return axioms.get(0);

  }

  /**
   * This method resolves all the anonymous restrictions occuring in an axiom by replacing them with their equivalent atomic restrictions.
   * Atomic restrictions are completed defined in terms of their onProperty, allValuesFrom/someValuesFrom/hasValue
   * @param a
   * @param pos
   * @param atomicRes
   * @param temp
   */

  public void resolveRestrictionsInAxioms(Axiom a, int pos, AxiomNode atomicRes,
		  Vector<AxiomNode> temp){
	  AxiomNode node = null;
	  if(pos==0){
		  node = a.getLeftAxiomNode();
		  if(node.getJoinType() == Ontology.ATOMICNODE && node.getAtomicNode().getId().equals(atomicRes.getRestrictionId())){
			  a.setLeftAxiomNode(atomicRes);
			  if(!temp.contains(atomicRes))
				  temp.add(atomicRes);
		  }
	  }else if(pos==1){
		  node = a.getRightAxiomNode();
		  if(node.getJoinType() == Ontology.ATOMICNODE && node.getAtomicNode().getId().equals(atomicRes.getRestrictionId())){
			  a.setRightAxiomNode(atomicRes);
			  if(!temp.contains(atomicRes))
				  temp.add(atomicRes);
		  }
	  }

	  if(node.getJoinType() > Ontology.ATOMICNODE){
		  Vector<AxiomNode> children = node.getChildren();
		  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
			  resolveRestrictionsInAxioms(node, it.next(), atomicRes, temp);
		  }
	  }
  }

  public void resolveRestrictionsInAxioms(AxiomNode parent, AxiomNode child, AxiomNode atomicRes,
		  Vector<AxiomNode> temp){

	  if(child.getJoinType() == Ontology.ATOMICNODE && child.getAtomicNode().getId().equals(atomicRes.getRestrictionId())){
		  int index = parent.getChildren().indexOf(child);
		  parent.getChildren().set(index, atomicRes);
		  if(!temp.contains(atomicRes))
			  temp.add(atomicRes);
	  }else if(child.getJoinType() > Ontology.ATOMICNODE){
		  Vector<AxiomNode> children = child.getChildren();
		  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
			  resolveRestrictionsInAxioms(child, it.next(), atomicRes, temp);
		  }
	  }
  }

  /** This method resolves the anonymous properties occuring within a restriction. ex: if the onProperty is the inverse of some other property.
   *
   * @param propRes
   * @param temp
   * @param ont
   */

  public void resolveAnonResProperties(AxiomNode propRes, Vector<AxiomNode> temp, Ontology ont){

	  Axiom axiom = getCompletelyDefinedDefinitions(propRes.getOnProperty().getAtomicNode(), ont);
	  AxiomNode prop = new AxiomNode();
	  prop.setJoinType(axiom.getType());
	  prop.addChild(axiom.getRightAxiomNode());
	  propRes.setOnProperty(prop);

	  ont.getAxiomByType(new Integer(axiom.getType())).remove(axiom);

	  ont.getAxiomBySubject(axiom.getLeftAxiomNode().getAtomicNode()).remove(axiom);

	  ont.getAxiomByObject(axiom.getRightAxiomNode().getAtomicNode()).remove(axiom);

	  ont.getAllAxioms().remove(axiom);

	  if(!isAnonRestriction(propRes)){
		  if(!temp.contains(propRes))
			  temp.add(propRes);
	  }

  }

  /**
   * This method resolves all the non-atomic restrictions. Restrictions are non-atomic if either of their someValuesFrom/allValuesFrom/hasValue
   * are anonymous.
   *
   * @param nonAtomicRestrictions
   * @param atomicRes
   * @param atomicRestrictions
   * @param tempA
   */
  public void resolveNonAtomicRestrictions(Vector<AxiomNode> nonAtomicRestrictions,
		  AxiomNode atomicRes, Vector<AxiomNode> atomicRestrictions, Vector<AxiomNode> tempA){

	  Vector<AxiomNode> temp = new Vector<AxiomNode>();
	  for(Iterator<AxiomNode> it=nonAtomicRestrictions.iterator();it.hasNext();){
		  AxiomNode nonAtomicRes = it.next();
		  if(nonAtomicRes.getAllValuesFrom()!=null &&
				  nonAtomicRes.getAllValuesFrom().getJoinType() == Ontology.ATOMICNODE){
			  ObjectId allVal = nonAtomicRes.getAllValuesFrom().getAtomicNode();
			  if(allVal.getId().equals(atomicRes.getRestrictionId())){
				  nonAtomicRes.setAllValuesFrom(atomicRes);
				  if(!tempA.contains(atomicRes))
					  tempA.add(atomicRes);
			  }
		  }else if(nonAtomicRes.getSomeValuesFrom()!=null &&
			  nonAtomicRes.getSomeValuesFrom().getJoinType() == Ontology.ATOMICNODE){
			  ObjectId allVal = nonAtomicRes.getSomeValuesFrom().getAtomicNode();
			  if(allVal.getId().equals(atomicRes.getRestrictionId())){
				  nonAtomicRes.setSomeValuesFrom(atomicRes);
				  if(!tempA.contains(atomicRes))
					  tempA.add(atomicRes);
			  }
		  }else if(nonAtomicRes.getHasValue()!=null &&
			  nonAtomicRes.getHasValue().getJoinType() == Ontology.ATOMICNODE){
			  ObjectId allVal = nonAtomicRes.getHasValue().getAtomicNode();
			  if(allVal.getId().equals(atomicRes.getRestrictionId())){
				  nonAtomicRes.setHasValue(atomicRes);
				  if(!tempA.contains(atomicRes))
					  tempA.add(atomicRes);
			  }
		  }

		  if(!isAnonRestriction(nonAtomicRes)){
			  if(!temp.contains(nonAtomicRes))
				  temp.add(nonAtomicRes);
		  }
	  }

	  nonAtomicRestrictions.removeAll(temp);
	  atomicRestrictions.addAll(temp);

  }

  /**
   * This method resolves all the non-atomic restrictions. Restrictions are non-atomic if either of their someValuesFrom/allValuesFrom/hasValue
   * are anonymous.
   *
   * @param nonAtomicRestrictions
   * @param atomicRes
   * @param atomicRestrictions
   * @param tempA
   */
  public void resolveNonAtomicRestrictions(Vector<AxiomNode> nonAtomicRestrictions,
		  Axiom axiom, Vector<AxiomNode> atomicRestrictions, Vector<Axiom> tempA, Ontology ont){

	  AxiomNode left = axiom.getLeftAxiomNode();
	  AxiomNode right = axiom.getRightAxiomNode();

	  Vector<AxiomNode> temp = new Vector<AxiomNode>();
	  for(Iterator<AxiomNode> it=nonAtomicRestrictions.iterator();it.hasNext();){
		  AxiomNode nonAtomicRes = it.next();
		  if(nonAtomicRes.getAllValuesFrom()!=null &&
				  nonAtomicRes.getAllValuesFrom().getJoinType() == Ontology.ATOMICNODE){
			  ObjectId allVal = nonAtomicRes.getAllValuesFrom().getAtomicNode();
			  if(allVal.getId().equals(left.getAtomicNode().getId())){

				  if(axiom.getType() == Ontology.EQUIVALENTCLASS || axiom.getType() == Ontology.EQUIVALENTPROPERTY){
					  nonAtomicRes.setAllValuesFrom(right);		// Ameet Oct 2 2008
				  }else{
					  AxiomNode aNode = new AxiomNode();
					  aNode.setJoinType(axiom.getType());
					  aNode.addChild(right);
					  nonAtomicRes.setAllValuesFrom(aNode);
				  }
                  if(ont.getAxiomByType(new Integer(axiom.getType()))!=null)  //zhp2 on 11/24/08
                      ont.getAxiomByType(new Integer(axiom.getType())).remove(axiom);
                  if(ont.getAxiomBySubject(left.getAtomicNode())!=null)                //zhp2 on 11/24/08
                      ont.getAxiomBySubject(left.getAtomicNode()).remove(axiom);

				  if(ont.getAxiomByObject(right.getAtomicNode())!=null)
					  ont.getAxiomByObject(right.getAtomicNode()).remove(axiom);

				  if(!tempA.contains(axiom))
					  tempA.add(axiom);
			  }
		  }else if(nonAtomicRes.getSomeValuesFrom()!=null &&
			  nonAtomicRes.getSomeValuesFrom().getJoinType() == Ontology.ATOMICNODE){
			  ObjectId someVal = nonAtomicRes.getSomeValuesFrom().getAtomicNode();
			  if(someVal.getId().equals(left.getAtomicNode().getId())){

				  if(axiom.getType() == Ontology.EQUIVALENTCLASS || axiom.getType() == Ontology.EQUIVALENTPROPERTY){
					  nonAtomicRes.setSomeValuesFrom(right);		// Ameet Oct 2 2008
				  }else{
					  AxiomNode aNode = new AxiomNode();
					  aNode.setJoinType(axiom.getType());
					  aNode.addChild(right);
					  nonAtomicRes.setSomeValuesFrom(aNode);
				  }


//				  AxiomNode aNode = new AxiomNode();	// commented by Ameet on Oct 2 2008
//                  //zhp2 on 7/3/08 start
//                  if(axiom.getType()==Ontology.EQUIVALENTCLASS){
//                      aNode.setJoinType(right.getJoinType());
//                      aNode.addChild(right.getChildren());
//                  }else
//                      continue;
                  //zhp2 on 7/3/08 end also comment out following two lines
                  /**
				  aNode.setJoinType(axiom.getType());
				  aNode.addChild(right);
                  **/
                  if (ont.getAxiomByType(new Integer(axiom.getType())) != null)
                    ont.getAxiomByType(new Integer(axiom.getType())).remove(axiom);
                  if (ont.getAxiomBySubject(left.getAtomicNode()) != null)
				    ont.getAxiomBySubject(left.getAtomicNode()).remove(axiom);

				  if(ont.getAxiomByObject(right.getAtomicNode())!=null)
					  ont.getAxiomByObject(right.getAtomicNode()).remove(axiom);

				  if(!tempA.contains(axiom))
					  tempA.add(axiom);
			  }
		  }else if(nonAtomicRes.getHasValue()!=null &&
			  nonAtomicRes.getHasValue().getJoinType() == Ontology.ATOMICNODE){
			  ObjectId hasVal = nonAtomicRes.getHasValue().getAtomicNode();
			  if(hasVal.getId().equals(left.getAtomicNode().getId())){

				  if(axiom.getType() == Ontology.EQUIVALENTCLASS || axiom.getType() == Ontology.EQUIVALENTPROPERTY){
					  nonAtomicRes.setHasValue(right);		// Ameet Oct 2 2008
				  }else{
					  AxiomNode aNode = new AxiomNode();
					  aNode.setJoinType(axiom.getType());
					  aNode.addChild(right);
					  nonAtomicRes.setHasValue(aNode);
				  }

				  ont.getAxiomByType(new Integer(axiom.getType())).remove(axiom);

				  ont.getAxiomBySubject(left.getAtomicNode()).remove(axiom);

				  if(ont.getAxiomByObject(right.getAtomicNode())!=null)
					  ont.getAxiomByObject(right.getAtomicNode()).remove(axiom);

				  if(!tempA.contains(axiom))
					  tempA.add(axiom);
			  }
		  }

		  if(!isAnonRestriction(nonAtomicRes)){
			  if(!temp.contains(nonAtomicRes))
				  temp.add(nonAtomicRes);
		  }
	  }

	  nonAtomicRestrictions.removeAll(temp);
	  atomicRestrictions.addAll(temp);

  }

  /**
   * This method returns true if the given restriction node is anonymous, false otherwise
   * @param res
   * @return
   */
  public boolean isAnonRestriction(AxiomNode res){

	  if(res.getOnProperty().getJoinType() == Ontology.ATOMICNODE &&
			isAnonymous(res.getOnProperty().getAtomicNode()))
		  return true;

	  if(res.getAllValuesFrom()!=null && res.getAllValuesFrom().getJoinType() == Ontology.ATOMICNODE &&
			isAnonymous(res.getAllValuesFrom().getAtomicNode()))
	  	  return true;

	  if(res.getSomeValuesFrom()!=null && res.getSomeValuesFrom().getJoinType() == Ontology.ATOMICNODE &&
			isAnonymous(res.getSomeValuesFrom().getAtomicNode()))
		  return true;

	  if(res.getHasValue()!=null && res.getHasValue().getJoinType() == Ontology.ATOMICNODE &&
			isAnonymous(res.getHasValue().getAtomicNode()))
		  return true;

	  return false;
  }

  /**
     * OilEd: Unpacks a RDF List into a collection of resources
     **/
  protected Collection unpackList(Model currentModel, Resource r) {
	    Collection result = new ArrayList();
	    /* Grab the first thing in the list */
	    Resource current = r;
	    int firstCount = count(currentModel.listObjectsOfProperty(r, RDF.first));
	    NodeIterator firsts = currentModel.listObjectsOfProperty(r, RDF.first);
	    while (firsts!=null && firstCount==1) {
              RDFNode fnode = firstNode(firsts);
              if(fnode.isResource()){
                    Resource listHead = (Resource) fnode;
                    result.add(listHead);
              }
              else return null;

		/* Find the next bit of the list */
        NodeIterator rests = currentModel.listObjectsOfProperty(current, RDF.rest);
        int restCount = count(currentModel.listObjectsOfProperty(current, RDF.rest));

        if (restCount==0) {
		    /* There's nothing else. Terminate the loop */
          firsts = null;
        }
        else {
          current = (Resource) firstNode(rests);

          if (currentModel.contains(current, RDF.type, RDF.nil)) {
			/* If the rest of the list is empty, terminate the loop */
            firsts = null;
          }
          else {
			/* otherwise, move along the list */
            firsts = currentModel.listObjectsOfProperty(current, RDF.first);
            firstCount = count(currentModel.listObjectsOfProperty(current, RDF.first));
          }
        }
	    }
	    return result;
  }

  /**
   * OilEd: Returns the first statement in a model. Assumes there is one, so
	 * you've better have checked that before calling this.
   */
  protected RDFNode firstNode(NodeIterator ri){
    if (ri.hasNext()) {
      return ri.nextNode();
    }
    else {
      return null;
    }
  }

  /**
   * Get the comment of a resource.
   */
  private String getComment(Resource r, Model model){
    NodeIterator ni = model.listObjectsOfProperty(r,RDFS.comment);
    if(ni.hasNext()){
      Literal l = (Literal)firstNode(ni);
      return l.getLexicalForm();
    }
    else
      return null;
  }

  /**
   * Get the label of a resource.
   */
  private String getLabel(Resource r, Model model){
    NodeIterator ni = model.listObjectsOfProperty(r,RDFS.label);
    if(ni.hasNext()){
      Literal l = (Literal)firstNode(ni);
      return l.getLexicalForm();
    }
    else
      return null;
  }

  /**
   * Add a resource as Class to the ontology.
   */
  private void addClass(Ontology ont, Resource r, String s, String comment
                        ,String label){
    ObjectId id = resourceToId(r);
    if(id==null)
      return;
//    if(!id.getNameSpace().replace('#',' ').trim().equalsIgnoreCase(ont.getXMLBase()))
      //return;  //pzx: 10/16/04: don't add foreign classes

     if(id.getNameSpace().equals("")&&id.getLocalName().equals("")) //zhp2 on 070709
         return;

    if(ont.getOntClass(id)==null){
      OntClass oc = ont.createClass();
      oc.setID(id);
      if(s.equalsIgnoreCase("deprecated"))
        oc.setDeprecated(true);
      if(comment!=null)
        oc.addDescription(comment);
      if(label!=null)
        oc.setLabel(label);
      if(!id.getNameSpace().replace('#',' ').trim().equalsIgnoreCase(ont.getXMLBase())) //zhp2 on 1/20/07: add 3 lines down
          ont.addImportedClass(id, oc);
      else
          ont.addOntClass(oc);
    }
  }

  /**
   * Add a resource as property.
   */
  private OntProperty addProperty(Ontology ont, Resource r, Model model, Boolean datatype){
    ObjectId id = resourceToId(r);
    if(id==null)
      return null;
 //   if(!r.getNameSpace().replace('#',' ').trim().equalsIgnoreCase(ont.getXMLBase()))
 //     return null;

    if(id.getNameSpace().equals("")&&id.getLocalName().equals("")) //zhp2 on 070709
         return null;

     OntProperty op = ont.getProperty(id);
    if(op == null){
      op = ont.createProperty();
      op.setID(id);
      op.addDescription(getComment(r,model));
      op.setLabel(getLabel(r,model));
      if(datatype!=null)
        op.setDatatype(datatype.booleanValue());
    //zhp2: suppport xml datatypes
//    StmtIterator si = null;										// Ameet, Oct 2 2008
//    si = model.listStatements(new SelectorImpl(r, RDFS.domain,	// Since domain and range are now being read as axioms we need not
//        (RDFNode) null));											// read domain and range again over here.
//    while(si.hasNext()){
//      Statement stmt = si.nextStatement();
//      Resource obj = (Resource)stmt.getObject();
//      ObjectId oid = resourceToId(obj);
//      if(oid==null)
//        continue;
//      op.setDomain(oid);
//    }
//
//    si = model.listStatements(new SelectorImpl(r, RDFS.range,
//        (RDFNode) null));
//    while(si.hasNext()){
//      Statement stmt = si.nextStatement();
//      ObjectId oid = null;
////      RDFNode objnode  = stmt.getObject();
////      if(objnode instanceof Resource){
//        Resource obj = (Resource)stmt.getObject();
//        oid = resourceToId(obj);
///*      }else if(objnode instanceof Literal){
//            Literal vl = (Literal)objnode;
//            String s = vl.getLexicalForm();
//            String dtype;
//            if(vl.getDatatype()!=null)
//              dtype = vl.getDatatype().getURI();
//            else
//              dtype = "";
//            DatatypeValue dv = new DatatypeValue(s,dtype);
//            oid = dv;
//          }
// */
//      if(oid==null)
//        continue;
//      op.setRange(oid);
//    }
    // The check for namespace is needed to prevent null pointer exception
    // Ameet, Oct 9 2007
    if(r.getNameSpace()!=null && !r.getNameSpace().replace('#',' ').trim().equalsIgnoreCase(ont.getXMLBase()))
        ont.addImportedProperty(id, op);
    else
        ont.addProperty(op);
    }
    return op;
  }


  /**
   * Wrap a resource into an ObjectId object.
   */
  private ObjectId resourceToId(Resource r){
    if(r==null)
      return null;
    if(!r.canAs(resource))
      return null;
    String s = r.toString();
    if(s==null)
      return null;
    if(r.isAnon())
      s = "["+s+"]";
    ObjectId id = new ObjectId(s);
    //id.setLocalName(r.getLocalName());
    //id.setNameSpace(r.getNameSpace());
    return id;
  }

  /**
   * OilEd: Count the number of elements in NodeIterator
   */
  private int count(NodeIterator ni){
    int c = 0;
    while (ni.hasNext()) {
	    ni.next();
	    c++;
    }
    return c;
  }

  /**
   * Combine two sets into one.
   */
  private Collection combineSets(Collection c1, Collection c2){
    ArrayList c = new ArrayList(c1);
    c.addAll(c2);
    return c;
  }

  private boolean isVocabularyResource(Resource r){
    if(r.equals(OWL.AllDifferent))
      return true;
    if(r.equals(OWL.Class))
      return true;
    if(r.equals(RDFS.Datatype))
      return true;
    if(r.equals(OWL.DatatypeProperty))
      return true;
    if(r.equals(OWL.DeprecatedClass))
      return true;
    if(r.equals(OWL.DeprecatedProperty))
      return true;
    if(r.equals(OWL.AnnotationProperty))
      return true;
    if(r.equals(OWL.FunctionalProperty))
      return true;
    if(r.equals(OWL.InverseFunctionalProperty))
      return true;
    if(r.equals(RDF.List))
      return true;
    if(r.equals(RDFS.Literal))
      return true;
    if(r.equals(RDF.nil))
      return true;
    if(r.equals(OWL.Nothing))
      return true;
    if(r.equals(OWL.ObjectProperty))
      return true;
    if(r.equals(OWL.Ontology))
      return true;
    if(r.equals(OWL.Restriction))
      return true;
    if(r.equals(OWL.SymmetricProperty))
      return true;
//   if(r.equals(OWL.Thing))
//     return true;
    if(r.equals(OWL.TransitiveProperty))
      return true;
    if(r.equals(RDF.Alt))
      return true;
    if(r.equals(RDF.Bag))
      return true;
    if(r.equals(RDF.Seq))
      return true;
    if(r.equals(RDF.Property))
      return true;
    if(r.equals(RDF.Statement))
      return true;
    if(r.equals(RDFS.Class))
      return true;
//    if(r.equals(RDFS.ConstraintProperty))
//      return true;
//    if(r.equals(RDFS.ConstraintResource))
//      return true;
    if(r.equals(RDFS.Container))
      return true;
    if(r.equals(RDFS.ContainerMembershipProperty))
      return true;
    return false;
  }

  private boolean isVocabularyProperty(Property r){
      if (r.equals(OWL.allValuesFrom))
          return true;
      if (r.equals(OWL.backwardCompatibleWith))
          return true;
      if (r.equals(OWL.cardinality))
          return true;
      if (r.equals(OWL.complementOf))
          return true;
      if (r.equals(OWL.differentFrom))
          return true;
      if (r.equals(OWL.disjointWith))
          return true;
      if (r.equals(OWL.distinctMembers))
          return true;
      if (r.equals(OWL.equivalentClass))
          return true;
      if (r.equals(OWL.equivalentProperty))
          return true;
      if (r.equals(OWL.hasValue))
          return true;
      if (r.equals(OWL.imports))
          return true;
      if (r.equals(OWL.incompatibleWith))
          return true;
      if (r.equals(OWL.intersectionOf))
          return true;
      if (r.equals(OWL.inverseOf))
          return true;
      if (r.equals(OWL.maxCardinality))
          return true;
      if (r.equals(OWL.minCardinality))
          return true;
      if (r.equals(OWL.oneOf))
          return true;
      if (r.equals(OWL.onProperty))
          return true;
      if (r.equals(OWL.priorVersion))
          return true;
      if (r.equals(OWL.sameAs))
          return true;
      if (r.equals(OWL.someValuesFrom))
          return true;
      if (r.equals(OWL.unionOf))
          return true;
      if (r.equals(OWL.versionInfo))
          return true;
      if (r.equals(RDFS.range))
          return true;
      if (r.equals(RDFS.domain))
          return true;
      if (r.equals(RDFS.seeAlso))
          return true;
      if (r.equals(RDFS.subClassOf))
          return true;
      if (r.equals(RDFS.subPropertyOf))
          return true;
      if (r.equals(RDFS.isDefinedBy))
          return true;
      if (r.equals(RDFS.member))
          return true;
      if (r.equals(RDF.first))
          return true;
      if(r.equals(RDF.rest))
          return true;

      return false;
  }

  public Proxy getProxy(){
	  return this.proxy;
  }
  public void setProxy(Proxy p){
	  this.proxy = p;
  }

  private StorageModel smodel;
  private Class resource = null; //to use REFNode.canAs(Class)
  private Vector blacklist;
  private Proxy proxy;

}
