/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/

package edu.lehigh.swat.hawk.storage.dlmemory;

//import java.util.*;
import java.io.*;
import java.net.*;
import edu.lehigh.swat.hawk.core.*;
import edu.lehigh.swat.hawk.owl.OwlDigWriter;
import org.kr.dl.dig.v1_1.*;
import org.kr.dl.dig.reasoner.v1_1.Reasoner;
import org.kr.dl.dig.reasoner.v1_1.impl.HTTPReasoner;



/**
 * An implementation of core.ontology, can do inference with a DL reasoner.
 * @author Z. Pan
 * @version 1.0
 */
public class Ontology extends edu.lehigh.swat.hawk.storage.simplememory.Ontology{

  /**
   * Constructor.
   */
  public Ontology(){
    super();
  }


  /**Use a DL reasoner to do inference on this ontology, then rearrange it based upon
   * the results. Currently we check the satisfibility and ask for parents of each
   * class. We also asks for the equivalent classes of each class.
   * Note: we don't delete the restrictions in the original superClass set.
   * Thus in one class's superClass set, it's possible that some property restrictions subsume one of the direct
   * superClass.
   * @param rUrl The url of the http reasoner, by default it is http://localhost:8080 .
   */
  public void computeSubsumptions(String rUrl, edu.lehigh.swat.hawk.core.Ontology[] imports){
    try{
      //System.setProperty("dig.shiqapp","D:\\OWLparser\\OWLParserreasoner\\shiq-app.exe");
      if(rUrl==null||rUrl.equals(""))
        rUrl = "http://localhost:8081";
      URL reasonerUrl = new URL(rUrl);
      Reasoner r = new HTTPReasoner(reasonerUrl);
      System.out.println("*****Reasoner URL is: "+rUrl+"\n");
      int timeOut = 300; //seconds
      long to = timeOut*1000;  //ms

      URI kburi = r.newKB();
      OwlDigWriter writer = new OwlDigWriter();
      StringWriter sw = new StringWriter();
      TellsDocument td = writer.writeOntology(this, sw, kburi.toString());
      sw.flush();
      /*  //test
      PrintStream ps = new PrintStream(new FileOutputStream("d:\\hawk\\test\\tells.xml"));
      ps.print(sw.toString());
      ps.close();
      */ //end test
      td.save(new File("tells.xml"));
     //td = TellsDocument.Factory.parse(sw.toString());

      ReasonerTeller rteller = new ReasonerTeller(r,td);
      long startTime = System.currentTimeMillis();
      rteller.start();
//      Timer timer = new Timer();
//      EndTask endthis = new EndTask(rteller);
//      timer.schedule(endthis,to);

      rteller.join(to);
      long duration = (System.currentTimeMillis()-startTime);
      System.out.println("Return from reasoner after "+duration);
      ResponseDocument rd = rteller.getResponse();
      //rd.save(new File("responseForTells.xml")); //test
      if(rd==null){
        r.releaseKB(kburi);
        return;
      }
//      endthis.cancel();
      IdRespType rt = rd.getResponse();
      if(rt.isSetOk())
        System.out.println("tells OK");
      else if(rt.isSetError()){
        System.out.println("tells error");
        ErrorDocument.Error er = rt.getError();
        System.out.println(er.getMessage());
        td.save(new File("tells.xml"));
        return;
      }

      if(imports!=null){
        int len = imports.length;
        for(int i=0;i<len;i++){
          edu.lehigh.swat.hawk.core.Ontology ionto = imports[i];
          if(ionto==null)
            continue;
          StringWriter isw = new StringWriter();
          TellsDocument itd = writer.writeOntology(ionto, isw, kburi.toString());
          isw.flush();
          System.out.println(ionto.getID().getId());
          itd.save(new File("tellsDoc"+i+".xml"));  //test

          startTime = System.currentTimeMillis();
          rteller = new ReasonerTeller(r,itd);
          rteller.start();
//          endthis = new EndTask(rteller);
//          timer.schedule(endthis,to);
          rteller.join(to);
          duration = (System.currentTimeMillis()-startTime);
          System.out.println("Return from reasoner after "+duration);
          ResponseDocument ird = rteller.getResponse();
/*          if(ird!=null){
              System.out.println("response "+ i +" done! ");
              ird.save(new File("responseForTells" + i + ".xml")); //test
          }
*/
           if(ird==null){
  //          timer.cancel();
            r.releaseKB(kburi);
            return;
          }
//          endthis.cancel();
          IdRespType irt = ird.getResponse();
          if(irt.isSetOk())
            System.out.println("tells OK");
          else if(irt.isSetError()){
            System.out.println("tells error");
            ErrorDocument.Error er = irt.getError();
            System.out.println(er.getMessage());
             itd.save(new File("tellsDoc"+i+".xml"));
          }
        }
     }


      AsksDocument askDoc = AsksDocument.Factory.newInstance();
      askDoc.addNewAsks().setUri(kburi.toString());
      Ask askSat = askDoc.getAsks();
      OntClass[] classes = this.getAllPlusImportedClasses();
      int len =0;
      if(classes!=null)
        len = classes.length;
      for(int i=0;i<len;i++){
        OntClass cls = classes[i];
        String cid = cls.getID().getId();
        ConceptId conid = askSat.addNewSatisfiable();
        conid.setId("q"+i);
        Named named = conid.addNewCatom();
        named.setName(cid);
      }
      //askDoc.save(new File("asksat.xml"));
      startTime = System.currentTimeMillis();
      ReasonerAsker rasker = new ReasonerAsker(r,askDoc);
      rasker.start();
//      endthis = new EndTask(rasker);
//      timer.schedule(endthis,to);
      rasker.join(to);
      duration = (System.currentTimeMillis()-startTime);
      System.out.println("Return from reasoner after "+duration);
      ResponsesDocument rSat = rasker.getResponse();
      if(rSat==null){
//        timer.cancel();
        r.releaseKB(kburi);
        return;
      }
//      endthis.cancel();
      ResponseType rtype = rSat.getResponses();
      //rtype.save(new File("d:\\responseSat.xml"));
      IdType[] ftypes = rtype.getFalseArray();
      int flen = ftypes.length;
      for(int i=0;i<flen;i++){
        String seq = ftypes[i].getId().substring(1);
        int rseq = Integer.parseInt(seq);
        classes[rseq].setSatisfiable(new Boolean(false));
      }
      IdType[] ttypes = rtype.getTrueArray();
      int tlen = ttypes.length;
      for(int i=0;i<tlen;i++){
        String seq = ttypes[i].getId().substring(1);
        int rseq = Integer.parseInt(seq);
        classes[rseq].setSatisfiable(new Boolean(true));
      }

      AsksDocument askDoc2 = AsksDocument.Factory.newInstance();
      askDoc2.addNewAsks().setUri(kburi.toString());
      Ask askSup = askDoc2.getAsks();

      for(int i=0;i<len;i++){
        OntClass cls = classes[i];
        String cid = cls.getID().getId();
        ConceptId cpid = askSup.addNewParents();
        cpid.setId("p"+i);
        ConceptId ceid = askSup.addNewEquivalents();
        ceid.setId("e"+i);
        ConceptId cdid = askSup.addNewDescendants();
        cdid.setId("d"+i);
        ConceptId ccid = askSup.addNewChildren();
        ccid.setId("c"+i);
        Named named = cpid.addNewCatom();
        named.setName(cid);
        Named named2 = ceid.addNewCatom();
        named2.setName(cid);
        Named named3 = cdid.addNewCatom();
        named3.setName(cid);
        Named named4 = ccid.addNewCatom();
        named4.setName(cid);
      }

      //ResponsesDocument rsup = r.asks(askDoc2);
      startTime = System.currentTimeMillis();
      rasker = new ReasonerAsker(r,askDoc2);
      rasker.start();
      rasker.join(to);
      duration = (System.currentTimeMillis()-startTime);
      System.out.println("Return from reasoner after "+duration);
      ResponsesDocument rsup = rasker.getResponse();
      if(rsup==null){
        r.releaseKB(kburi);
        return;
      }
      ResponseType stype = rsup.getResponses();
      Csynonyms[] csys = stype.getConceptSetArray();

    //  rsup.save(new File("superResponse.xml"));
      int clen = csys.length;
      //System.out.println("Subs and supers: " + clen + " responses");
      //System.out.println(stype.sizeOfErrorArray() + " error responses");

      for(int i=0;i<clen;i++){
        Csynonyms csy = csys[i];
        String rid = csy.getId();
        String seq = rid.substring(1);
        int rseq = Integer.parseInt(seq);

        Concepts[] ca = csy.getSynonymsArray();
        int sylen = ca.length;

        if(rid.startsWith("p")){
//            System.out.println("super classes for "+classes[rseq].getID().getId()+": "+sylen);
   //       emptySupers(classes[rseq]); //comment out by zhp2 on 06/14/09
          for(int j=0;j<sylen;j++){
            Named[] cnames = ca[j].getCatomArray();
            int nlen = cnames.length;
            for(int k=0;k<nlen;k++){
              ObjectId newSuper = new ObjectId(cnames[k].getName());
              classes[rseq].addSuperClass(newSuper);
            }

/*        if(ca[j].sizeOfAndArray()>0){
                System.out.println("from reasoner: some values!");
                RoleConceptPair[] rcs = ca[j].getSomeArray();
                int rclen = rcs.length;
                for(int k=0;k<rclen;k++){
                    RoleConceptPair rc = rcs[k];
                    PropertyRestriction res = this.createRestriction();
                    ObjectId oid = new ObjectId("["+String.valueOf(Math.random())+"]");
                    res.setID(oid);
                    res.setOnProperty(new ObjectId(rc.getRatom().getName()));
                    res.setSomeValuesFrom(new ObjectId(rc.getCatom().getName()));
                    classes[rseq].addSuperClass(oid);
                }
            }
*/
          }
        }
        if(rid.startsWith("e")){
          for(int j=0;j<sylen;j++){
            Named[] cnames = ca[j].getCatomArray();
            int nlen = cnames.length;
            for(int k=0;k<nlen;k++){
              ObjectId newEqu = new ObjectId(cnames[k].getName());
              classes[rseq].addEquivalentClass(newEqu);
//              classes[rseq].addSuperClass(newEqu); //zhp2 added on 06/14/2009
//              classes[rseq].addSubClass(newEqu);  //zhp2 added on 06/14/2009
            }
          }
        }

        if(rid.startsWith("d")){
          for(int j=0;j<sylen;j++){
            Named[] cnames = ca[j].getCatomArray();
            int nlen = cnames.length;
            for(int k=0;k<nlen;k++){
              ObjectId newDes = new ObjectId(cnames[k].getName());
              ((edu.lehigh.swat.hawk.storage.simplememory.OntClass)classes[rseq]).addDescendant(newDes);
            }
          }
        }
      }


//properties
      OntProperty[] properties = this.getAllPlusImportedProperties();
      int plen =0;
      if(properties!=null)
         plen = properties.length;
      AsksDocument askDoc3 = AsksDocument.Factory.newInstance();
      askDoc3.addNewAsks().setUri(kburi.toString());
      Ask askSupp = askDoc3.getAsks();
      for(int i=0;i<plen;i++){
        OntProperty prop = properties[i];
        String cid = prop.getID().getId();
        RoleId cpid = askSupp.addNewRparents();
        cpid.setId("pp"+i);
//        ConceptId ceid = askSupp.addNewRequivalents();
//        ceid.setId("pe"+i);
        Named named = cpid.addNewRatom();
        named.setName(cid);
//        Named named2 = ceid.addNewRatom();
 //       named2.setName(cid);
      }

      //ResponsesDocument rsupp = r.asks(askDoc3);
      startTime = System.currentTimeMillis();
      rasker = new ReasonerAsker(r,askDoc3);
      rasker.start();
      rasker.join(to);
      duration = System.currentTimeMillis()-startTime;
      System.out.println("Return from reasoner after "+duration);
      ResponsesDocument rsupp = rasker.getResponse();
      if(rsupp==null){
        r.releaseKB(kburi);
        return;
      }
      ResponseType pstype = rsupp.getResponses();
      Rsynonyms[] pcsys = pstype.getRoleSetArray();
      int pclen = pcsys.length;
      for(int i=0;i<pclen;i++){
        Rsynonyms csy = pcsys[i];
        String rid = csy.getId();
        String seq = rid.substring(2);
        int rseq = Integer.parseInt(seq);

        Roles[] ca = csy.getSynonymsArray();
        int sylen = ca.length;
        if(rid.startsWith("pp")){
          emptySupers(properties[rseq]);
          for(int j=0;j<sylen;j++){
            Named[] cnames = ca[j].getRatomArray();
            int nlen = cnames.length;
            for(int k=0;k<nlen;k++){
              ObjectId newSuper = new ObjectId(cnames[k].getName());
              properties[rseq].addSuperProperty(newSuper);
            }
          }
        }
      }

//     timer.cancel();
     r.releaseKB(kburi);

    }
    catch(Exception e){
      System.out.println(e.getMessage());
//      e.printStackTrace();
    }
  }

  /**
   * Empty the superClasses.
   * @param c
   */
  private void emptySupers(OntObject x){
    if(x instanceof OntClass){
      OntClass c = (OntClass)x;
      ObjectId[] supers = c.getDirectSuperClasses();
      int len = supers.length;
      for(int i=0;i<len;i++){
        if(!supers[i].getLocalName().equals("")) //z.pan: be careful here.
          c.deleteSuperClass(supers[i]);
      }
    }
    if(x instanceof OntProperty){
      OntProperty p = (OntProperty)x;
      ObjectId[] supers = p.getDirectSuperProperties();
      int len = supers.length;
      for(int i=0;i<len;i++){
        if(!supers[i].getLocalName().equals("")) //z.pan: be careful here.
          p.deleteSuperProperty(supers[i]);
      }
    }
  }

}
