/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/

package edu.lehigh.swat.hawk.storage.simpledb;

/**
 * A logical atom. An atom consists of a predicate and a set of
 * arguments. The arguments may be variables or literals. If they
 * are variables, then they are preceded by a question mark.
 * @author Z. Pan
 * @version 1.0
 */

public class Atom {

  String predicate;
  String[] arguments;

  /* Constructor for an atom with an arbitrary number of arguments. */
  public Atom(String predicate, String[] arguments) {

    this.predicate = predicate;
    this.arguments = arguments;
  }

  /** A shorthand constructor for an atom with a binary predicate. */
  public Atom(String predicate, String arg1, String arg2) {

    this.predicate = predicate;
    arguments = new String[2];
    arguments[0] = arg1;
    arguments[1] = arg2;
  }

  /** A shorthand constructor for an atom with a unary predicate. */
  public Atom(String predicate, String arg) {

    this.predicate = predicate;
    arguments = new String[1];
    arguments[0] = arg;
  }

  /** Return the name of the atom's predicate.*/
  public String getPredicate() {

    return predicate;
  }

  /** Return an array of the atom's arguments. */
  public String[] getArguments() {

    return arguments;
  }

  public String toString() {

    StringBuffer sbuf;
    int i;

    sbuf = new StringBuffer();
    sbuf.append("(" + predicate);
    for (i=0; i < arguments.length; i++)
      sbuf.append(" " + arguments[i]);
    sbuf.append(")");
    return sbuf.toString();
  }

}
