/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/

package edu.lehigh.swat.hawk.storage.simpledb;

/**
 * Database storage model.
 * @author Z. Pan
 * @version 1.0
 */

import java.io.*;
import java.sql.*;
import java.util.*;

import edu.lehigh.swat.hawk.core.DocId;
import edu.lehigh.swat.hawk.core.ObjectId;
import edu.lehigh.swat.hawk.storage.*;
import edu.lehigh.swat.hawk.fol.Atom;
import edu.lehigh.swat.hawk.helpers.SparqlQueryDB;
//import edu.lehigh.swat.hawk.storage.simplememory.Document;
//import edu.lehigh.swat.hawk.core.Ontology;


public class DatabaseStorage extends PersistentStorageModel {//implements QueryableStorageModel{

  static final String WRITE_OBJECT_SQL = "INSERT INTO ontology_depot(id, object) VALUES (?, ?)";

  static final String READ_OBJECT_SQL = "SELECT object FROM ontology_depot WHERE id = ?";

  /**
   * Constructor.
   * @param DBName
   */
  public DatabaseStorage(String DBName) {
    dbName = DBName;
    queryOnto = "";
  }

  /**
   * Open a connection to the database.
   */
  public void open(){
    String db_prefix = "";
    String fname = "";
    String[] login = {"",""};
    try{
      if(dbName.endsWith(".mdb")){
        java.lang.Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
        db_prefix = "jdbc:odbc:MS Access Database;DBQ=";
        emptyTableState = "S0002";
        fname = dbName;
      }
      else if(dbName.indexOf("postgresql")>0){
        java.lang.Class.forName("org.postgresql.Driver");
        emptyTableState = "42P01";
        String[] names = dbName.split("@");
        fname = names[1];
        login = names[0].split(":",2);
        System.out.println("postgresql!");
      }else if(dbName.indexOf("mysql")>0){
        Class.forName("com.mysql.jdbc.Driver").newInstance();
        emptyTableState = "42S02";
        String[] names = dbName.split("@");
        fname = names[1]+"?autoReconnect=true";
        login = names[0].split(":",2);
        System.out.println("mysql!");
      }else if(dbName.indexOf("jdbc:db2:")>0){
        Class.forName("com.ibm.db2.jcc.DB2Driver").newInstance();
        emptyTableState = "42704";
        String[] names = dbName.split("@");
        fname = names[1];
        login = names[0].split(":",2);
        System.out.println("db2!");
     } else if(dbName.indexOf("sqlite")>0){
     //   System.out.println("new SQLITE package name.!!!");
      Class.forName("SQLite.JDBCDriver").newInstance(); // Class.forName("org.sqlite.JDBC").newInstance();
      emptyTableState = "S0002"; //used to determine if db is empty
      String[] names = dbName.split("@");
      fname = names[1];
      login[0] = "";
      login[1] = "";
      System.out.println("SQLite!");
    }
      else if(dbName.indexOf("hsqldb")>0){
        Class.forName("org.hsqldb.jdbcDriver").newInstance();
        emptyTableState = "S0002"; //used to determine if db is empty
        String[] names = dbName.split("@");
        fname = names[1];
        //login = names[0].split(":",2);
        login[0] = "sa";
        login[1] = "";
        System.out.println("hsqldb!");
      }else if(dbName.indexOf("h2")>0){
        Class.forName("org.h2.Driver").newInstance();
        emptyTableState = "42S02"; //used to determine if db is empty
        String[] names = dbName.split("@");
        fname = names[1];
        //login = names[0].split(":",2);
        login[0] = "sa";
        login[1] = "";
        System.out.println("h2!");
    }else if(dbName.indexOf("derby")>0){
        Class.forName("org.apache.derby.jdbc.EmbeddedDriver").newInstance();
        emptyTableState = "42X05"; //used to determine if db is empty
        String[] names = dbName.split("@");
        fname = names[1];
        System.out.println("derby!");
    }


     }
    catch(Exception e){
      if(e instanceof java.lang.ClassNotFoundException)
        System.out.print("ClassNotFoundException: ");
      System.out.println(e.getMessage());
    }

    try{
      String s_open = db_prefix+fname;
      System.out.println(s_open);
      con = java.sql.DriverManager.getConnection(s_open,login[0],login[1]);
    }
    catch(SQLException e){
      System.out.println("error when connecting database");
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
      stmt = con.createStatement(java.sql.ResultSet.TYPE_FORWARD_ONLY,java.sql.ResultSet.CONCUR_READ_ONLY);
     }
    catch(SQLException e){
      System.out.println("error_in_stmt");
    }
/*
    if(dbName.indexOf("mysql")>0){
        try {
            if(stmt.execute("SET SQL_LOG_BIN = 0;"))
                System.out.println("mysql bin log off!");
            else
                System.out.println("can't turn off mysql bin log!");
        } catch (SQLException e) {
            System.out.println(e.getMessage());
        }
    }
*/
    knownOntos = getKnownOntologies();
    if(!queryOnly)
       getSourceNum();
    getKnownUrls();
//    getPropertyTypes();
    cachedDocs = new Hashtable();
    loadedDocs = new Hashtable();
    loadedTripleNum = 0;
  }

  /**
   * open kb with arguments
   */
  public void open(boolean readOnly){
     queryOnly = readOnly;
     open();
  }

  /**
   * Close the database connection.
   */
  public void close(){
    try{
      flushLoadedDocs(true);
      stmt.close();
      con.close();
      }
    catch(SQLException e){
      System.out.println("error when close kb");
    }
  }

  /**
   * Get the sql statement.
   */
  public Statement getStatement(){
    return stmt;
  }

   protected void flushLoadedDocs(boolean doneLoading){
/*     int slash = dbName.lastIndexOf('\\');
     int sat = dbName.indexOf('@');
     String dbpath = "";
     if(slash>0){
       dbpath = dbName.substring(sat+1,slash+1);
     }
*/
     long t0 = System.currentTimeMillis();
     try{

         Vector iontos = new Vector();
         iontos.addAll(cachedDocs.keySet()); //z.pan: 3/8/07
         Iterator it = loadedDocs.values().iterator();
         while (it.hasNext()) {
             Document doc = (Document) it.next();
             DocId did = doc.getID();
             if (doc instanceof Ontology && !iontos.contains(did))
                 iontos.add(did);
             //     doc.flush();
             DocId[] dids = doc.getImports();
             if (dids != null) {
                 int dlen = dids.length;
                 for (int i = 0; i < dlen; i++) {
                     if (!iontos.contains(dids[i])&&knownOntos.contains(dids[i].getId()))
                         iontos.add(dids[i]);
                 }
             }
             knownDocs.put(doc.getID().getId(), new Integer(doc.getDocSeq()));
         }

    if(doneLoading){
         Hashtable<Integer, DocId> flushOntos = new Hashtable();
         int[] seqs = new int[iontos.size()];
         int is = iontos.size();
         for(int i=0; i<is; i++){
             DocId did = (DocId)iontos.get(i);
             Integer seq = (Integer)knownOntos.get(did.getId());
             if(seq!=null&&seq>0){
                 seqs[i] = seq.intValue();
                 flushOntos.put(seq, did);
             }
         }

         Arrays.sort(seqs);

     //    long t0 = System.currentTimeMillis();
         is = seqs.length;
         for(int j=0;j<is;j++) {
             if(seqs[j]<1)
                 continue;
             DocId DId= flushOntos.get(seqs[j]);
             if(DId==null)
                 continue;
             String did = DId.getId();
             Document idoc = (Document)getDocument(did);
             if(!(idoc instanceof Ontology))
                 continue;
             Ontology onto = (Ontology) idoc;
     //System.out.println(onto.getID().getId());

             edu.lehigh.swat.hawk.core.OntProperty[] props = onto.
                     getAllPlusImportedProperties();
             int len = props.length;
             for (int i = 0; i < len; i++) {
                 edu.lehigh.swat.hawk.core.OntProperty prop = props[i];
     //System.out.println(prop.getID().getId());
                 String tableName = getTableName(prop.getID());
                 if (tableName == null)
                     continue;
                 tableName = tableName + "_p";
     //System.out.println(tableName);
                 String viewName = tableName+"_view";
                 if(this instanceof edu.lehigh.swat.hawk.storage.ontper.dldb.DldbStorage)
                     viewName = tableName+"_"+onto.getOntSeq()+"_view";
                 if (prop.isTransitive()) {
                	 if(prop.isSymmetric()){
                		 String qs = "insert ignore into "+tableName+" select object, subject, source, "+onto.getOntSeq()+" from "+tableName;
                         try {
                             long t00 = System.currentTimeMillis();
                             stmt.execute(qs);
                             System.out.println("done symmetric used: " +
                                                (System.currentTimeMillis() - t00));
                         } catch (SQLException e) {
                             System.out.println("error when do symmetric");
                             System.out.println(qs);
                             System.out.println("SQLState:" + e.getSQLState());
                             System.out.println("Message :" + e.getMessage());
                             System.out.println("Vendor  :" + e.getErrorCode());
                         }
                	 }
                     String qs = "call computeTC('" + tableName + "')";
                     if(this instanceof edu.lehigh.swat.hawk.storage.ontper.dldb.DldbStorage)
                         qs = "call computeTCp('" + tableName + "'," +
                                 onto.getOntSeq() + ")";
                     System.out.println(qs);
                     try {
                         long t00 = System.currentTimeMillis();
                         stmt.execute(qs);
                         System.out.println("computeTC used: " +
                                            (System.currentTimeMillis() - t00));
                     } catch (SQLException e) {
                         System.out.println("error when call computeTC");
                         System.out.println(qs);
                         System.out.println("SQLState:" + e.getSQLState());
                         System.out.println("Message :" + e.getMessage());
                         System.out.println("Vendor  :" + e.getErrorCode());
                     }
                 }
                 if (prop.isInverseFunctional()) {
                     String qs = "call findSameAs('" + viewName+"')";
                     if(prop.isDatatype()==null||prop.isDatatype().booleanValue())
                         continue;
                     System.out.println(qs);
                     try {
                         long t00 = System.currentTimeMillis();
                         stmt.execute(qs);
                         System.out.println("findSameAs used: " +
                                            (System.currentTimeMillis() - t00));
                     } catch (SQLException e) {
                         System.out.println("error when call findSameAs");
                         System.out.println(qs);
                         System.out.println("SQLState:" + e.getSQLState());
                         System.out.println("Message :" + e.getMessage());
                         System.out.println("Vendor  :" + e.getErrorCode());
                     }
                 }
                 if (prop.isFunctional()) {
                     String qs = "call findSameAs2('" + viewName + "')";
                     if(prop.isDatatype()==null||prop.isDatatype().booleanValue())
                         continue;
                     System.out.println(qs);
                     try {
                         long t00 = System.currentTimeMillis();
                         stmt.execute(qs);
                         System.out.println("findSameAs used: " +
                                            (System.currentTimeMillis() - t00));
                     } catch (SQLException e) {
                         System.out.println("error when call findSameAs2");
                         System.out.println(qs);
                         System.out.println("SQLState:" + e.getSQLState());
                         System.out.println("Message :" + e.getMessage());
                         System.out.println("Vendor  :" + e.getErrorCode());
                     }
                 }

                 if(prop.getDomain()!=null||prop.getRange()!=null){
                     long t00 = System.currentTimeMillis();
                     String tempProp = "CREATE TEMPORARY TABLE IF NOT EXISTS " +
                                       viewName
                                       + " (subject int, object int, source int, onto int, index(subject), index(object))"
                                       + " select * from " + viewName;
                     try {
                         stmt.execute(tempProp);
                     } catch (SQLException e) {
                     }

                     //domain
                     if (prop.getDomain() != null) {
                         ObjectId domainId = prop.getDomain();
                         String tname = getTableName(domainId);
                         if (tname != null) {
                             String vname = tname+"_view";
                             if(this instanceof edu.lehigh.swat.hawk.storage.ontper.dldb.DldbStorage)
                                 vname = tname + "_" + onto.getOntSeq() +"_view";
                             String tempClass =
                                     "CREATE TEMPORARY TABLE IF NOT EXISTS " +
                                     vname
                                     +
                                     " (id int, source int, onto int, index(id))"
                                     + " select * from " + vname;
                             String qs = "insert ignore into " + tname +
                                         " select subject, source,"
                                         + onto.getOntSeq() + " from " +
                                         viewName
                                         +
                                     " where subject not in (select id from " +
                                         vname + ")";
                             try {
                                 stmt.execute(tempClass);
                             } catch (SQLException e) {
                             }

                             try {
                                 stmt.execute(qs);
                             } catch (SQLException e) {
                                 System.out.println("error when do domain");
                                 System.out.println(qs);
                                 System.out.println("SQLState:" + e.getSQLState());
                                 System.out.println("Message :" + e.getMessage());
                                 System.out.println("Vendor  :" +
                                         e.getErrorCode());
                             }
                         }
                     }

                     //range
                     if (prop.getRange() != null) {
                         ObjectId rangeId = prop.getRange();
                         String tname = getTableName(rangeId);
                         if (tname != null) {
                             String vname = tname+"_view";
                             if(this instanceof edu.lehigh.swat.hawk.storage.ontper.dldb.DldbStorage)
                                 vname = tname + "_" + onto.getOntSeq() +
                                            "_view";
                             String tempClass =
                                     "CREATE TEMPORARY TABLE IF NOT EXISTS " +
                                     vname
                                     +
                                     " (id int, source int, onto int, index(id))"
                                     + " select * from " + vname;
                             String qs = "insert ignore into " + tname +
                                         " select object, source,"
                                         + onto.getOntSeq() + " from " +
                                         viewName
                                         +
                                     " where object not in (select id from " +
                                         vname + ")";
                             try {
                                 stmt.execute(tempClass);
                             } catch (SQLException e) {
                             }

                             try {
                                 stmt.execute(qs);
                             } catch (SQLException e) {
                                 System.out.println("error when do range");
                                 System.out.println(qs);
                                 System.out.println("SQLState:" + e.getSQLState());
                                 System.out.println("Message :" + e.getMessage());
                                 System.out.println("Vendor  :" +
                                         e.getErrorCode());
                             }
                         }
                     }
                     System.out.println("Domain and Range used: " +
                                         (System.currentTimeMillis() - t00));
                 }
             }

             Hashtable<String, Vector<Vector<Atom>>> ccr = onto.getCyclicRules();
             if(ccr == null)
            	 continue;
             boolean fixPoint = false;
             System.out.println("Entering fixpoint...");
	         while(!fixPoint){
	        	 fixPoint = true;
	             Enumeration<String> ecc =  ccr.keys();
	             while(ecc.hasMoreElements()){
	            	 String cc = ecc.nextElement();
	            	 ObjectId cid = new ObjectId(cc);
	            	 String tname = getTableName(cid);
	            	 if(tname==null)
	            		continue;
	            	 String vname = tname+"_view";
	            	 String cvname = tname+"_cycle_view";
                     if(this instanceof edu.lehigh.swat.hawk.storage.ontper.dldb.DldbStorage){
                         vname = tname+"_"+onto.getOntSeq()+"_view";
                         cvname = tname+"_"+onto.getOntSeq()+"_cycle_view";
                     }
	            	 int before = 0;
	            	 int after = 0;

                     System.out.println("compute "+vname);
	            	 try{
	            		String qs = "select count(0) from "+vname;
	            		java.sql.ResultSet rs = stmt.executeQuery(qs);
                        if(rs.next())
                            before = rs.getInt(1);
	            		String qs2 = "insert ignore into "+tname+" select * from "+cvname;
	            		stmt.executeUpdate(qs2);
	            		rs = stmt.executeQuery(qs);
                        if(rs.next())
	            		    after = rs.getInt(1);
	            	 }catch(Exception e){
	            		 System.out.println(e.getMessage());
	            	 }
	            	 if(after>before){
	            		 fixPoint = false;
                         System.out.println(after+">"+before);
	            	 }
	             }
             }
         }
    } //end if
    // stmt.execute("flush tables;");
     }catch(Exception e)
     {
         System.out.println(e.getMessage());
         e.printStackTrace();
     }
    System.out.println("time in flushing: "+ (System.currentTimeMillis()-t0));
    loadedDocs.clear();
    loadedTripleNum = 0;
    knownUrls.clear(); //zhp2 on 7/16/08

   }

   public void storeOntology(Ontology ontd){
      // System.out.println("storing ontology...");
       if(!serializeOnto)
           return;
//         if(!iontos.contains(doc.getID()))
//              iontos.add(doc.getID());
         // Ontology ontd = (Ontology)doc;
//        String sname = dbpath+"onto"+ontd.getOntSeq()+".dat";
         try{
       //   FileOutputStream out = new FileOutputStream(sname);
       //   ObjectOutputStream s = new ObjectOutputStream(out);
           edu.lehigh.swat.hawk.storage.simplememory.Ontology monto = ontd.getMemOnto();
           monto.deleteAllIndividual();  //zhp2: we don't want serialization to include individuals.
       //   s.writeObject(monto);
       //   s.flush();
       //   s.close();
       //   out.close();
           PreparedStatement pstmt = con.prepareStatement(WRITE_OBJECT_SQL);
           pstmt.setInt(1, ontd.getOntSeq());

           ByteArrayOutputStream baos = new ByteArrayOutputStream();
           ObjectOutputStream oos = new ObjectOutputStream(baos);
           oos.writeObject(monto);
           oos.flush();
 //          byte[] by = baos.toByteArray();
           pstmt.setBytes(2, baos.toByteArray());
 //          System.out.println(by.length);
           pstmt.executeUpdate();
           oos.close();
           baos.close();
        }catch(Exception e){
          System.out.println(e.getMessage());
          e.printStackTrace();
        }
 //       knownOntos.put(ontd.getID().getId(),new Integer(ontd.getOntSeq()));
   }

  /**
   * Add a document to this model.
   */
  public void addDocument(edu.lehigh.swat.hawk.core.Document d){
    if(d.getID()==null){
      System.out.println("Error: ID not found in a document");
      return;
    }
    loadedDocs.remove(d.getID());
    loadedDocs.put(d.getID(),d);
    int size = d.getSize();
    loadedTripleNum += size;
//    System.out.println("sizeeeeeee "+loadedTripleNum+"   Memmmm "+Runtime.getRuntime().freeMemory());
    if(loadedTripleNum>1500000 && (Runtime.getRuntime().freeMemory()<500000000)){
        System.out.println("flush loaded docs!");
        flushLoadedDocs(false);
    }
    if(d instanceof Ontology){
      Ontology ontd = (Ontology)d;
      updateOntIndex(ontd, ontd.getOntSeq());
      updateSourceIndex(ontd, ontd.getDocSeq());
      cachedDocs.put(d.getID(), ontd); //zhp2: new as of 5/15/08
    }
    else{
      Document docd = (Document)d;
      updateSourceIndex(docd, docd.getDocSeq());
    }
  }

  /**
   * Get the specified document.
   */
  public edu.lehigh.swat.hawk.core.Document getDocument(String url){
    Document doc = null;
    DocId did = new DocId(url);
    if(loadedDocs.containsKey(did))
      return (edu.lehigh.swat.hawk.core.Document)loadedDocs.get(did);
    if(cachedDocs.containsKey(did))
      return (edu.lehigh.swat.hawk.core.Document)cachedDocs.get(did);
    if(knownOntos.containsKey(url)){
       Integer i = (Integer)knownOntos.get(url);
       int seq = i.intValue();
  /*     int slash = dbName.lastIndexOf('\\');
       String dbpath = "";
       if(slash>0){
        dbpath = dbName.substring(0,slash+1);
       }
       String sname = dbpath+"onto"+seq+".dat";
   */
       try{
         /*
         FileInputStream ins = new FileInputStream(sname);
         ObjectInputStream ois = new ObjectInputStream(ins);
         edu.lehigh.swat.hawk.storage.simplememory.Ontology mOnto =
             (edu.lehigh.swat.hawk.storage.simplememory.Ontology) ois.readObject();
         ois.close();
          */
         PreparedStatement pstmt = con.prepareStatement(READ_OBJECT_SQL);
         pstmt.setInt(1, seq);
         java.sql.ResultSet rs = pstmt.executeQuery();

         rs.next();
      //   edu.lehigh.swat.hawk.storage.simplememory.Ontology mOnto
      //     = (edu.lehigh.swat.hawk.storage.simplememory.Ontology)
         Blob blob;
         blob = rs.getBlob(1);
     //    int len = bytes.length;
     //    System.out.println("length:"+len);
        // byte[] by = blob.getBytes(1, (int) blob.length());
         BufferedInputStream bis = new BufferedInputStream(blob.getBinaryStream());
         //ByteArrayInputStream bai = new ByteArrayInputStream(by);
         ObjectInputStream ois = new ObjectInputStream(bis);
//         System.out.println(blob.length());
         Object obj = ois.readObject();
         edu.lehigh.swat.hawk.storage.simplememory.Ontology mOnto = null;
         if(obj instanceof edu.lehigh.swat.hawk.storage.dlmemory.Ontology){
            mOnto = (edu.lehigh.swat.hawk.storage.dlmemory.Ontology) obj;
         }else
             mOnto = (edu.lehigh.swat.hawk.storage.simplememory.Ontology) obj;
         bis.close();
         ois.close();

    //    String className = object.getClass().getName();
     //   System.out.println("object is an instance of "+className);
         Ontology ont = new Ontology(mOnto);
         ont.setID(mOnto.getID());
         ont.setOntSeq(seq);
         doc = ont;
         cachedDocs.put(doc.getID(), doc);
       }
       catch(Exception e){
           System.out.println("error when retrieve ontology object");
           if(e instanceof SQLException){
               SQLException se = (SQLException)e;
               System.out.println("SQLState:" + se.getSQLState());
               System.out.println("Message :" + se.getMessage());
               System.out.println("Vendor  :" + se.getErrorCode());
           }
           System.out.println(e.getMessage());
           e.printStackTrace();
         Ontology ont2 = new Ontology(new edu.lehigh.swat.hawk.storage.simplememory.Ontology());
         ont2.setID(did);
         ont2.setOntSeq(seq);
         return ont2;
       }return doc;
    }
    if(knownDocs.containsKey(url)){
      doc = new Document();
      doc.setID(did);
      int seq = ((Integer)knownDocs.get(url)).intValue();
      doc.setDocSeq(seq);
      doc.setModel(this);
      doc.setSQLStatement(stmt);
      return doc;
    }
    return null;
  }


  /**
   * Update a document.
   */
  public void updateDocument(edu.lehigh.swat.hawk.core.Document d){

  }

  /**
   * Get all the documents in the model.
   */
  public Iterator getDocumentIterator(){
    return null;
  }

  /**
   * Set whether or not include implictly imported documents.
   */
  public void setImplictImports(boolean imp){

  }


  /**
   * Create an Ontology object.
   */
  public edu.lehigh.swat.hawk.core.Ontology createOntology(){
    edu.lehigh.swat.hawk.storage.simplememory.Ontology mOnto =
        new edu.lehigh.swat.hawk.storage.simplememory.Ontology();
    Ontology onto = new Ontology(mOnto);
    onto.setDocSeq(++sourceSeqNum);
    onto.setOntSeq(++ontSeqNum);
    onto.setSQLStatement(stmt);
    onto.setModel(this);
    return onto;
  }

  /**
   * Create an Document object.
   */
  public edu.lehigh.swat.hawk.core.Document createDocument(){
    Document doc = new Document();
    doc.setDocSeq(++sourceSeqNum);
    doc.setSQLStatement(stmt);
    doc.setModel(this);
    return doc;
  }

  /**
   *retrieval ontologies already exist in database
   */
  public Hashtable getKnownOntologies()
  {
    int seqNum = 0;
    String ontoUrl;
    String qs = "SELECT * \nFROM Ontologies_index";
    String qs2 = "SELECT MAX(seqnum) FROM Ontologies_index";
    Hashtable knownOntologies = new Hashtable();
    String errState = "";
    java.sql.ResultSet rs = null;

    try{
        System.out.println("Access Ontologies_index..");
        rs = stmt.executeQuery(qs);
     }
    catch(SQLException e){
      //errcode = e.getErrorCode();
     System.out.println("Access Ontologies_index...  SQLState:"+e.getSQLState());
       errState = e.getSQLState();
       if(errState==null||errState.equals(emptyTableState)){
         createKb();
         ontSeqNum = 0;
         return knownOntologies;
       }else{
        System.out.println("error");
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
    }
    System.out.println("Populate knownOntos..");
    try{
      while (rs.next()){
        ontoUrl = rs.getString("url");
        seqNum = rs.getInt("seqnum");
        knownOntologies.put(ontoUrl, new Integer(seqNum));
      }
      rs.close();
      rs = stmt.executeQuery(qs2);
      if(rs.next())
        ontSeqNum = rs.getInt(1);
      rs.close();
    }

    catch (SQLException e){
        System.out.println(e.getMessage());
     }

    knownOntos = knownOntologies;
    return knownOntologies;
  }


  /**
   * Get the maximum source sequence number in the databade.
   */
  public void getSourceNum()
  {
    knownDocs = new Hashtable();
    int seqNum = 0;
    String qs = "SELECT * \nFROM source_index";
    String qs2 = "SELECT MAX(seqnum) FROM source_index";
    String errState = "";
    java.sql.ResultSet rs = null;
    try{
      rs = stmt.executeQuery(qs);
     }
    catch(SQLException e){
      System.out.println("error");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
      errState = e.getSQLState();
    }
    if (errState.equals(emptyTableState)){
          createKb();
          sourceSeqNum = 0;
          return;
    }
    try{
        System.out.println("Populate knownSources..");
      while (rs.next()){
        String url = rs.getString("url");
        seqNum = rs.getInt("seqnum");
        knownDocs.put(url,new Integer(seqNum));
      }
      rs.close();
      rs = stmt.executeQuery(qs2);
      if(rs.next())
        sourceSeqNum = rs.getInt(1);
      rs.close();
    }catch (SQLException e){
        System.out.println(e.getMessage());
    }
  }


  /**
   * Initialize the database. Create the index tables and utility tables.
   */
  public void createKb(){
    try{
      stmt.executeUpdate("CREATE TABLE Ontologies_index ("+
                               "url VARCHAR (255),"+
                               "SEQNUM INTEGER not null,"+
                               "PRIMARY KEY (SEQNUM)"+
                               ")");
    } catch(SQLException e){
            System.out.println("error when creat table 'Ontologies_index'");
            System.out.println("SQLState:"+e.getSQLState());
            System.out.println("Message :"+e.getMessage());
            System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
    stmt.executeUpdate("CREATE TABLE source_index ("+
                               "url VARCHAR (255),"+
                               "SEQNUM INTEGER not null,"+
                               "PRIMARY KEY (SEQNUM)"+
                               ")");
    }catch(SQLException e){
      System.out.println("error when creat table 'source_index'");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
      stmt.executeUpdate("CREATE TABLE uri_index ("+
                         "uri VARCHAR (255) not null,"+
                         "ID INTEGER,"+
                         "PRIMARY KEY (uri)"+
                               ")");
    }catch(SQLException e){
      System.out.println("error when creat table 'uri_index'");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
      stmt.executeUpdate("CREATE index uriIndex ON uri_index (ID)");
    } catch(SQLException e){
      System.out.println("error when index url");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
      stmt.executeUpdate("CREATE TABLE propertytypes ("+
                                  "pname VARCHAR (255) not null,"+
                                  "type VARCHAR (20),"+
                                  "PRIMARY KEY (pname))");
      stmt.executeUpdate("INSERT INTO propertytypes values ('label_rdfs_p', 'datatype')");
    }catch(SQLException e){
      System.out.println("error when create propertytype table");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
    stmt.executeUpdate("CREATE TABLE thing_owl ("+
                       "ID VARCHAR (255) not null,"+
                       "source INTEGER, onto INTEGER, PRIMARY KEY (ID))");
    }catch(SQLException e){
     System.out.println("error when create thing_owl table");
     System.out.println("SQLState:"+e.getSQLState());
     System.out.println("Message :"+e.getMessage());
     System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
    stmt.executeUpdate("CREATE TABLE label_rdfs_p ("+
                       "subject INTEGER not null, object VARCHAR(255) not null,"+
                       "source INTEGER not null, onto INTEGER, PRIMARY KEY (subject, object, source))");
    }catch(SQLException e){
     System.out.println("error when create label_rdfs table");
     System.out.println("SQLState:"+e.getSQLState());
     System.out.println("Message :"+e.getMessage());
     System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
        if(dbName.endsWith(".mdb")){
            stmt.executeUpdate("CREATE TABLE tables_index ("+
                       "tablename VARCHAR (255) not null,"+
                       "id real, PRIMARY KEY (tablename))");
        }
        else
            stmt.executeUpdate("CREATE TABLE tables_index ("+
                       "tablename VARCHAR (255) not null,"+
                       "ontid SMALLINT, termid SMALLINT, PRIMARY KEY (tablename))");
    }catch(SQLException e){
     System.out.println("error when create tables_index table");
     System.out.println("SQLState:"+e.getSQLState());
     System.out.println("Message :"+e.getMessage());
     System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
        if(dbName.endsWith(".mdb")){
            stmt.executeUpdate("CREATE TABLE individual_occurrence ("+
                            "ID INTEGER, "+
                            "INTABLE real, PRIMARY KEY (ID, INTABLE))");
        }else
        stmt.executeUpdate("CREATE TABLE individual_occurrence ("+
                           "ID INTEGER not null, "+
                           "ontid SMALLINT not null, termid SMALLINT not null, position SMALLINT not null, PRIMARY KEY (ID, ontid, termid,position))");
        }catch(SQLException e){
         System.out.println("error when create individual_occurrence table");
         System.out.println("SQLState:"+e.getSQLState());
         System.out.println("Message :"+e.getMessage());
         System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
    if(dbName.indexOf("mysql")>0)
      stmt.executeUpdate("CREATE TABLE ontology_depot ("+
                                  "id int not null,"+
                                  "object BLOB(524287),"+  //zhp2: mysql default length was 65535.
                                  "PRIMARY KEY (id))");
    else
        stmt.executeUpdate("CREATE TABLE ontology_depot ("+
                             "id int not null,"+
                             "object BLOB,"+
                             "PRIMARY KEY (id))");

    }catch(SQLException e){
      System.out.println("error when create ontology_depot table");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
    }

    try{
    if(dbName.endsWith(".mdb")){
      stmt.executeUpdate("CREATE TABLE view_definition ("+
                       "viewname VARCHAR (255),"+
                       "PRIMARY KEY (viewname),"+
           "definition MEMO)");

  }else if(dbName.indexOf("hsqldb")>0){
        stmt.executeUpdate("CREATE TABLE view_definition ("+
                "viewname VARCHAR (255),"+
                "definition LONGVARCHAR,"+
                "PRIMARY KEY (viewname));");
    }
  else if(dbName.indexOf("derby") >0){
      stmt.executeUpdate("CREATE TABLE view_definition (" +
                         "viewname VARCHAR (255)," + //view is apparently a key word in derby, will prob
                         "PRIMARY KEY (viewname)," + //need to fix elsewhere
                         "definition LONG VARCHAR)");
  }else if(dbName.indexOf("jdbc:db2:")>0){
      stmt.executeUpdate("CREATE TABLE view_definition ("+
                       "viewname VARCHAR (255) not null,"+
                       "definition CLOB (4096),"+
           "PRIMARY KEY (viewname))");
  }else
      stmt.executeUpdate("CREATE TABLE view_definition ("+
                       "viewname VARCHAR (255) not null,"+
                       "definition TEXT,"+
           "PRIMARY KEY (viewname))");

/*  stmt.executeUpdate("CREATE TABLE META_DATA("+
                  "tablename varchar (255),"+
                  "prop_id varchar (255),"+
                  "PRIMARY KEY (tablename,prop_id))"
                 // +  "KEY(tablename))"
          );
*/

  String sp = "CREATE PROCEDURE computeTC(IN tableName varchar(255))"
  +"\nbeg:BEGIN"
  +  "\nDECLARE num1 int default 0;"
  +  "\nDECLARE num2 int default 0;"
  +  "\nset @tempTable := concat(tableName,'temp');"
  +  "\nset @deltaTable := concat(tableName,'delta');"
  +  "\nset @viewName := concat(tableName,'_view');"
  +  "\nset @viewName2 := concat(tableName,'_view2');"

  +     "\nset @sql_text:=concat('SELECT count(*) INTO @num1 FROM ',@viewName);"
  +     "\nprepare stmt from @sql_text;"
  +     "\nexecute stmt;  "
//  +     "\nSELECT @num1;"
  +     "\nIF @num1>300000 THEN LEAVE beg;"
  +      "END IF;"

  +  "\nset @sql_text:=concat('DROP TEMPORARY TABLE IF EXISTS ',@tempTable);"
  +  "\nprepare stmt from @sql_text;"
  +  "\nexecute stmt;"

  +  "\nset @sql_text:=concat('DROP TEMPORARY TABLE IF EXISTS ',@viewName);"
  +  "\nprepare stmt from @sql_text;"
  +  "\nexecute stmt;"

  +  "\nset @sql_text:=concat('DROP TEMPORARY TABLE IF EXISTS ',@viewName2);"
  +  "\nprepare stmt from @sql_text;"
  +  "\nexecute stmt;"

   + "\nset @sql_text:=concat('DROP TABLE IF EXISTS ',@deltaTable);"
   + "\nprepare stmt from @sql_text;"
   + "\nexecute stmt;  "

   + "\nset @sql_text:=concat('CREATE TEMPORARY TABLE ',@tempTable, '(subject int, object int, source int, onto int, primary key(subject, object), index(object))');"
   + "\nprepare stmt from @sql_text;"
   + "\nexecute stmt; "

   + "\nset @sql_text:=concat('CREATE TABLE ',@deltaTable, '(subject int, object int, source int, onto int, primary key(subject, object), index(object))');"
   + "\nprepare stmt from @sql_text;"
   + "\nexecute stmt;        "

   + "\nset @sql_text:=concat('CREATE TEMPORARY TABLE IF NOT EXISTS ',@viewName2, '(subject int, object int, source int, onto int, index(subject), index(object)) select * from ', @viewName);"
   + "\nprepare stmt from @sql_text;"
   + "\nexecute stmt; "

   + "\nset @sql_text:=concat('CREATE TEMPORARY TABLE IF NOT EXISTS ',@viewName, '(subject int, object int, source int, onto int, index(subject), index(object)) select * from ', @viewName);"
   + "\nprepare stmt from @sql_text;"
   + "\nexecute stmt; "

   + "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p1.subject, p2.object, 0, 0'"
   +"\n' from ',@viewName,' as p1,',@viewName2,' as p2 ',"
   + "\n'where p1.object = p2.subject');"
   +"\nprepare stmt from @sql_text;"
   +"\nexecute stmt; "

   + "\nset @sql_text:=concat('insert ignore into ',@deltaTable, ' select * from ',@tempTable,"
   + "\n ' where not exists (select * from ',@viewName, ' where ',@tempTable,'.subject=',@viewName,'.subject and ',@tempTable,'.object=',@viewName,'.object)');"
   + "\nprepare stmt from @sql_text;"
   + "\nexecute stmt;          "

   + "\nl1:LOOP"
   +     "\nIF @num2>0 THEN set @num1 = @num2;"
   +     "\nELSE"
   +     "\nset @sql_text:=concat('SELECT count(*) INTO @num1 FROM ',@deltaTable);"
   +     "\nprepare stmt from @sql_text;"
   +     "\nexecute stmt;   "
   +      "\nEND IF;"

//+"\n select @num1;"
   +     "\nset @sql_text:=concat('DELETE FROM ',@tempTable);"
   +     "\nprepare stmt from @sql_text;"
   +     "\nexecute stmt;         	    "

   + "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p1.subject, p2.object,0,0'"
   +"\n' from ',@deltaTable,' as p1,',@deltaTable,' as p2 ',"
   + "\n'where p1.object = p2.subject');"
   +"\nprepare stmt from @sql_text;"
   +"\nexecute stmt; "

    +    "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p1.subject, p2.object,0,0'"
    +    "\n' from ',@viewName,' as p1,',@deltaTable,' as p2 ',"
    +    "\n'where p1.object = p2.subject');"
    +    "\nprepare stmt from @sql_text;"
    +    "\nexecute stmt; "

    +    "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p2.subject, p1.object,0,0'"
    +    "\n' from ',@viewName,' as p1,',@deltaTable,' as p2 ',"
    +    "\n'where p2.object = p1.subject');"
    +    "\nprepare stmt from @sql_text;"
    +    "\nexecute stmt; "

    +    "\nset @sql_text:=concat('insert ignore into ',@deltaTable, ' select * from ',@tempTable,"
    +    "\n' where not exists (select * from ',@viewName, ' where ',@tempTable,'.subject=',@viewName,'.subject and ',@tempTable,'.object=',@viewName,'.object)');"
    +    "\nprepare stmt from @sql_text;"
    +    "\nexecute stmt;    	"

    +    "\nset @sql_text:=concat('SELECT count(*) INTO @num2 FROM ',@deltaTable);"
    +    "\nprepare stmt from @sql_text;"
    +    "\nexecute stmt;   "
//+"\n select @num2;"
    +"\nIF @num1=@num2 THEN LEAVE l1;"
    +"\nEND IF;"

    +"\nIF @num2>50000 THEN LEAVE l1;"
    +"\nEND IF;"

    +"\nEND LOOP l1;"

    +"\nset @sql_text:=concat('insert ignore into ',tableName,' select * from ',@deltaTable);"
    +"\nprepare stmt from @sql_text;"
    +"\nexecute stmt;   "
  +"\nEND";

//zhp2: perspective version of Transitive Closure computation.
    String sp1 = "CREATE PROCEDURE computeTCp(IN tableName varchar(255), IN view int)"
    +"\nbeg:BEGIN"
    +  "\nDECLARE num1 int default 0;"
    +  "\nDECLARE num2 int default 0;"
    +  "\nset @tempTable := concat(tableName,'temp');"
    +  "\nset @deltaTable := concat(tableName,'delta');"
    +  "\nset @viewName := concat(tableName,'_',view,'_view');"
    +  "\nset @viewName2 := concat(tableName,'_',view,'_view2');"

    +     "\nset @sql_text:=concat('SELECT count(*) INTO @num1 FROM ',@viewName);"
    +     "\nprepare stmt from @sql_text;"
    +     "\nexecute stmt;  "
//    +     "\nSELECT @num1;"
    +     "\nIF @num1>300000 THEN LEAVE beg;"
    +      "END IF;"

    +  "\nset @sql_text:=concat('DROP TEMPORARY TABLE IF EXISTS ',@tempTable);"
    +  "\nprepare stmt from @sql_text;"
    +  "\nexecute stmt;"

    +  "\nset @sql_text:=concat('DROP TEMPORARY TABLE IF EXISTS ',@viewName);"
    +  "\nprepare stmt from @sql_text;"
    +  "\nexecute stmt;"

    +  "\nset @sql_text:=concat('DROP TEMPORARY TABLE IF EXISTS ',@viewName2);"
    +  "\nprepare stmt from @sql_text;"
    +  "\nexecute stmt;"

     + "\nset @sql_text:=concat('DROP TABLE IF EXISTS ',@deltaTable);"
     + "\nprepare stmt from @sql_text;"
     + "\nexecute stmt;  "

     + "\nset @sql_text:=concat('CREATE TEMPORARY TABLE ',@tempTable, '(subject int, object int, source int, onto int, primary key(subject, object), index(object))');"
     + "\nprepare stmt from @sql_text;"
     + "\nexecute stmt; "

     + "\nset @sql_text:=concat('CREATE TABLE ',@deltaTable, '(subject int, object int, source int, onto int, primary key(subject, object), index(object))');"
     + "\nprepare stmt from @sql_text;"
     + "\nexecute stmt;        "

     + "\nset @sql_text:=concat('CREATE TEMPORARY TABLE IF NOT EXISTS ',@viewName2, '(subject int, object int, source int, onto int, index(subject), index(object)) select * from ', @viewName);"
     + "\nprepare stmt from @sql_text;"
     + "\nexecute stmt; "

     + "\nset @sql_text:=concat('CREATE TEMPORARY TABLE IF NOT EXISTS ',@viewName, '(subject int, object int, source int, onto int, index(subject), index(object)) select * from ', @viewName);"
     + "\nprepare stmt from @sql_text;"
     + "\nexecute stmt; "

     + "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p1.subject, p2.object,', view,',',view,"
     +"\n' from ',@viewName,' as p1,',@viewName2,' as p2 ',"
     + "\n'where p1.object = p2.subject');"
     +"\nprepare stmt from @sql_text;"
     +"\nexecute stmt; "

     + "\nset @sql_text:=concat('insert ignore into ',@deltaTable, ' select * from ',@tempTable,"
     + "\n ' where not exists (select * from ',@viewName, ' where ',@tempTable,'.subject=',@viewName,'.subject and ',@tempTable,'.object=',@viewName,'.object)');"
     + "\nprepare stmt from @sql_text;"
     + "\nexecute stmt;          "

     + "\nl1:LOOP"
     +     "\nIF @num2>0 THEN set @num1 = @num2;"
     +     "\nELSE"
     +     "\nset @sql_text:=concat('SELECT count(*) INTO @num1 FROM ',@deltaTable);"
     +     "\nprepare stmt from @sql_text;"
     +     "\nexecute stmt;   "
     +      "\nEND IF;"

     +     "\nset @sql_text:=concat('DELETE FROM ',@tempTable);"
     +     "\nprepare stmt from @sql_text;"
     +     "\nexecute stmt;         	    "

     + "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p1.subject, p2.object,', view,',',view,"
     +"\n' from ',@deltaTable,' as p1,',@deltaTable,' as p2 ',"
     + "\n'where p1.object = p2.subject');"
     +"\nprepare stmt from @sql_text;"
     +"\nexecute stmt; "

      +    "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p1.subject, p2.object,', view,',',view,"
      +    "\n' from ',@viewName,' as p1,',@deltaTable,' as p2 ',"
      +    "\n'where p1.object = p2.subject');"
      +    "\nprepare stmt from @sql_text;"
      +    "\nexecute stmt; "

      +    "\nset @sql_text:=concat('insert ignore into ',@tempTable,' select p2.subject, p1.object,', view,',',view,"
      +    "\n' from ',@viewName,' as p1,',@deltaTable,' as p2 ',"
      +    "\n'where p2.object = p1.subject');"
      +    "\nprepare stmt from @sql_text;"
      +    "\nexecute stmt; "

      +    "\nset @sql_text:=concat('insert ignore into ',@deltaTable, ' select * from ',@tempTable,"
      +    "\n' where not exists (select * from ',@viewName, ' where ',@tempTable,'.subject=',@viewName,'.subject and ',@tempTable,'.object=',@viewName,'.object)');"
      +    "\nprepare stmt from @sql_text;"
      +    "\nexecute stmt;    	"

      +    "\nset @sql_text:=concat('SELECT count(*) INTO @num2 FROM ',@deltaTable);"
      +    "\nprepare stmt from @sql_text;"
      +    "\nexecute stmt;   "

      +"\nIF @num1=@num2 THEN LEAVE l1;"
      +"\nEND IF;"

      +"\nIF @num2>50000 THEN LEAVE l1;"
      +"\nEND IF;"

      +"\nEND LOOP l1;"

      +"\nset @sql_text:=concat('insert ignore into ',tableName,' select * from ',@deltaTable);"
      +"\nprepare stmt from @sql_text;"
      +"\nexecute stmt;   "
  +"\nEND";
/*
String sp = "CREATE PROCEDURE computeTC(IN tableName varchar(255), IN view int)"
+"\nBEGIN"
+  "\nDECLARE num1 int default 0;"
+  "\nDECLARE num2 int default 0;"

+  "\nset @viewName := concat(tableName,'_',view,'_view');"

 + "\nREPEAT"
 +     "\nset @sql_text:=concat('SELECT count(*) INTO @num1 FROM ',@viewName);"
 +     "\nprepare stmt from @sql_text;"
 +     "\nexecute stmt;         	    "

  +    "\nset @sql_text:=concat('insert ignore into ',@viewName,' select p1.subject, p2.object,', view,',',view,"
  +    "\n' from ',@viewName,' as p1,',@viewName,' as p2 ',"
  +    "\n'where p1.object = p2.subject',"
  +    "' and not exists (select * from ',@viewName, ',p1,p2 where p1.subject=',@viewName,'.subject and p2.object=',@viewName,'.object)');"
  +    "\nprepare stmt from @sql_text;"
  +    "\nexecute stmt; "

  +    "\nset @sql_text:=concat('SELECT count(*) INTO @num2 FROM ',@viewName);"
  +    "\nprepare stmt from @sql_text;"
  +    "\nexecute stmt;   "

  +"\nUNTIL num1=num2"
  +"\nEND REPEAT;"

  +"\nset @sql_text:=concat('insert ignore into ',tableName,' select * from ',@deltaTable);"
  +"\nprepare stmt from @sql_text;"
  +"\nexecute stmt;   "
  +"\nEND";


*/



   String sp2 =
  "CREATE PROCEDURE substituteid(IN orig INT, IN targ INT) \n"
  +"BEGIN \n"
  +"  DECLARE done INT DEFAULT 0; \n"
  +"  DECLARE tname VARCHAR(255); \n"
  +"  DECLARE tid, oid, pos INT; \n"
  +"  DECLARE cur1 CURSOR FOR  \n"
  +"    select tablename, tables_index.termid, tables_index.ontid , individual_occurrence.position \n"
  +"    from individual_occurrence, tables_index \n"
  +"    where id=orig \n"
  +"        and individual_occurrence.termid=tables_index.termid \n"
  +"        and individual_occurrence.ontid=tables_index.ontid; \n"
  +"  DECLARE CONTINUE HANDLER FOR SQLSTATE '02000' SET done = 1;      \n"
  +" \n"
  +"  set @targ=targ; \n"
  +"  set @orig=orig;   \n"
  +"  \n"
  +"  OPEN cur1; \n"
  +"  REPEAT \n"
  +"     FETCH cur1 INTO tname, tid, oid, pos; \n"
  +"     IF tid>0 THEN \n"
  +"         set @sql_text:=concat('update ignore ',tname,' set id=@targ where id=@orig'); \n"
  +"    	 prepare stmt from @sql_text; \n"
  +"		 execute stmt; \n"
  +" \n"
  +"         insert ignore individual_occurrence values(targ, oid, tid,0); \n"
  +"        \n"
  +"         set @sql_text:=concat('delete from ',tname,' where id=@orig'); \n"
  +"    	 prepare stmt from @sql_text; \n"
  +"    	 execute stmt; \n"
  +"     END IF; \n"
  +"     IF tid<0 and pos=1 THEN \n"
  +"         set @sql_text:=concat('update ignore ',tname,' set subject=@targ where subject=@orig'); \n"
  +"    	 prepare stmt from @sql_text; \n"
  +"    	 execute stmt;      \n"
  +" \n"
  +"     	 insert ignore individual_occurrence values(targ, oid, tid,1); \n"
  +"     	  \n"
  +"         set @sql_text:=concat('delete from ',tname,' where subject=@orig'); \n"
  +"    	 prepare stmt from @sql_text; \n"
  +"    	 execute stmt; \n"
  +"     END IF;      \n"
  +"     IF tid<0 and pos=2 THEN \n"
  +"         set @sql_text:=concat('update ignore ',tname,' set object=@targ where object=@orig'); \n"
  +"    	 prepare stmt from @sql_text; \n"
  +"    	 execute stmt;      \n"
  +" \n"
  +"     	 insert ignore individual_occurrence values(targ, oid, tid,2); \n"
  +"     	  \n"
  +"         set @sql_text:=concat('delete from ',tname,' where object=@orig'); \n"
  +"    	 prepare stmt from @sql_text; \n"
  +"    	 execute stmt; \n"
  +"     END IF;           \n"
  +"     delete from individual_occurrence where id=orig; \n"
  +"     update uri_index set id=targ where id=orig; \n"
  +"  UNTIL done \n"
  +"  END REPEAT; \n"
  +"  CLOSE cur1; \n"
  +"END";

String sp3 =
     "CREATE PROCEDURE findSameAs(IN tableName VARCHAR(255)) \n"
     +"BEGIN \n"
     +"  DECLARE done INT DEFAULT 0; \n"
     +"  DECLARE a VARCHAR(255); \n"
     +"  DECLARE b,c,num INT; \n"

     +"  DECLARE cur1 CURSOR FOR select object from sameas_temp1;       \n"
     +"  DECLARE cur2 CURSOR FOR select subject from sameas_temp2;\n"
     +"  DECLARE CONTINUE HANDLER FOR SQLSTATE '02000' SET done = 1; \n"

     +"  SET @tableName:=tableName;\n"
     +"  DROP TEMPORARY TABLE IF EXISTS sameas_temp1;       \n"
     +"  DROP TEMPORARY TABLE IF EXISTS sameas_temp2;    \n"
     +"  CREATE TEMPORARY TABLE sameas_temp1(object varchar(255), primary key(object));\n"
     +"  CREATE TEMPORARY TABLE sameas_temp2(subject int, primary key(subject));  \n"

     +"  set @sql_text:=concat('insert into sameas_temp1 select object from ',@tableName,' group by object having count(distinct subject)>1');\n"
     +"  prepare stmt from @sql_text;  \n"
     +"  execute stmt;  \n"
     +"  OPEN cur1;\n"
     +"  loop1:LOOP\n"
     +"  	FETCH cur1 into a;\n"
     +"  	IF done THEN\n"
     +"  	  CLOSE cur1;\n"
     +"  	  LEAVE loop1;\n"
     +"  	END IF;\n"
     +"  	SET num = 0;\n"
     +"  	set @sql_text:=concat('insert ignore into sameas_temp2 select distinct subject from ',@tableName,' where object=\\\'',a,'\\\'');\n"
     +"    prepare stmt from @sql_text;\n"
     +"    execute stmt;\n"
     +"   	OPEN cur2;\n"
     +"  	loop2:LOOP\n"
     +"  	  FETCH cur2 into c;\n"
     +"  	  IF done THEN\n"
     +"  	    CLOSE cur2;\n"
     +"  	    SET done = 0;\n"
     +"  	    LEAVE loop2;\n"
     +"  	  END IF;\n"
     +"  	  IF num=0 THEN\n"
     +"  	    SET b=c;\n"
     +"  	  ELSE\n"
     +"  	  	CALL substituteid(c,b);\n"
     +"  	  END IF;\n"
     +"  	  SET num = num+1;	  \n"
     +"  	END LOOP loop2;\n"
     +"  	delete from sameas_temp2;  	\n"
     +"  END LOOP loop1;	  \n"
+"END ";

    String sp4 =
         "CREATE PROCEDURE findSameAs2(IN tableName VARCHAR(255)) \n"
         +"BEGIN \n"
         +"  DECLARE done INT DEFAULT 0; \n"
         +"  DECLARE a VARCHAR(255); \n"
         +"  DECLARE b,c,num INT; \n"

         +"  DECLARE cur1 CURSOR FOR select object from sameas_temp1;       \n"
         +"  DECLARE cur2 CURSOR FOR select subject from sameas_temp2;\n"
         +"  DECLARE CONTINUE HANDLER FOR SQLSTATE '02000' SET done = 1; \n"

         +"  SET @tableName:=tableName;\n"
         +"  DROP TEMPORARY TABLE IF EXISTS sameas_temp1;       \n"
         +"  DROP TEMPORARY TABLE IF EXISTS sameas_temp2;    \n"
         +"  CREATE TEMPORARY TABLE sameas_temp1(object varchar(255), primary key(object));\n"
         +"  CREATE TEMPORARY TABLE sameas_temp2(subject int, primary key(subject));  \n"

         +"  set @sql_text:=concat('insert into sameas_temp1 select subject from ',@tableName,' group by subject having count(distinct object)>1');\n"
         +"  prepare stmt from @sql_text;  \n"
         +"  execute stmt;  \n"
         +"  OPEN cur1;\n"
         +"  loop1:LOOP\n"
         +"  	FETCH cur1 into a;\n"
         +"  	IF done THEN\n"
         +"  	  CLOSE cur1;\n"
         +"  	  LEAVE loop1;\n"
         +"  	END IF;\n"
         +"  	SET num = 0;\n"
         +"  	set @sql_text:=concat('insert ignore into sameas_temp2 select distinct object from ',@tableName,' where subject=\\\'',a,'\\\'');\n"
         +"    prepare stmt from @sql_text;\n"
         +"    execute stmt;\n"
         +"   	OPEN cur2;\n"
         +"  	loop2:LOOP\n"
         +"  	  FETCH cur2 into c;\n"
         +"  	  IF done THEN\n"
         +"  	    CLOSE cur2;\n"
         +"  	    SET done = 0;\n"
         +"  	    LEAVE loop2;\n"
         +"  	  END IF;\n"
         +"  	  IF num=0 THEN\n"
         +"  	    SET b=c;\n"
         +"  	  ELSE\n"
         +"  	  	CALL substituteid(c,b);\n"
         +"  	  END IF;\n"
         +"  	  SET num = num+1;	  \n"
         +"  	END LOOP loop2;\n"
         +"  	delete from sameas_temp2;  	\n"
         +"  END LOOP loop1;	  \n"
+"END ";

 if(dbName.indexOf("mysql")>0){
     stmt.executeUpdate(sp);
     stmt.executeUpdate(sp1);
     stmt.executeUpdate(sp2);
     stmt.executeUpdate(sp3);
     stmt.executeUpdate(sp4);
 }
    }catch(SQLException e){
     System.out.println("error when create stored procedures.");
     System.out.println("SQLState:"+e.getSQLState());
     System.out.println("Message :"+e.getMessage());
     System.out.println("Vendor  :"+e.getErrorCode());
    }
  }

  /**
   * Get the table name for giving id.
   */
  public String getTableName(ObjectId id){
   if (id==null)
       return null;
    String local = id.getLocalName();
    String ontoName = id.getNameSpace();
    ontoName = ontoName.replace('#',' ').trim();
/*
    Document doc = (Document)getDocument(ontoName);

    if(doc==null||!(doc instanceof Ontology)){
      System.out.println("+++ Can not find onto that defined term: "+id.getId());
      return null;
    }
    Ontology ont = (Ontology)doc;
    if(!ont.hasClass(local)&&!ont.hasProperty(local)){
      return null;
    }
*/ //pan on 10/28/05 and 01/10/06
    Integer seq = (Integer)knownOntos.get(ontoName);
    if(ontoName.equals(""))
      System.out.println("getTableName: "+id.getId());
    //System.out.println("getTableName: "+ontoName);
    if(seq!=null)
      return local.replace('-','_').replace('.','_')+"_"+seq.intValue();
    if(local.equalsIgnoreCase("Thing")&&ontoName.startsWith("http://www.w3.org/2002/07/owl"))
      return "thing_owl";
    if(local.equalsIgnoreCase("Resource")&&ontoName.startsWith("http://www.w3.org/"))
      return "thing_owl";
    if(local.equalsIgnoreCase("label")&&ontoName.startsWith("http://www.w3.org/"))
      return "label_rdfs";
//  System.out.println("can't find table name for "+id.getId());
    return null;
  }

  /**
   * Retrieval Urls already exist in database
   */
  public Hashtable getKnownUrls()
  {
    if(knownUrls!=null)
        return knownUrls;
    else{
        java.sql.ResultSet rs = null;
        int id = 0;
        String url;
        String qs = "SELECT MAX(id) FROM uri_index";
        knownUrls = new Hashtable();

        try {
            System.out.println("get known urls..");
            rs = stmt.executeQuery(qs);
            if (rs.next()) {
                id = rs.getInt(1);
            }
            rs.close();
            maxUrlNum = id;
        } catch (SQLException e) {
            System.out.println("error when getting known URIs!");
            System.out.println("SQLState:" + e.getSQLState());
            System.out.println("Message :" + e.getMessage());
            System.out.println("Vendor  :" + e.getErrorCode());
        }

        return knownUrls;
    }
  }


  /**
   * Add the new ontology to index.
   */
  private void updateOntIndex(Ontology onto, int seqnum)
  {

     String qs = "INSERT INTO Ontologies_index VALUES ('"+onto.getID().getId()+"', "
                   +String.valueOf(seqnum)+")";
     try {
     stmt.executeUpdate(qs);
     }
     catch(SQLException e)
               {
                 System.out.println("error when update onto index");
                 System.out.println(qs);
                 System.out.println("SQLState:"+e.getSQLState());
                 System.out.println("Message :"+e.getMessage());
                 System.out.println("Vendor  :"+e.getErrorCode());
               }
     knownOntos.put(onto.getID().getId(), new Integer(seqnum));
  }

  /**
   * Add a new document to the index.
   */
  private void updateSourceIndex(Document onto, int seqnum)
  {
     String qs = "INSERT INTO source_index VALUES ('"+onto.getID().getId()+"', "
                   +String.valueOf(seqnum)+")";
   //  log.println(qs);
     try {
     stmt.executeUpdate(qs);
     }
     catch(SQLException e)
               {
                 System.out.println("error when update source index");
                 System.out.println(qs);
                 System.out.println("SQLState:"+e.getSQLState());
                 System.out.println("Message :"+e.getMessage());
                 System.out.println("Vendor  :"+e.getErrorCode());
               }
  }

  /**
   * Get the index for given uri.
   */
  protected int getUrlIndex(String url){
    int urlId = -1;
    url = url.replace("'","\\'"); //zhp2: added for database escape
    if(knownUrls.get(url)!=null){
      urlId = ( (Integer) knownUrls.get(url)).intValue();
    }
    else{
      String qus = "SELECT id FROM uri_index WHERE uri='"+url+"'";
      try{
        java.sql.ResultSet rs = stmt.executeQuery(qus);
        if(rs.next())  {
          urlId = rs.getInt(1);
          knownUrls.put(url,new Integer(urlId));
        }
        else  {
          urlId = addUrl(url);
        }
      }
      catch(SQLException e){
                  System.out.println("error");
                  System.out.println("SQLState:"+e.getSQLState());
                  System.out.println("Message :"+e.getMessage());
                  System.out.println("Vendor  :"+e.getErrorCode());
      }
    }
    return urlId;
  }

  /**
   * Add a new url to the url index.
   */
  private int addUrl(String url)
  {

    String qs = "INSERT INTO uri_index VALUES ('"+url+"', "
                   +String.valueOf(++maxUrlNum)+")";
     try {
     stmt.executeUpdate(qs);
     knownUrls.put(url,new Integer(maxUrlNum));
     return maxUrlNum;
     }
     catch(SQLException e)
               {
                 System.out.println("error when update source index");
                 System.out.println(url);
                 System.out.println("SQLState:"+e.getSQLState());
                 System.out.println("Message :"+e.getMessage());
                 System.out.println("Vendor  :"+e.getErrorCode());
                 return -1;
               }
  }

  /**
   *
   * Get the sequence number for giving Document
   */
  public int getSeqNum(DocId did){
    if(knownDocs.get(did)!=null){
      return ((Integer)knownDocs.get(did)).intValue();
    }
    else{
      int seq = sourceSeqNum++;
      knownDocs.put(new Integer(seq),did);
      return seq;
    }
  }


  public void setViewDefinition(String vName, String vDef){
      viewDef.put(vName,vDef);
  }

  public Hashtable getUpdatedViewDefs(){
      return viewDef;
  }

  public void clearViewDefinitions(){
      viewDef.clear();
  }

  /**
   * Get sql definition for giving view name.
   */
  public String getViewDefinition(String vName){
    if(viewDef.get(vName)!=null)          //cache view definitions that has been changed in memory but hasn't been committed to database.
          return viewDef.get(vName);
    String qs = "SELECT definition FROM \n"+"view_definition WHERE viewname='"
        +vName+"'";
    try{
      java.sql.ResultSet rs = stmt.executeQuery(qs);
      while(rs.next()){
        return rs.getString("definition");
      }
      return null;
    }
    catch(SQLException e){
      System.out.println("error when get view definition");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
      return null;
    }
  }

  /**
   *
   * Get the sequence number for giving ontology.
   */
  public int getOntSeq(String ns){
    Integer in = (Integer)knownOntos.get(ns.replace('#',' ').trim());
    if(in==null)
      return -1;
    else
      return in.intValue();
  }

  /**
   * Forced exit.
   */
  public void forcedExit(){
      try{
          con.rollback();
      }
      catch(SQLException e){
          System.out.println("error when forced exit");
          System.out.println("SQLState:"+e.getSQLState());
          System.out.println("Message :"+e.getMessage());
          System.out.println("Vendor  :"+e.getErrorCode());
      }
      try{
          close();
      }
      catch(Exception e){
          System.err.println(e.getMessage());
          //e.printStackTrace();
      }
  }

  /**
   * Execute query and return results.
   */
  public edu.lehigh.swat.hawk.storage.ResultSet issueQuery(edu.lehigh.swat.hawk.storage.simpledb.Query qry){
    edu.lehigh.swat.hawk.storage.ResultSet rset;
    java.sql.ResultSet rs;

    Query query = (Query)qry;
    DocId persp = query.getPerspective();
    String nameSpace = getQueryOntology().concat("#");
    if(persp!=null)
      nameSpace = persp.getId().concat("#");
//    int perspSeq = this.getOntSeq(nameSpace);
    //ObjectId oid = new ObjectId("");
    if(propertyTypes==null)
      getPropertyTypes();

    int ptype=0;
    Vector a = query.getAtoms();
    int l = a.size();

    StringBuffer colName = new StringBuffer("SELECT DISTINCT ");
    StringBuffer tableName = new StringBuffer("FROM ");
    StringBuffer condition = new StringBuffer("WHERE ");
//    int hasname=0;
    int colnum = 0;
    Vector vcol = new Vector();
    vcol.clear();

    Hashtable tables = new Hashtable();
    tables.clear();

  //  Vector columns = new Vector();//new for alias
  //  columns.clear(); //--

    Hashtable classes = new Hashtable();
    Hashtable properties = new Hashtable();

    for(int i=0;i<l;i++){
      ptype=0;
      edu.lehigh.swat.hawk.storage.simpledb.Atom am = (edu.lehigh.swat.hawk.storage.simpledb.Atom) a.get(i);
      String sp = am.toString();
      System.out.println(sp);

      String amp = am.getPredicate();
      String[] ama = am.getArguments();
      if(amp.startsWith("@every")||amp.startsWith("type"))
      ptype=1;
      if(amp.startsWith("@string"))
      ptype=2;

      String tn;

      switch(ptype){
      case 1:
              ObjectId objid;
              if(ama[0].indexOf('#')>0||ama[0].indexOf('/')>0)
                objid = new ObjectId(ama[0]);
              else
                objid = new ObjectId(nameSpace+ama[0]);

              tn = getTableName(objid);
            //  tn = tn+"_"+perspSeq;
              tn = tn + "_view";
              tableName.append(tn).append(", ");
              tables.put(tn,"class");
              if(properties.get(ama[1])==null){
                tableName.append("uri_index AS ui" + i + ", ");
                colName.append("ui" + i + ".uri, ");
                condition.append(tn + ".id = ui" + i + ".id AND ");
                vcol.add(ama[1]);
                classes.put(ama[1], tn + ".id");
              }else{
                condition.append(properties.get(ama[1])).append("=");
                condition.append(tn).append(".id AND ");
              }
      break;
      case 2:

              //colName.append("url_index.url, url_index.id, ");
              if(properties.get(ama[0])!=null){
                String fullname = (String)properties.get(ama[0]);
                int dot = fullname.indexOf('.');
                String tname = fullname.substring(0,dot-5);
           //     String cname = fullname.substring(dot);
                String type = "";
                try{
   /*               rs = con.getMetaData().getColumns("","",tname,cname);
                  if(rs.next())
                    type = rs.getString("TYPE_NAME");
    */
                 type = (String)propertyTypes.get(tname);
                }
                catch(Exception e)
                {
                 System.out.println(e.getMessage());
                }


                if(type.equalsIgnoreCase("object"))
                {
                  tableName.append("uri_index AS ui"+i+", ");
                  condition.append("ui"+i+".uri");
                  condition.append(" LIKE '").append(ama[1]).append("' AND ");
                  condition.append(properties.get(ama[0]));
                  condition.append("=").append("ui"+i+".id AND ");
                }
                else
                {
                  condition.append(properties.get(ama[0]));
                  condition.append(" LIKE '").append(ama[1]).append("' AND ");
                }
              }
      break;
      default:
              ObjectId oid;
              if(amp.indexOf('#')>0||amp.indexOf('/')>0)
                oid = new ObjectId(amp);
              else
                oid = new ObjectId(nameSpace+amp);
              tn = getTableName(oid)+"_p";
          //    tn = tn+"_"+perspSeq;

              if(((String)propertyTypes.get(tn)).equals("hasinverse"))
                  tn = tn+"_inv";
              String vn = tn+"_view";
              if(!tables.contains(vn)){
                tableName.append(vn).append(", ");
                tables.put(vn,"property");
              }

              //second argument, object.
              if(ama[1].startsWith("?")){
                if(classes.get(ama[1])!=null){
                  condition.append(classes.get(ama[1])).append("=");
                  condition.append(vn).append(".object AND ");
                }
                if(properties.get(ama[1])!=null){
                  condition.append(properties.get(ama[1])).append("=");
                  condition.append(vn).append(".object AND ");
                }
                if(properties.get(ama[1])==null&&classes.get(ama[1]) == null){
                    String select_column_name = vn+".object";
                    String type = (String)propertyTypes.get(tn);
                    if(type!=null&&type.equalsIgnoreCase("datatype")){
                      colName.append(select_column_name).append(", ");
                      tables.remove(vn);
                      tables.put(vn, "datatype");
                    }
                    else{
                      tableName.append("uri_index AS uio"+i+", ");
                      colName.append("uio"+i+".uri, ");
                      condition.append(select_column_name+" = uio"+i+".id AND ");
                    }
                    properties.put(ama[1], select_column_name);
                    vcol.add(ama[1]);
                }
              }
              else{// argument is not variable
                tableName.append("uri_index AS uio"+i+", ");
                condition.append("uio"+i+".uri='").append(ama[1]).append("' AND ");
                condition.append(vn).append(".object=uio"+i+".id AND ");
              }

              //first argument, subject.
              if(ama[0].startsWith("?")){
                if (properties.get(ama[0]) != null) {
                  condition.append(properties.get(ama[0])).append("=");
                  condition.append(vn).append(".subject ").append(" AND ");
                }
                if (classes.get(ama[0]) != null){
                  condition.append(classes.get(ama[0])).append("=");
                  condition.append(vn).append(".subject AND ");
                }
                if(properties.get(ama[0])==null&&classes.get(ama[0]) == null){
                  String select_column_name = vn + ".subject";
                  tableName.append("uri_index AS uis" + i + ", ");
                  colName.append("uis" + i + ".uri, ");
                  condition.append(select_column_name).append(" = ");
                  condition.append("uis" + i + ".id AND ");
                  vcol.add(ama[0]);
                  properties.put(ama[0], select_column_name);
                }
              }
              else{  //argument is not variable
                  tableName.append("uri_index AS uis"+i+", ");
                  condition.append("uis"+i+".uri='").append(ama[0]).append("' AND ");
                  condition.append(vn).append(".subject=uis"+i+".id AND ");
              }
          }//end switch

    }//end for
    int len = colName.length();
    colName.delete(len-2,len);
    colName.append("\n");

    len = tableName.length();
    tableName.delete(len-2,len);
    tableName.append("\n");

    len = condition.length();
    condition.delete(len-4,len);
    condition.append("\n");

    String colNames = colName.toString();
    String tableNames = tableName.toString();
    String conditions = condition.toString();

    if(conditions.length()>6)
      System.out.println(colNames+tableNames+conditions);
    else
      System.out.println(colNames+tableNames);

    colnum = vcol.size();
    String[] varnames = new String[vcol.size()];
    for(int i=0;i<colnum;i++)
    varnames[i] = (String) vcol.elementAt(i);

    rset = new edu.lehigh.swat.hawk.storage.ResultSet();

    long t0 = System.currentTimeMillis();// test
    if(dbName.indexOf("mysql")>0&&(0>1)){
 //     rs.close();  //zpan
      try{
        Enumeration en = tables.keys();
        stmt.clearBatch();
        while(en.hasMoreElements()){
          String tname = (String)en.nextElement();
          String ttype = (String)tables.get(tname);
          String temptable = "create temporary table "+tname;
          if(ttype.equals("property")){
            temptable =
            temptable+" (subject int, object int, source int, onto int, key(subject), key(object))";
          }else if(ttype.equals("datatype")){
            temptable =
            temptable+" (subject int, object text, source int, onto int, key(subject), key(object))";
          }else
            temptable = temptable+" (id int, source int, onto int, key(id))";
          temptable = temptable+" select * from "+tname+";";
 //         System.out.println(temptable);
          stmt.addBatch(temptable);
        }
        stmt.executeBatch();
     }catch(SQLException e){
     //System.out.println("error when do query");
     System.out.println(e.getMessage());
     //System.out.println(e.getErrorCode());
     }
   }

   //zhp2: for hsqsldb
/*   if(dbName.indexOf("hsql")>0){

           Enumeration en = tables.keys();
           while (en.hasMoreElements()) {
               String tname = (String) en.nextElement();
               String qs0 = "drop view "+tname;
               String qs = this.getViewDefinition(tname);
               try{
                   stmt.execute(qs0);
                   stmt.execute("drop view undergraduatestudent_1_view");
             } catch(Exception e){
                       e.printStackTrace();
                 }
               try{
                     System.out.println(qs);
                       stmt.execute(qs);
       //                System.out.println(qs);
                   } catch(Exception e){
                           e.printStackTrace();
                   }

          }
   }
*/
   try{
    if(conditions.length()>6)
      rs = stmt.executeQuery(colNames+tableNames+conditions);
    else
      rs = stmt.executeQuery(colNames+tableNames);
    //test
    long t1 = System.currentTimeMillis();
    long t = t1-t0;
    System.out.println("\n\ntime in SQL query: "+t+" ms.");
    //end test

    while(rs.next())
    {
      String[] data = new String[colnum];
      for(int i=0;i<colnum;i++)
      {
        data[i] = rs.getString(i+1);
      }

      rset.addResult(new Result(data));
    }

//    rset.setSqlResult(rs);
  //  rs.close();
    }

    catch(SQLException e){
        System.out.println("error when do query");
        System.out.println(e.getMessage());
      //  System.out.println(e.getErrorCode());
    }

    return rset;

  }


  /**
    * Execute query and return results.
    */
   public edu.lehigh.swat.hawk.storage.ResultSet issueQuery(edu.lehigh.swat.hawk.storage.simpledb.SparqlQuery query){
     edu.lehigh.swat.hawk.storage.ResultSet rset;
     java.sql.ResultSet rs;


//     SparqlQuery query = (SparqlQuery)qry;
     //if the sql has not been created yet, create it now
     if("".equals(query.getSql())){
         SparqlQueryDB sqdb = new SparqlQueryDB();
         sqdb.setStorage(this);
 /*        String fileName = "tempFile";
         File tempFile = new File(fileName);
         try {
             FileOutputStream fw = new FileOutputStream(tempFile);
             fw.write(query.getOrigSparql().getBytes());
             fw.close();
         }catch(IOException ie) {
             System.out.println("IOException creating tempFile: " + ie.getMessage());
         }
         ArrayList queryList = sqdb.parseQuery(fileName); */
         ArrayList queryList = sqdb.parseQueryString(query.getOrigSparql());
         query = (SparqlQuery)queryList.get(0);

     }

     StringBuffer sql = new StringBuffer(query.getSql());
     //DocId persp = query.getPerspective();
     long t0 = System.currentTimeMillis();// test
     Map tables = query.getDBTables();
     if(dbName.indexOf("mysql")>0&&(tables.size()>1||tables.containsKey("Student_1_1_view")||tables.containsKey("PeopleWithHobby_1_1_view"))){
//     rs.close();  //zpan
       try{
         Iterator en = tables.keySet().iterator();
         stmt.clearBatch();
        // Vector pass = new Vector();
         while(en.hasNext()){
           String tname = (String)en.next();
           String ttype = (String)tables.get(tname);

           String temptable = "create temporary table if not exists "+tname;
           if(ttype.equals("property")){
             temptable =
             temptable+" (subject int, object int, source int, onto int, key(subject), key(object))";
           //}else if(ttype.equals("datatype")){
            // temptable =
           //  temptable+" (subject int, object TEXT, source int, onto int, key(subject))";
             //continue;
           }else if(ttype.equals("class"))
             temptable = temptable+" (id int, source int, onto int, key(id))";

           if(!tname.endsWith("w")){
               tname = tname.substring(0, tname.length() - 1);
               //temptable = "drop temporary table IF EXISTS "+ tname;
               temptable = temptable+" select * from "+tname;
               int sind = sql.indexOf(tname+" AS");
               int eind = sind+tname.length()+3;
               sql.delete(sind, eind);
           }else
               temptable = temptable+" select * from "+tname;
           System.out.println("create temp table for "+tname);
           try{
               stmt.execute(temptable);
           }catch(SQLException e){
               System.out.println(e.getMessage());
               continue;
           }
         }
        // stmt.executeBatch();
      }catch(Exception e){
       // System.out.println("error when create temp tables");
        System.out.println(e.getMessage());
       //System.out.println(e.getErrorCode());
      }
  }


     rset = new edu.lehigh.swat.hawk.storage.ResultSet();
//     long t0 = System.currentTimeMillis();// test
     try{
      //   System.out.println(query.getSql());
         String qs = sql.toString();
     //  qs = qs+" limit 1000"; //zhp2 on 11/12/07: for hawkeye demo
         System.out.println(qs);
        // qs="select count(*) from Student_1_view";
         rs = stmt.executeQuery(qs+"\n");

         //test
         long t1 = System.currentTimeMillis();
         long t = t1-t0;
     //    System.out.println("\n\ntime in SQL query: "+t+" ms.");
         //end test
         ResultSetMetaData rsmd = rs.getMetaData();
         int colnum = rsmd.getColumnCount();
         String[] headers = query.getHeaders();
         if(headers!=null){
             rset.setHeaders(headers);
//             int n = headers.length;
 //            System.out.println();
 //            for(int i=0;i<n;i++){
 //              System.out.print(headers[i]+".............");
 //            }
         }
         else
             System.out.println("no headers in sparql query!");

         while(rs.next())
         {
           String[] data = new String[colnum];
           for(int i=0;i<colnum;i++)
           {
             data[i] = rs.getString(i+1);
           }
           rset.addResult(new Result(data));
         }

         //test
         long t2 = System.currentTimeMillis();
         t = t2-t0;
  //       System.out.println("\n\ntime in ResultSet: "+t+" ms.");
         //end test

     //    rset.setSqlResult(rs);
       //  rs.close();
     } catch(SQLException e){
         System.out.println("error when do query");
         System.out.println(e.getMessage());
         System.out.println(e.getErrorCode());
        System.out.println("SQLState:" + e.getSQLState());
     }

     return rset;

  }


  /**
   * Get the types of properties.
   */
  public Hashtable getPropertyTypes(){
    if(propertyTypes!=null)
        return propertyTypes; //zhp2: add two lines on 1/4/06
    updatePropertyTypes();
    return propertyTypes;
  }

  public void updatePropertyTypes(){
      String qs = "SELECT * \nFROM propertytypes";
      java.sql.ResultSet rs = null;
      try
      {
          rs = stmt.executeQuery(qs);
      }
      catch(SQLException e)
      {
          System.out.println("error");
          System.out.println("SQLState:"+e.getSQLState());
          System.out.println("Message :"+e.getMessage());
          System.out.println("Vendor  :"+e.getErrorCode());
                       // errcode = e.getErrorCode();
      }

      Hashtable propertytypes = new Hashtable();
      try{
          System.out.println("update property types..");
          while (rs.next())
          {
              String pname = rs.getString("pname");
              String ptype = rs.getString("type");
              propertytypes.put(pname, ptype);

          }
          rs.close();
      }catch (SQLException e)
      {
          System.out.println(e.getMessage());
      }
      propertyTypes = propertytypes;
  }

 /**
  * Decide which ontology would commit to giving the resource doc and objectId.
  */
 public int[] decideCommitOnto(Document doc, ObjectId id){
   String ns = id.getNameSpace().replace('#',' ').trim();
   DocId did = new DocId(ns);
   if(doc instanceof Ontology){
     Ontology on = (Ontology)doc;
     if(on.getOntClass(id)!=null||on.getProperty(id)!=null)
       return new int[]{on.getOntSeq()};
   }
   DocId[] dids = doc.getImports();
   if(dids==null)
     return new int[]{this.getOntSeq(ns)};
   int len = dids.length;
   int[] seqs = new int[len];
   int num = 0;
   for(int i=0;i<len;i++){
     if(dids[i].equals(did)){
       return new int[]{this.getOntSeq(ns)};
     }
     edu.lehigh.swat.hawk.core.Document idoc = this.getDocument(dids[i].getId());
     if(idoc instanceof edu.lehigh.swat.hawk.core.Ontology){
         edu.lehigh.swat.hawk.core.Ontology ont =
         (edu.lehigh.swat.hawk.core.Ontology)this.getDocument(dids[i].getId());
         if(ont.getOntClass(id)!=null||ont.getProperty(id)!=null){
             seqs[num] = this.getOntSeq(dids[i].getId());
             num++;
        }
     }
   }
   if(num>0){
     int[] rets = new int[num];
     System.arraycopy(seqs,0,rets,0,num);
     return rets;
   }
   return new int[]{this.getOntSeq(ns)};
 }

 /**
  * Set the query ontology.
  */
  public void setQueryOntology(String uri){
    queryOnto = uri;
  }

  /**
   * Get the query ontology.
   */
  public String getQueryOntology(){
    return queryOnto;
  }

  /**
   * Get the database name.
   */
  public String getDBName(){
    return dbName;
  }

  /**
   * Sets the hashtable that contains the known ontologies.
   * @param KnownOntos the hashtable of known ontologies
   */
  public void setKnownOntologies(Hashtable KnownOntos){
      knownOntos = KnownOntos;
  }

  /**
   * Sets the connection to the database.
   * @param Conn a Connection object.
   */
  public void setConnection(Connection Conn)throws SQLException{
      con = Conn;
      stmt = con.createStatement(java.sql.ResultSet.TYPE_FORWARD_ONLY,java.sql.ResultSet.CONCUR_READ_ONLY);
  }

  /**
   * Add imports objects to ontology, for the use of memory-based storage model.
   * @param ont Ontology
   */
  public void pushImports(edu.lehigh.swat.hawk.core.Ontology ont){
  }



  /**
   * Get the imports closure.
   */
  public DocId[] getImportsClosure(edu.lehigh.swat.hawk.core.Document doc){
    DocId[] imports = doc.getImports();

    if(imports==null)
      return null;

    Vector<DocId> ic = new Vector();
    for(int i =0;i<imports.length;i++){
          ic.add(imports[i]);
          DocId did = imports[i];
          edu.lehigh.swat.hawk.core.Document idoc = this.getDocument(did.getId());
          DocId[] importIds = getImportsClosure(idoc);
          if(importIds==null)
              continue;
          int len = importIds.length;
          for(int j=0;j<len;j++){
              if(!ic.contains(importIds[j]))
                  ic.add(importIds[j]);
          }
    }
     int size = ic.size();
     return ic.toArray(new DocId[size]);
  }

  protected int ontSeqNum=0, sourceSeqNum=0, maxUrlNum=0;
  protected Hashtable knownOntos, knownDocs, knownUrls, propertyTypes;
  protected String dbName, queryOnto;
  protected Connection con;
  protected Statement stmt;
  protected Hashtable cachedDocs, loadedDocs;
  private boolean serializeOnto = true;
  protected String emptyTableState = "";
  protected int loadedTripleNum;
  protected boolean queryOnly = false;
  private Hashtable<String,String> viewDef = new Hashtable();
}
