/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/

package edu.lehigh.swat.hawk.storage.simplememory;
import edu.lehigh.swat.hawk.core.Axiom;
import edu.lehigh.swat.hawk.core.AxiomNode;
import edu.lehigh.swat.hawk.core.ObjectId;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;
/**
 * A simple implementation of core.OntProperty in memory.
 * @author Z. Pan
 * @version 1.0
 */

public class OntProperty extends OntObject
    implements edu.lehigh.swat.hawk.core.OntProperty{

  /**
   * Constructor
   */
  public OntProperty() {
    transitive = false;
    symetric = false;
    functional = false;
    inverseFunctional = false;
    deprecated = false;
    datatype = null;
    annotation = false;
    inverseOf = new Vector();
    subProperties = new Vector();
    superProperties = new Vector();
    equivalent = new Vector();
  }

  /**
   * get the domain
   */
  public ObjectId getDomain(){
    return domain;
  }

  /**
   * set the domain
   */
  public void setDomain(ObjectId c){
    domain = c;
  }

  /**
   * get the range
   */
  public ObjectId getRange(){
    return range;
  }

  /**
   * set the domain
   */
  public void setRange(ObjectId c){
    range = c;
  }

  /**
   * test if this property is transitive
   */
  public boolean isTransitive(){
    return transitive;
  }

  /**
   * set the property to be transitive or not
   */
  public void setTransitive(boolean b){
    transitive = b;
  }

  /**
   * test if this property is symetric
   */
  public boolean isSymmetric(){
    return symetric;
  }

  /**
   * set the property to be symetric or not
   */
  public void setSymmetric(boolean b){
    symetric = b;
  }

  /**
   * add a direct subProperty
   * $memo: currently we only set bi-directional links to those porperties who are
   * in the same document.
   */
  public void addSubProperty(ObjectId p){
      if(p.equals(this.getID()))
        return;

	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBPROPERTYOF));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(p) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(this.getID()))
	    		return;
	    }
	    }
	    Axiom axiom = new Axiom(Ontology.SUBPROPERTYOF);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(p);
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(this.getID());
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(p, axiom);
		ont.addAxiomByObject(this.getID(), axiom);
		ont.addAxiomByType(new Integer(Ontology.SUBPROPERTYOF), axiom);

  }

  /**
   * get the direct subProperties
   */
  public ObjectId[] getDirectSubProperties(){
    /*int len = subProperties.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = subProperties.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBPROPERTYOF));
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	   	Axiom a = it.next();
	   	if(a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(this.getID()) &&
			a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE){
	   		od.add(a.getLeftAxiomNode().getAtomicNode());
	   	}
	}
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * delete a direct subProperty
   */
  public void deleteSubProperty(ObjectId p){
    /*if(subProperties.contains(p)){
      subProperties.remove(p);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntProperty pp = ont.getProperty(p);
      if(pp!=null)
        pp.deleteSuperProperty(this.getID());
    }*/
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBPROPERTYOF));
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	   	Axiom a = it.next();
	   	if(a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(this.getID()) &&
			a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getLeftAxiomNode().getAtomicNode().equals(p)){
	   		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(p).remove(a);
	   		ont.getAxiomByObject(this.getID()).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.SUBPROPERTYOF)).remove(a);
	   		break;
	   	}
	}
	}
  }

  /**
   * add a direct super Property
   */
  public void addSuperProperty(ObjectId p){
    /*if(!superProperties.contains(p)){
      superProperties.add(p);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntProperty pp = ont.getProperty(p);
      if(pp!=null&&!pp.equals(this))
        pp.addSubProperty(this.getID());
    }*/

	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBPROPERTYOF));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(p))
	    		return;
	    }
	    }

	    Axiom axiom = new Axiom(Ontology.SUBPROPERTYOF);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(p);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(p, axiom);
		ont.addAxiomByType(new Integer(Ontology.SUBPROPERTYOF), axiom);

  }

  /**
   * get the direct super Properties
   */
  public ObjectId[] getDirectSuperProperties(){
    /*int len = superProperties.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = superProperties.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.SUBPROPERTYOF && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   		od.add(a.getRightAxiomNode().getAtomicNode());
	}
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * delete a direct super Property
   */
  public void deleteSuperProperty(ObjectId p){
    /*if(superProperties.contains(p)){
      superProperties.remove(p);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntProperty pp = ont.getProperty(p);
      if(pp!=null)
        pp.deleteSubProperty(this.getID());
    }*/
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.SUBPROPERTYOF &&
			a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(p)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(p).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.SUBPROPERTYOF)).remove(a);
	   		break;
	   	}
	}
	}
  }

  /**
   * Test if this property is deprecated.
   */
  public boolean isDeprecated(){
    return deprecated;
  }

  /**
   * Set the property to be deprecated.
   */
  public void setDeprecated(boolean dep){
    deprecated = dep;
  }

  /**
   * Test if this property is functional.
   */
  public boolean isFunctional(){
    return functional;
  }

  /**
   * Set the property to be functional.
   */
  public void setFunctional(boolean f){
    functional = f;
  }

  /**
   * Test if this property is inverse functional.
   */
  public boolean isInverseFunctional(){
    return inverseFunctional;
  }

  /**
   * Set the property to be inverse functional.
   */
  public void setInverseFuncional(boolean f){
    inverseFunctional = f;
  }

  /**
   * Add a 'inverseOf' property
   */
  public void addInverseOf(ObjectId p){
    /*if(!inverseOf.contains(p))
      inverseOf.add(p);*/

	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.INVERSEOF));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(p))
	    		return;
	    }
	    }

	    Axiom axiom = new Axiom(Ontology.INVERSEOF);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(p);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(p, axiom);
		ont.addAxiomByType(new Integer(Ontology.INVERSEOF), axiom);
  }

  /**
   * delete a 'inverseOf' property
   */
  public void deleteInverseOf(ObjectId p){
    //inverseOf.remove(p);
    Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.INVERSEOF &&
			a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(p)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(p).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.INVERSEOF)).remove(a);
	   		break;
	   	}
	}
	}
  }

  /**
   * get the 'inverseOf' properties
   */
  public ObjectId[] getInverseOf(){
    /*int len = inverseOf.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = inverseOf.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.INVERSEOF && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   		od.add(a.getRightAxiomNode().getAtomicNode());
	    }
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * Add a equivalent property.
   */
  public void addEquivalentProperty(ObjectId ec){
    /*if(!equivalent.contains(ec)&&!ec.equals(this.getID()))
      equivalent.add(ec);*/
	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.EQUIVALENTPROPERTY));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE &&
	    	   ((temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) && temp.getRightAxiomNode().getAtomicNode().equals(ec)) ||
	    	    (temp.getLeftAxiomNode().getAtomicNode().equals(ec) && temp.getRightAxiomNode().getAtomicNode().equals(this.getID()))) )
	    		return;
	    }
	    }
	    Axiom axiom = new Axiom(Ontology.EQUIVALENTPROPERTY);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(ec);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(ec, axiom);
		ont.addAxiomByType(new Integer(Ontology.EQUIVALENTPROPERTY), axiom);
  }

  /**
   * Delete a equivalent property.
   */
  public void deleteEquivalentProperty(ObjectId r){
    //equivalent.remove(r);
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTPROPERTY &&
			a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(r)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(r).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.EQUIVALENTPROPERTY)).remove(a);
	   		return;
	   	}
	}
	}
	axioms = (Vector<Axiom>)ont.getAxiomByObject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTPROPERTY &&
			a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getLeftAxiomNode().getAtomicNode().equals(r)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(r).remove(a);
	   		ont.getAxiomByObject(this.getID()).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.EQUIVALENTPROPERTY)).remove(a);
	   		return;
	   	}
	}
	}
  }

  /**
   * Get the equivalent properties.
   */
  public ObjectId[] getEquivalentProperties(){
    /*int len = equivalent.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = equivalent.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();

	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTPROPERTY && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   			od.add(a.getRightAxiomNode().getAtomicNode());
	}
	}
	axioms = (Vector<Axiom>)ont.getAxiomByObject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTPROPERTY && a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   			od.add(a.getLeftAxiomNode().getAtomicNode());
	}
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }


  /**
   * Tell this property is ObjectProperty or DatatypeProperty.
   * Null means unknown.
   */
  public Boolean isDatatype(){
    return datatype;
  }

  /**
   * Set this property to be DatatypeProperty or ObjectProperty.
   */
  public void setDatatype(boolean b){
    datatype = new Boolean(b);
  }
  /**
   * Tell this property is AnnotationProperty or not.
   *
   */
  public boolean isAnnotation(){
    return annotation;
  }

  /**
   * Set this property to be DatatypeProperty or not.
   */
  public void setAnnotation(boolean b){
    annotation = b;
  }

  private boolean transitive, symetric, functional, inverseFunctional,
                   deprecated, annotation;
  private Boolean datatype;

  private Vector inverseOf, subProperties, superProperties,equivalent;
  private ObjectId domain, range;



}
