/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.storage.simplememory;

import java.util.*;

import edu.lehigh.swat.hawk.core.ObjectId;
import edu.lehigh.swat.hawk.core.DocId;
import edu.lehigh.swat.hawk.core.Axiom;
import edu.lehigh.swat.hawk.core.AxiomNode;
import edu.lehigh.swat.hawk.fol.HornRule;
import edu.lehigh.swat.hawk.fol.Atom;
//import edu.lehigh.swat.hawk.core.Annotation;

/**
 * A simple implementation of core.ontology in memory.
 * @author Z. Pan
 * @version 1.0
 */

public class Ontology extends Document implements edu.lehigh.swat.hawk.core.Ontology{

    private static final long serialVersionUID = 1L;

  /**
   * Constructor
   */
  public Ontology() {
	axioms = new Vector<Axiom>();
	axiomsBySubject = new Hashtable<ObjectId, Vector<Axiom>>();
	axiomsByObject = new Hashtable<ObjectId, Vector<Axiom>>();
	axiomsByType = new Hashtable<Integer, Vector<Axiom>>();
	axiomNodeRestrictions = new Vector<AxiomNode>();
    classes = new Hashtable();
    properties = new Hashtable();
    restrictions = new Hashtable();
    compatiblePriorOnts = new Vector();
    priorOnts = new Vector();
    version = null;
    importedClasses = new Hashtable();
    importedProperties = new Hashtable();
    classIds = new Vector();
    propertyIds = new Vector();
    assertedClasses = new Vector();
    assertedProperties = new Vector();
    annotations = new Vector<edu.lehigh.swat.hawk.core.Annotation>();
    annsBySubject = new Hashtable<ObjectId, Vector<edu.lehigh.swat.hawk.core.Annotation>>();
    annsByPredicate = new Hashtable<ObjectId, Vector<edu.lehigh.swat.hawk.core.Annotation>>();
  }

  public void addAxiom(Axiom a){
      if(!axioms.contains(a))
          axioms.add(a);
  }

  public List<Axiom> getAllAxioms(){
	  return axioms;
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsBySubject(){
	  return axiomsBySubject;
  }

  public Hashtable<Integer, Vector<Axiom>> getAllAxiomsByType(){
	  return axiomsByType;
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsByObject(){
      return axiomsByObject;
  }

  public void addAxiomBySubject(ObjectId sub, Axiom a){
	  if(axiomsBySubject.containsKey(sub))
		  axiomsBySubject.get(sub).add(a);
	  else{
		  Vector<Axiom> axm = new Vector<Axiom>();
		  axm.add(a);
		  axiomsBySubject.put(sub, axm);
	  }
  }

  public void addAxiomByObject(ObjectId obj, Axiom a){
	  if(axiomsByObject.containsKey(obj))
		  axiomsByObject.get(obj).add(a);
	  else{
		  Vector<Axiom> axm = new Vector<Axiom>();
		  axm.add(a);
		  axiomsByObject.put(obj, axm);
	  }
  }

  public void addAxiomByType(Integer type, Axiom a){
	  if(axiomsByType.containsKey(type)){
              if (!axiomsByType.get(type).contains(a))
                  axiomsByType.get(type).add(a);
          }
	  else{
		  Vector<Axiom> axm = new Vector<Axiom>();
		  axm.add(a);
		  axiomsByType.put(type, axm);
	  }
  }

  public List<Axiom> getAxiomBySubject(ObjectId subject){
	  if(subject == null)
		  return null;
	  if(axiomsBySubject.containsKey(subject))
		  return axiomsBySubject.get(subject);
	  return null;
 }

  public List<Axiom> getAxiomByObject(ObjectId object){
	  if(object == null)
		  return null;
	  if(axiomsByObject.containsKey(object))
		  return axiomsByObject.get(object);
	  return null;
  }

  public List<Axiom> getAxiomByType(Integer type){
	  if(axiomsByType.containsKey(type))
		  return axiomsByType.get(type);
	  return null;
  }

  public Vector<AxiomNode> getAxiomNodeRestrictions(){
	  return axiomNodeRestrictions;
  }

  public void addAxiomNodeRestriction(AxiomNode res){
	  axiomNodeRestrictions.add(res);
  }

  public void addAllAxiomNodeRestrictions(Vector<AxiomNode> res){
	  axiomNodeRestrictions.addAll(res);
  }

  /**
   * add a class to this ontology
   */
  public void addOntClass(edu.lehigh.swat.hawk.core.OntClass c){
    classes.put(c.getID(),c);
    classIds.add(c.getID());
  }

  /**
   *
   * get the index number for a class
   */
  public int getClassIndex(ObjectId cid){
      return classIds.indexOf(cid)+1;
  }

  /**
   * get the object (class/property) given the ID
   */
  public edu.lehigh.swat.hawk.core.OntObject getOntObject(ObjectId id){
    if(classes.containsKey(id))
    	return (edu.lehigh.swat.hawk.core.OntClass)classes.get(id);
    else if(importedClasses.containsKey(id))
    	return (edu.lehigh.swat.hawk.core.OntClass) importedClasses.get(id);
    else if(properties.containsKey(id))
    	return (edu.lehigh.swat.hawk.core.OntProperty)properties.get(id);
    else if(importedProperties.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntProperty)importedProperties.get(id);
    return null;
  }

  /**
   * get the specified class
   */
  public edu.lehigh.swat.hawk.core.OntClass getOntClass(ObjectId id){
    if(classes.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntClass)classes.get(id);
    if(importedClasses.containsKey(id)){
      return (edu.lehigh.swat.hawk.core.OntClass) importedClasses.get(id);
    }
    return null;
  }

  /**
   * delete the specified class
   */
  public void deleteOntClass(ObjectId id){
    if(classes.containsKey(id))
      classes.remove(id);
  }

  /**
   * get all the classes in this ontology
   */
  public edu.lehigh.swat.hawk.core.OntClass[] getAllClasses(){
    Collection c = classes.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.OntClass[] oc = new edu.lehigh.swat.hawk.core.OntClass[len];
    System.arraycopy(oa,0,oc,0,len);
    return oc;

  }

  /**
   * Get the Iterator of all the classes.
   */
  public Iterator getClassIterator(){
    return classes.values().iterator();
  }


  /**
   * add a property to this ontology
   */
  public void addProperty(edu.lehigh.swat.hawk.core.OntProperty p){
    properties.put(p.getID(),p);
    propertyIds.add(p.getID());
  }

  /**
   *
   * get the index number for a property
   */
  public int getPropertyIndex(ObjectId pid){
      int i = propertyIds.indexOf(pid)+1;
      String s = "-"+String.valueOf(i);
      return Integer.parseInt(s);
  }

  /**
   * get the specified property
   */
  public edu.lehigh.swat.hawk.core.OntProperty getProperty(ObjectId id){
    if(properties.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntProperty)properties.get(id);
    else if(importedProperties.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntProperty)importedProperties.get(id);
    return null;

  }

  /**
   * delete a specified property
   */
  public void deleteProperty(ObjectId id){
    properties.remove(id);
  }

  /**
   * get all the properties in this ontology
   */
  public edu.lehigh.swat.hawk.core.OntProperty[] getAllProperties(){
    Collection c = properties.values();
    Object[] oa = c.toArray();
    int len = oa.length;;
    edu.lehigh.swat.hawk.core.OntProperty[] oc = new edu.lehigh.swat.hawk.core.OntProperty[len];
    System.arraycopy(oa,0,oc,0,len);
    return oc;
  }

  /**
   * Get the Iterator of all the properties.
   */
  public Iterator getPropertyIterator(){
    return properties.values().iterator();
  }


  /**
   * add a property to this ontology
   */
  public void addRestriction(edu.lehigh.swat.hawk.core.PropertyRestriction p){
    restrictions.put(p.getID(),p);
  }

  /**
   * get the specified property
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction getRestriction(ObjectId id){
    if(restrictions.containsKey(id))
      return (edu.lehigh.swat.hawk.core.PropertyRestriction)restrictions.get(id);
    else
      return null;
  }

  /**
   * delete a specified property
   */
  public void deleteRestriction(ObjectId id){
    restrictions.remove(id);
  }

  /**
   * get all the properties in this ontology
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction[] getAllRestrictions(){
    Collection c = restrictions.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.PropertyRestriction[] oc = new edu.lehigh.swat.hawk.core.PropertyRestriction[len];
    System.arraycopy(oa,0,oc,0,len);
    return oc;
  }

  /**
   * Get the priorVersion ontologies which are backwardCompatibleWith
   */
  public DocId[] getCompatiblePriorOnts(){
    if(compatiblePriorOnts.size()<=0)
      return null;
    else{
     Object[] oa = compatiblePriorOnts.toArray();
     int len = oa.length;
     DocId[] oc = new DocId[len];
     System.arraycopy(oa,0,oc,0,len);
     return oc;
    }
  }

  /**
   * Add a priorVersion ontology which is backwardCompatibleWith.
   */
  public void addCompatiblePriorOnts(DocId id){
    compatiblePriorOnts.add(id);
  }

  /**
   * Delete a priorVersion ontology which is backwardCompatibleWith.
   */
  public void deleteCompatiblePriorOnts(DocId id){
    compatiblePriorOnts.remove(id);
  }

  /**
   * Get the priorVersion ontologies may be incompatibleWith
   */
  public DocId[] getPriorOnts(){
    if(priorOnts.size()<=0)
      return null;
    else{
     Object[] oa = priorOnts.toArray();
     int len = oa.length;
     DocId[] oc = new DocId[len];
     System.arraycopy(oa,0,oc,0,len);
     return oc;
    }
  }

  /**
   * Add a priorVersion ontology may be incompatibleWith.
   */
  public void addPriorOnts(DocId id){
    priorOnts.add(id);
  }

  /**
   * Delete a priorVersion ontology may be incompatibleWith.
   */
  public void deletePriorOnts(DocId id){
    priorOnts.remove(id);
  }

  /**
   * get the version of this ontology
   */
  public String getVersion(){
    return version;
  }

  /**
   * Set the version of this ontology.
   */
  public void setVersion(String ver){
    version = ver;
  }

  /**
   * Create a Class object and set its ontology.
   */
  public edu.lehigh.swat.hawk.core.OntClass createClass(){
    OntClass cls = new OntClass();
    cls.setOntology(this);
    return cls;
  }

  /**
   * Create a Property object and set its ontology.
   */
  public edu.lehigh.swat.hawk.core.OntProperty createProperty(){
    OntProperty property = new OntProperty();
    property.setOntology(this);
    return property;

  }

  /**
   * Create a PropertyRestriction object and set its ontology.
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction createRestriction(){
    PropertyRestriction pr = new PropertyRestriction();
    pr.setOntology(this);
    return pr;
  }

  /**
   * To see if ontology has the class.
   */
  public boolean hasClass(String name){
    Enumeration e = classes.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(oid.getLocalName().equalsIgnoreCase(name))
        return true;
    }
    return false;
  }

  /**
   * To see if ontology has the property.
   */
  public boolean hasProperty(String name){
    Enumeration e = properties.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(oid.getLocalName().equalsIgnoreCase(name))
        return true;
    }
    return false;
  }

  /**
   * Add an imported class.
   */
  public void addImportedClass(ObjectId cid, edu.lehigh.swat.hawk.core.OntClass c){
    importedClasses.put(cid,c);
  }

  /**
   * Add an imported property.
   */
  public void addImportedProperty(ObjectId pid, edu.lehigh.swat.hawk.core.OntProperty p){
    importedProperties.put(pid,p);
  }


  /**
   * Get All the classes plus the imported Classes.
   */
  public edu.lehigh.swat.hawk.core.OntClass[] getAllPlusImportedClasses(){
    edu.lehigh.swat.hawk.core.OntClass[] ac = this.getAllClasses();
    int num = ac.length;
    Collection c = importedClasses.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.OntClass[] oc = new edu.lehigh.swat.hawk.core.OntClass[len+num];
    System.arraycopy(ac,0,oc,0,num);
    System.arraycopy(oa,0,oc,num,len);
    return oc;
  }

  /**
   * Get All the classes plus the imported properties.
   */
  public edu.lehigh.swat.hawk.core.OntProperty[] getAllPlusImportedProperties(){
    edu.lehigh.swat.hawk.core.OntProperty[] ac = this.getAllProperties();
    int num = ac.length;
    Collection c = importedProperties.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.OntProperty[] oc = new edu.lehigh.swat.hawk.core.OntProperty[len+num];
    System.arraycopy(ac,0,oc,0,num);
    System.arraycopy(oa,0,oc,num,len);
    return oc;
  }

  /**
   * To tell if the ontology imorted specified class.
   */
  public boolean importedClass(ObjectId cid){
    Enumeration e = importedClasses.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(cid.equals(oid))
        return true;
    }
    return false;
  }

  /**
   * To tell if the ontology imported specified property.
   */
  public boolean importedProperty(ObjectId pid){
    Enumeration e = importedProperties.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(pid.equals(oid))
        return true;
    }
    return false;
  }

  /**
   * To tell if the ontology asserted specified class.
   */
  public boolean assertedClass(ObjectId cid){
      return assertedClasses.contains(cid);
  }

  /**
   * To tell if the ontology asserted specified property.
   */
  public boolean assertedProperty(ObjectId pid){
      return assertedProperties.contains(pid);
  }

  /**
   * Add an asserted class.
   */
  public void addAssertedClass(ObjectId cid){
      if(!assertedClasses.contains(cid))
          assertedClasses.add(cid);
  }

  /**
   * Add an asserted property.
   */
  public void addAssertedProperty(ObjectId pid){
      if(!assertedProperties.contains(pid))
      assertedProperties.add(pid);
  }


  /**
   * Get the imported class.
   */
//  public edu.lehigh.swat.hawk.core.OntClass getImportedClass(ObjectId id){
//    if(importedClasses.containsKey(id))
//      return (edu.lehigh.swat.hawk.core.OntClass)importedClasses.get(id);
//    else
//      return null;
//  }

  /**
   * Get the imported property.
   */
//  public edu.lehigh.swat.hawk.core.OntProperty getImportedProperty(ObjectId id){
//    if(importedProperties.containsKey(id))
//      return (edu.lehigh.swat.hawk.core.OntProperty)importedProperties.get(id);
//    else
//      return null;
//  }

  /**
   * Add annotations.
   */
  public boolean addAnnotations(ObjectId sId, ObjectId pId, String value) {
	  Vector<edu.lehigh.swat.hawk.core.Annotation> vecSubAnns = annsBySubject.get(sId);
	  for (int i=0; vecSubAnns != null && i<vecSubAnns.size(); i++) {
		  if (vecSubAnns.get(i).getPredicate().equals(pId) && vecSubAnns.get(i).getObject().getId().equals(value) )
			  return false;
	  }

	  Annotation temp = new Annotation(sId, pId, new ObjectId(value));
	//  temp.setSubject(sId);
	//  temp.setPredicate(pId);
	//  temp.setObject(new ObjectId(value));

	  annotations.add(temp);
	  if (annsBySubject.containsKey(sId))
		  annsBySubject.get(sId).add(temp);
	  else {
		  Vector<edu.lehigh.swat.hawk.core.Annotation> subVector = new Vector<edu.lehigh.swat.hawk.core.Annotation>();
		  subVector.add(temp);
		  annsBySubject.put(sId, subVector);
	  }

	  if (annsByPredicate.containsKey(pId))
		  annsByPredicate.get(pId).add(temp);
	  else {
		  Vector<edu.lehigh.swat.hawk.core.Annotation> objVector = new Vector<edu.lehigh.swat.hawk.core.Annotation>();
		  objVector.add(temp);
		  annsBySubject.put(pId, objVector);
	  }

	  return true;
  }
  /**
   * Get annotations by subject ID.
   */
  public Vector<edu.lehigh.swat.hawk.core.Annotation> getAnnotationsBySubject(ObjectId pid) {
	  Vector<edu.lehigh.swat.hawk.core.Annotation> temp = null;
	  if (annsBySubject.containsKey(pid)) {
		  return annsBySubject.get(pid);
	  }
	  return temp;
  }
  /**
   * Get annotations by predicate ID.
   */
  public Vector<edu.lehigh.swat.hawk.core.Annotation> getAnnotationsByObject(ObjectId pid) {
	  Vector<edu.lehigh.swat.hawk.core.Annotation> temp = null;
	  if (annsByPredicate.containsKey(pid)) {
		  return annsByPredicate.get(pid);
	  }
	  return temp;
  }
  /**
   * Get annotations by object ID.
   */
  public Vector<edu.lehigh.swat.hawk.core.Annotation> getAllAnnotations() {
	  return annotations;
  }

  public void addImportedAxiom(Axiom a){
      importedAxioms.add(a);
      if(a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE)
          addImportedAxiomBySubject(a.getLeftAxiomNode().getAtomicNode(), a);
      if(a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
          addImportedAxiomByObject(a.getRightAxiomNode().getAtomicNode(), a);

      addImportedAxiomByType(new Integer(a.getType()), a);
  }

  public List<Axiom> getAllPlusImportedAxioms(){

      Vector<Axiom> allPlusImportedAxioms = new Vector<Axiom>();
      allPlusImportedAxioms.addAll(axioms);
      allPlusImportedAxioms.addAll(importedAxioms);
      return allPlusImportedAxioms;
  }

  public Hashtable<Integer, Vector<Axiom>> getAllPlusImportedAxiomsByType(){
      Hashtable<Integer, Vector<Axiom>> allPlusImportedAxiomsByType = new Hashtable<Integer, Vector<Axiom>>();
      allPlusImportedAxiomsByType.putAll(axiomsByType);
      allPlusImportedAxiomsByType.putAll(importedAxiomsByType);
      return allPlusImportedAxiomsByType;
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllPlusImportedAxiomsByObject(){
      Hashtable<ObjectId, Vector<Axiom>> allPlusImportedAxiomsByObject = new Hashtable<ObjectId, Vector<Axiom>>();
      allPlusImportedAxiomsByObject.putAll(axiomsByObject);
      allPlusImportedAxiomsByObject.putAll(importedAxiomsByObject);
      return allPlusImportedAxiomsByObject;
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllPlusImportedAxiomsBySubject(){

      Hashtable<ObjectId, Vector<Axiom>> allPlusImportedAxiomsBySubject = new Hashtable<ObjectId, Vector<Axiom>>();
      allPlusImportedAxiomsBySubject.putAll(axiomsBySubject);
      allPlusImportedAxiomsBySubject.putAll(importedAxiomsBySubject);
      return allPlusImportedAxiomsBySubject;
  }

  public void addImportedAxiomByObject(ObjectId obj, Axiom a){
      if(importedAxiomsByObject.containsKey(obj)){
          if(!importedAxiomsByObject.get(obj).contains(a))
              importedAxiomsByObject.get(obj).add(a);
      }else{
          Vector<Axiom> axm = new Vector<Axiom>();
          axm.add(a);
          importedAxiomsByObject.put(obj, axm);
      }
  }

  public void addImportedAxiomByType(Integer type, Axiom a){
      if(importedAxiomsByType.containsKey(type)){
          if(!importedAxiomsByType.get(type).contains(a))
              importedAxiomsByType.get(type).add(a);
      }else{
          Vector<Axiom> axm = new Vector<Axiom>();
          axm.add(a);
          importedAxiomsByType.put(type, axm);
      }
  }

  public void addImportedAxiomBySubject(ObjectId sub, Axiom a){
      if(importedAxiomsBySubject.containsKey(sub)){
          if(!importedAxiomsBySubject.get(sub).contains(a))
              importedAxiomsBySubject.get(sub).add(a);
      }else{
          Vector<Axiom> axm = new Vector<Axiom>();
          axm.add(a);
          importedAxiomsBySubject.put(sub, axm);
      }
  }



  public void convertAxiomsToRules(){
      System.out.println("all axiom size: "+getAllAxioms().size());
      List<Axiom> equiaxioms = this.getAllAxiomsByType().get(edu.lehigh.swat.hawk.core.Ontology.EQUIVALENTCLASS);
      if(equiaxioms!=null){
          Iterator<Axiom> it = equiaxioms.iterator();
          while (it.hasNext()) {
              Axiom axiom = it.next();
              AxiomNode left = axiom.getLeftAxiomNode();
              AxiomNode right = axiom.getRightAxiomNode();
              if (left.getJoinType() == Ontology.ATOMICNODE &&
                  right.getJoinType() == Ontology.ATOMICNODE)
                  continue;
              Axiom sub1 = new Axiom(Ontology.SUBCLASSOF);
              sub1.setLeftAxiomNode(axiom.getLeftAxiomNode());
              sub1.setRightAxiomNode(axiom.getRightAxiomNode());

              Axiom sub2 = new Axiom(Ontology.SUBCLASSOF);
              sub2.setRightAxiomNode(axiom.getLeftAxiomNode());
              sub2.setLeftAxiomNode(axiom.getRightAxiomNode());

              addAxiom(sub1);
              addAxiomByType(Integer.valueOf(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF), sub1);

              addAxiom(sub2);
              addAxiomByType(Integer.valueOf(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF), sub2);

          }
      }

      List<Axiom> subaxioms = this.getAllAxiomsByType().get(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF);
      if(subaxioms!=null){
          int ss = subaxioms.size();
          for(int j=0;j<ss;j++){
              Axiom axiom = subaxioms.get(j);
              AxiomNode left = axiom.getLeftAxiomNode();
              AxiomNode right = axiom.getRightAxiomNode();
              // if left is union, split LHS
              if(left.getJoinType()==edu.lehigh.swat.hawk.core.Ontology.UNIONOF){
                  Vector<AxiomNode> childern = left.getChildren();
                  int cs = childern.size();
                  for(int i=0;i<cs;i++){
                      Axiom sub1 = new Axiom(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF);
                      sub1.setLeftAxiomNode(childern.get(i));
                      sub1.setRightAxiomNode(right);
                      addAxiom(sub1);
                      addAxiomByType(Integer.valueOf(edu.lehigh.swat.hawk.core.
                              Ontology.SUBCLASSOF), sub1);
                  }
              }
              // if rigth is intersection, split RHS
              if(right.getJoinType()==Ontology.INTERSECTIONOF){
                  Vector<AxiomNode> childern = right.getChildren();
                  int cs = childern.size();
                  for(int i=0;i<cs;i++){
                      Axiom sub1 = new Axiom(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF);
                      sub1.setLeftAxiomNode(left);
                      sub1.setRightAxiomNode(childern.get(i));
                      addAxiom(sub1);
                      addAxiomByType(Integer.valueOf(edu.lehigh.swat.hawk.core.
                              Ontology.SUBCLASSOF), sub1);
                  }
              }
          }
      }

      rawRules = new Hashtable();
      subaxioms = this.getAllAxiomsByType().get(edu.lehigh.swat.hawk.core.Ontology.SUBCLASSOF);
      //System.out.println("subc size: "+subaxioms.size());
      if(subaxioms!=null){
          Iterator<Axiom> it = subaxioms.iterator();
          while (it.hasNext()) {
              Axiom axiom = it.next();
              AxiomNode left = axiom.getLeftAxiomNode();
              AxiomNode right = axiom.getRightAxiomNode();

              if(right.getJoinType()==Ontology.ATOMICNODE){
                  ObjectId rid = right.getAtomicNode();
                  if(rid.getNameSpace().equals("")||rid.getNameSpace().startsWith("http://www.w3.org/2002/07/owl"))
                      continue;
                  HornRule rule = new HornRule();
                  Atom head = new Atom(rid.getId(), rule.getVariable());
                  rule.setHead(head);
                  if(!nodeToAtoms(left, rule, rule.getVariable()))
                      continue;
                  System.out.println(rule.toString());
                  if(rawRules.get(head.getPredicate())==null){
                      Vector<HornRule> commonHeadRules = new Vector<HornRule>();
                      commonHeadRules.add(rule);
                      rawRules.put(head.getPredicate(), commonHeadRules);
                  }else{
                      Vector<HornRule> commonHeadRules = rawRules.get(head.getPredicate());
                      commonHeadRules.add(rule);
                      rawRules.put(head.getPredicate(), commonHeadRules);
                  }
              }else if(right.getJoinType()==Ontology.ALLVALUERESTRICTION){
                  AxiomNode cls = right.getAllValuesFrom();
                  AxiomNode prop = right.getOnProperty();
                  if(cls.getAtomicNode()==null||prop.getAtomicNode()==null)
                      continue;
                  ObjectId cid = cls.getAtomicNode();
                  ObjectId pid = prop.getAtomicNode();
                 if(cid.getNameSpace().equals("")||cid.getNameSpace().startsWith("http://www.w3.org/2002/07/owl"))
                     continue;
                 HornRule rule = new HornRule();
                 Atom head = new Atom(cid.getId(), HornRule.X);
                 rule.setHead(head);
                 Atom lprop = new Atom(pid.getId(), rule.getNewVariable(), HornRule.X);
                 rule.addToBody(lprop);
                 if(!nodeToAtoms(left, rule, rule.getVariable()))
                     continue;
                 System.out.println(rule.toString());
                 if(rawRules.get(head.getPredicate())==null){
                     Vector<HornRule> commonHeadRules = new Vector<HornRule>();
                     commonHeadRules.add(rule);
                     rawRules.put(head.getPredicate(), commonHeadRules);
                 }else{
                     Vector<HornRule> commonHeadRules = rawRules.get(head.getPredicate());
                     commonHeadRules.add(rule);
                     rawRules.put(head.getPredicate(), commonHeadRules);
                 }
              }
          }
      }


      /*properties. Maybe the views work, so FOL is not necessary.
      List<Axiom> subpaxioms = this.getAllAxiomsByType().get(edu.lehigh.swat.hawk.core.Ontology.SUBPROPERTYOF);
      if(subaxioms!=null){
          Iterator<Axiom> it = subaxioms.iterator();
          while (it.hasNext()) {
              Axiom axiom = it.next();
              AxiomNode left = axiom.getLeftAxiomNode();
              AxiomNode right = axiom.getRightAxiomNode();
          }
      }*/
  }

  protected boolean nodeToAtoms(AxiomNode node, HornRule rule, int var){
      switch(node.getJoinType()){

      case Ontology.ATOMICNODE:
          if(node.getAtomicNode().getNameSpace().equals(""))
              return false;
          Atom b = new Atom(node.getAtomicNode().getId(), var);
          rule.addToBody(b);
          break;

      case Ontology.INTERSECTIONOF:
          Vector<AxiomNode> children = node.getChildren();
          Iterator<AxiomNode> it = children.iterator();
          while (it.hasNext()) {
              AxiomNode child = it.next();
              if(!nodeToAtoms(child, rule, var))
                  return false;
          }
          break;

      case Ontology.SOMEVALUERESTRICTION:
          AxiomNode property = node.getOnProperty();
          if (property.getJoinType() == Ontology.ATOMICNODE) {
              Atom a = new Atom(property.getAtomicNode().getId(),
                                rule.getVariable(), rule.getNewVariable());
              rule.addToBody(a);
          } else {
              //inverse property
          }
          AxiomNode restr = node.getSomeValuesFrom();
          if(!nodeToAtoms(restr, rule, rule.getVariable()))
              return false;
          break;

      default:
         return false;
      }

      return true;
  }

  public void expandRules(){
      expandedAtoms = new Hashtable<String, Vector<Vector<Atom>>>();
      cyclicRules = new Hashtable<String, Vector<Vector<Atom>>>();
      Enumeration enu = rawRules.keys();
      while (enu.hasMoreElements()) {
          String pred = (String) enu.nextElement();
          Atom p = new Atom(pred, HornRule.X);
          Vector<Vector<Atom>> cycleDisj = new Vector<Vector<Atom>>();
          inPlay = new Vector<String>();
          Vector<Vector<Atom>> DNF = expandAtom(p);
          expandedAtoms.put(pred, DNF);
          Iterator<Vector<Atom>> dit = DNF.iterator();
          while (dit.hasNext()) {
              Vector<Atom> conj = dit.next();
              Iterator<Atom> cit = conj.iterator();
              while (cit.hasNext()) {
                  Atom atom = cit.next();
                  if (atom.getPredicate().equals(p.getPredicate())
                      &&(atom.getArguments()[0] != p.getArguments()[0] ||
                       atom.getArguments()[1] != p.getArguments()[1])) {
                      cycleDisj.add(conj);
                  }
              }
          }
          if (!cycleDisj.isEmpty()) {
              cyclicRules.put(pred, cycleDisj);
              DNF.removeAll(cycleDisj);
              System.out.println("---Cyclic axioms found for " +
                                 pred);
          }
      }
  }

  //returns a deep copy of the DNF stored in expandedAtom, this copy is expected to be changed.
  protected Vector<Vector<Atom>> expandAtom(Atom a){
//      System.out.println("expanding... "+a.getPredicate());
/*
      if(expandedAtoms.get(a.getPredicate())!=null){
           Vector<Vector<Atom>> nochange = expandedAtoms.get(a.getPredicate());
           Vector<Vector<Atom>> temp = new Vector<Vector<Atom>>();
           int diff = a.getFirstArgument()-HornRule.X;
           for(Iterator<Vector<Atom>> i = nochange.iterator();i.hasNext();){
               Vector<Atom> nochangev = i.next();
               Vector<Atom> tempv = new Vector<Atom>();
               for(Iterator<Atom> j = nochangev.iterator(); j.hasNext(); ){
                   Atom nochangea = j.next();
                   Atom tempa;
                   if(nochangea.isBinary()){
                       tempa = new Atom(nochangea.getPredicate(),
                                        nochangea.getFirstArgument() + diff,
                                        nochangea.getSecondArgument() + diff);
                       tempa.setPrimitive(nochangea.isPrimitive());
                   }
                   else{
                       tempa = new Atom(nochangea.getPredicate(),
                                        nochangea.getFirstArgument() + diff);
                       tempa.setPrimitive(nochangea.isPrimitive());
                   }
                   tempv.add(tempa);
               }
               temp.add(tempv);
            }
//                  System.out.println("Done expanding... "+a.getPredicate());
           return temp;
      }
      else{
*/
          Vector<Vector<Atom>> expanded = new Vector();
          Vector<Atom> va = new Vector<Atom>();
          va.add(a);
          expanded.add(va);
          Vector<HornRule> rr = rawRules.get(a.getPredicate());
          if(rr==null)
              return expanded;

//          expandedAtoms.put(a.getPredicate(), expanded);

          Iterator<HornRule> it = rr.iterator();
          inPlay.add(a.getPredicate());
          int diff = a.getFirstArgument()-HornRule.X;
          while(it.hasNext()){
              HornRule rule = it.next();
              Vector<Atom> body = rule.getBody();
              Iterator<Atom> ait = body.iterator();

              Vector<Vector<Atom>> disj = null;
              while(ait.hasNext()){
                  Atom ai = ait.next();
                  Atom b;
                  if(ai.isBinary())
                    b = new Atom(ai.getPredicate(), ai.getFirstArgument()+diff, ai.getSecondArgument()+diff);
                  else
                    b = new Atom(ai.getPredicate(), ai.getFirstArgument()+diff);

                  if(b.equals(a))
                      break;
                  Vector<Vector<Atom>> holder;
                  if(b.getPredicate().equals(a.getPredicate())||inPlay.contains(b.getPredicate())||b.isBinary()){
                      holder = new Vector<Vector<Atom>>();
                      Vector<Atom> holderc = new Vector<Atom>();
                      Atom bnp = new Atom(b.getPredicate(), b.getArguments());
                      bnp.setPrimitive(false);
                      holderc.add(bnp);
                      holder.add(holderc);
                  }
                  else{
                     holder =  expandAtom(b);
//                     boolean hasPredicate = false;
                     boolean allPrimitive = true;
                     int hs = holder.size();
                     for(int k=0;k<hs;k++){
//                         if(hasPredicate(holder.get(k), a.getPredicate()))
//                             hasPredicate = true;
                         if(!allPrimitive(holder.get(k)))
                             allPrimitive = false;
                         if(holder.get(k).contains(a)){   //no tautology here!
                             holder.remove(k);
                             k--;
                             hs--;
                         }
                     }
/*
                    if(!hasPredicate&&allPrimitive&&hs>2){
                         holder = new Vector<Vector<Atom>>();
                         Vector<Atom> holderc = new Vector<Atom>();
                         Atom bnp = new Atom(b.getPredicate(), b.getArguments());
                         bnp.setPrimitive(false);
                         holderc.add(bnp);
                         holder.add(holderc);
                     }
 */

                    if((hs>30||(allPrimitive&&hs>1)||(!a.getPredicate().endsWith("#Person")&&b.getPredicate().endsWith("#Person")))&&!inPlay.get(0).endsWith("#Person")){
                        holder = new Vector<Vector<Atom>>();
                        Vector<Atom> holderc = new Vector<Atom>();
                        Atom bnp = new Atom(b.getPredicate(), b.getArguments());
                        bnp.setPrimitive(false);
                        holderc.add(bnp);
                        holder.add(holderc);
                    }
                  }

                  if(disj==null){
                      disj = holder;
                      continue;
                  }
                  int size1 = disj.size();
                  int size2 = holder.size();
                  if(size1*size2>60){
                      System.out.println("expanding... " + a.getPredicate() +
                                         "  adding..." + b.getPredicate());
                      System.out.println("DNF size: " + size1 + "  new size: " +
                                         size2);
                  }
                  Vector<Vector<Atom>> tdnf = new Vector<Vector<Atom>>();
                  for(int i=0; i<size1; i++){
                      for(int j=0; j<size2; j++){
                          Vector<Atom> conj = (Vector<Atom>) disj.get(i).clone();
                          conj.addAll(holder.get(j));
                          tdnf.add(conj); //conj to DNF
                      }
                  }
                  disj = tdnf;
                  //conjunc
              }
              //add to DNF
              if(disj==null)
                  continue;
//              System.out.println("start eliminate duplicates");
//              System.out.println("disj size: "+disj.size()+" expanded size: "+expanded.size());
              for(Iterator<Vector<Atom>> djit = disj.iterator();djit.hasNext();){
                  Vector<Atom> conj1 = djit.next();
                  boolean isThere = false;
                  int es = expanded.size();
                  for(int m= 0; m<es; m++){
                      Vector<Atom> conj2 = expanded.get(m);
                      boolean sub1 = compareConjuncts(conj2,conj1);
                      boolean sub2 = compareConjuncts(conj1,conj2);
                      if(sub2){ //if conj2 \subseteq  conj1, meaning conj2 less restrictions, more members, include conj1
                          isThere = true;
                      }else if(sub1){ //if conj1 \subset conj2, meaning conj1 less restrctions, more members, include conj2
                          expanded.remove(conj2);
                          es--;
                          m--;
                      }
                  }
                  if(!isThere){
                      expanded.add(conj1);
                  }
              }
//              System.out.println("done eliminate duplicates");
          }
//          System.out.println("start copy & return");
          //expandedAtoms.put(a.getPredicate(), expanded);
        /*  int diff = a.getFirstArgument()-HornRule.X;
          Vector<Vector<Atom>> temp = new Vector<Vector<Atom>>();
          for(Iterator<Vector<Atom>> i = expanded.iterator();i.hasNext();){
              Vector<Atom> nochangev = i.next();
              Vector<Atom> tempv = new Vector<Atom>();
              for(Iterator<Atom> j = nochangev.iterator(); j.hasNext(); ){
                  Atom nochangea = j.next();
                  Atom tempa;
                  if(nochangea.isBinary()){
                      tempa = new Atom(nochangea.getPredicate(),
                                       nochangea.getFirstArgument() + diff,
                                       nochangea.getSecondArgument() + diff);
                      tempa.setPrimitive(nochangea.isPrimitive());
                  }
                  else{
                      tempa = new Atom(nochangea.getPredicate(),
                                       nochangea.getFirstArgument() + diff);
                      tempa.setPrimitive(nochangea.isPrimitive());
                  }
                   tempv.add(tempa);
              }
              temp.add(tempv);
          }
*/
          inPlay.remove(a.getPredicate());
          if(expanded.size()>60){
                System.out.println("Done expanding.. "+a.getPredicate()+" ||size: "+expanded.size());

            }
          return expanded;
//      }
  }

  /**
   * Determine if c2 is subset of or equivalent to c1
   * @param c1 Vector
   * @param c2 Vector
   * @return boolean
   */
  public boolean compareConjuncts(Vector<Atom> c1, Vector<Atom>c2){
      int size1 = c1.size();
      int size2 = c2.size();
      if(size1<size2)
          return false;
      for(int i=0; i<size2; i++){
          if(!c1.contains(c2.get(i)))
              return false;
      }
      return true;
  }

  public boolean hasPredicate(Vector<Atom> v1, String pred){
      for(int i=0; i<v1.size(); i++){
          if(v1.get(i).getPredicate().equalsIgnoreCase(pred))
              return true;
      }
      return false;
  }

  public boolean allPrimitive(Vector<Atom> v1){
      for(int i=0; i<v1.size(); i++){
          if(!v1.get(i).getPredicate().endsWith("#Person")&&!v1.get(i).isPrimitive()&&!v1.get(i).isBinary()&&!inPlay.contains(v1.get(i).getPredicate()))
              return false;
      }
      return true;
  }

  public Hashtable<String, Vector<Vector<Atom>>> getCyclicRules(){
	  return cyclicRules;
  }

  public Hashtable<String, Vector<Vector<Atom>>> getExpandedAtoms(){
	  return expandedAtoms;
  }

  private Hashtable classes, properties, restrictions,
  importedClasses, importedProperties;
  private Vector assertedClasses, assertedProperties;
  private Vector priorOnts, compatiblePriorOnts;
  private String version;
  private Vector classIds, propertyIds;
  private Vector<Axiom> axioms;
  private Hashtable<ObjectId, Vector<Axiom>> axiomsBySubject;
  private Hashtable<ObjectId, Vector<Axiom>> axiomsByObject;
  private Hashtable<Integer, Vector<Axiom>> axiomsByType;
  private Vector<AxiomNode> axiomNodeRestrictions;
  private Vector<edu.lehigh.swat.hawk.core.Annotation> annotations;
  private Hashtable<ObjectId, Vector<edu.lehigh.swat.hawk.core.Annotation>> annsBySubject;
  private Hashtable<ObjectId, Vector<edu.lehigh.swat.hawk.core.Annotation>> annsByPredicate;

  private Vector<Axiom> importedAxioms;
  private Hashtable<ObjectId, Vector<Axiom>> importedAxiomsBySubject;
  private Hashtable<ObjectId, Vector<Axiom>> importedAxiomsByObject;
  private Hashtable<Integer, Vector<Axiom>> importedAxiomsByType;


  private Vector<String> inPlay;
  private Hashtable<String, Vector<HornRule>> rawRules;
  private Hashtable<String, Vector<Vector<Atom>>> expandedAtoms;
  private Hashtable<String, Vector<Vector<Atom>>> cyclicRules;

}
